/* -*- c++ -*-
// ****************************************************************************
//  Name: jit_debug.cpp
//  Athr: Umit V. Catalyurek (UVC)
//  Date: 10/05/2006: UVC - created based on Mike Beynon's dc_debug.cpp,
//                          THANKS MIKE!!!!
//
//  Desc: Implementation of a dynamic array class.
// ****************************************************************************
*/
#include "jit_debug.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#define DEBUGGER_USE_GDB
#define DBG_JITD_USE_BINARY_IN_PATH


typedef unsigned int uint;

#define DeleteNull(v) { delete [] v; v = NULL; }
#define TestDelete(v) if (v) { delete [] v; }
#define TestDeleteNull(v) if (v) { delete [] v; v = NULL; }

using namespace std;

class Argv {
protected:
    char   **rgElem;
    uint    maxElem;
    uint    nElem;
    bool *rgfDelete;
    
    void Extend(void);
    
public:
    Argv(void);
    Argv(int argc, char *argv[], bool fDelete = false);
    ~Argv(void);

    void Copy(int argc, char *argv[]);
    void Set(int argc, char *argv[], bool fDelete = false);
//    void Set(DArray<char*> da, bool fDelete);
    int Append(char *sbNewArg, bool fDelete = false);
//    int Append(DArray<char*> da, bool fDelete = false);
    int AppendDup(char *sbNewArg) { return Append(strdup(sbNewArg), true); };
    void Flush(void);
  
    int argc(void) const            { return nElem; }
    char **argv(void)               { return rgElem; }
    char *Get(uint i)                    { return (i>maxElem) ? (char*)NULL : rgElem[i]; }
    const char *Get(uint i) const        { return (i>maxElem) ? (char*)NULL : rgElem[i]; }
    char *operator[](uint i)             { return Get(i); }
    const char *operator[](uint i) const { return Get(i); }
    
    bool Find(const char *sbFlag);
    char *NextFind(const char *sbFlag);
    bool Extract(const char *sbFlag);
    char *NextExtract(const char *sbFlag);
    
/*    int toString(DCString_t &strOut, bool fAddQuotes=true, int startArg=0) const; */
    
    friend std::ostream &operator<<(std::ostream &os, Argv &argv);

private:
    Argv(const Argv& a);
    Argv& operator=(const Argv& a);
};


#define ARGV_DefaultExtentSize 10


// ****************************************************************************
//  Class: Argv
// ****************************************************************************

Argv::Argv(void)
{
    nElem = maxElem = 0; 
    rgElem = NULL;
    rgfDelete = NULL;
}


Argv::Argv(int argc, char *argv[], bool fDelete )
{
    nElem = maxElem = 0; 
    rgElem = NULL;
    rgfDelete = NULL;
    Set(argc, argv, fDelete);
}


Argv::~Argv(void)
{ 
    Flush();
    TestDeleteNull(rgElem);
    TestDeleteNull(rgfDelete);
}

void Argv::Extend(void)
{
    uint nNewSize = maxElem + ARGV_DefaultExtentSize;
    char **rgTmp = rgElem;
    bool *rgfTmp = rgfDelete;
    rgElem = new char*[(maxElem = nNewSize)];
    rgfDelete = new bool[maxElem];
    if (rgTmp) {
        // copy over old elements
        for (uint i=0; i<nElem; i++) {
            rgElem[i] = rgTmp[i];
            rgfDelete[i] = rgfTmp[i];
        }
        rgElem[nElem] = NULL;
        delete [] rgTmp;
        delete [] rgfTmp;
    }
}


void Argv::Copy(int argc, char *argv[])
{
    Flush();
    TestDelete(rgElem);
    TestDelete(rgfDelete);
    nElem = argc;
    maxElem = nElem + 1;
    rgElem = new char*[maxElem];
    rgfDelete = new bool[maxElem];
    for (uint i=0; i<nElem; i++) {
        rgElem[i] = new char[strlen(argv[i])+1];
        strcpy(rgElem[i], argv[i]);
        rgfDelete[i] = true;
    }
    rgElem[nElem] = NULL;
}


// ****************************************************************************
//  Name: Set
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/**
 * Use a given argc and argv.
 * @param argc
 * @param argv - arg list to use
 */
void Argv::Set(int argc, char *argv[], bool fDelete)
{
    TestDelete(rgElem);
    TestDelete(rgfDelete);
    nElem = argc;
    maxElem = nElem + 1;
    rgElem = new char*[maxElem];
    rgfDelete = new bool[maxElem];
    for (uint i=0; i<nElem; i++) {
        rgElem[i] = argv[i];
        rgfDelete[i] = fDelete;
    }
    rgElem[nElem] = NULL;
}

/*
void Argv::Set(DArray<char*> da, bool fDelete)
{
    TestDelete(rgElem);
    TestDelete(rgfDelete);
    nElem = da.NumElem();
    maxElem = nElem + 1;
    rgElem = new char*[maxElem];
    rgfDelete = new bool[maxElem];
    for (uint i=0; i<nElem; i++) {
        rgElem[i] = da.Get(i);
        rgfDelete[i] = fDelete;
    }
    rgElem[nElem] = NULL;
}
*/

// ****************************************************************************
//  Name: Append
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/**
 * Add an arg to the end of the list.
 * @param sbNewArg - arg to add
 */
int Argv::Append(char *sbNewArg, bool fDelete)
{
    if (nElem+1 >= maxElem)
        Extend();
    
    rgElem[nElem] = sbNewArg;
    rgfDelete[nElem] = fDelete;
    nElem++;
    rgElem[nElem] = NULL;
    return nElem;
}

/*
int Argv::Append(DArray<char*> da, bool fDelete)
{
    uint n = da.NumElem();
    for (uint i=0; i<n; i++) {
        Append(da.Get(i), fDelete);
    }
    return nElem;
}
*/


// ****************************************************************************
//  Name: Flush
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/**
 * Empties all in list
 */
void Argv::Flush(void)
{
    if (rgElem != NULL) {
        for (uint i=0; i<nElem; i++) {
            if (rgfDelete[i]) TestDelete(rgElem[i]);
        }
        nElem = 0;
        rgElem[nElem] = NULL;
    }
}


// ****************************************************************************
//  Name: Find
//  Date: 06/12/1997 : mdb - Created.
//        07/23/1997 : mdb - Moved into Argv class.
// ****************************************************************************
/** 
 * Detects the occurrence of sbFlag.
 * @param sbFlag - flag to look for
 * @return bool - true=found, false=not found
 */
bool Argv::Find(const char *sbFlag)
{
    for (uint i=0; i<nElem; i++) {
        if (strcmp(sbFlag, rgElem[i]) == 0)
            return true;
    }
    return false;
}


// ****************************************************************************
//  Name: NextFind
//  Date: 05/19/1997 : mdb - Created.
//        07/23/1997 : mdb - Moved into Argv class.
// ****************************************************************************
/**
 * Find the first occurrence of sbFlag and the following argument.
 * @param sbFlag - flag to look for
 * @return char* - NULL=not found, else=argument
 */
char *Argv::NextFind(const char *sbFlag)
{
    for (uint i=0; i<nElem; i++) {
        if (strcmp(sbFlag, rgElem[i]) == 0) {  /* found? */
            return ((i+1 < nElem) ? rgElem[i+1] : (char*)NULL);
        }
    }
    return NULL;
}



// ****************************************************************************
//  Name: Extract
//  Date: 01/30/2003 : uvc - Created.
// ****************************************************************************
/** 
 * Remove the first occurrence of sbFlag 
 * @param sbFlag - flag to look for
 * @return bool - true=found, false=not found
 */
bool Argv::Extract(const char *sbFlag)
{
    bool fFound=false;
    uint i, j;
    
    /* i - the next one to look at
     * j - the next spot in result
     */
    for (i=0,j=0; i<nElem; i++) {
        if (!fFound && strcmp(sbFlag, rgElem[i]) == 0) {  /* found? */
            fFound = true;
        } else {
            rgElem[j] = rgElem[i];
            rgfDelete[j] = rgfDelete[i];
            j++;  /* leave in result argv */
        }
    }
    nElem -= (i-j);  /* reduce by how many skipped */
    rgElem[nElem] = NULL;
    
    return fFound;
}


// ****************************************************************************
//  Name: NextExtract
//  Date: 05/19/1997 : mdb - Created.
//        07/23/1997 : mdb - Moved into Argv class.
// ****************************************************************************
/** 
 * Remove the first occurrence of sbFlag and the following argument.
 * @param sbFlag - flag to look for
 * @return char* - NULL=not found, else=argument
 */
char *Argv::NextExtract(const char *sbFlag)
{
    char *sbRtn=NULL;
    uint i, j, fFound = 0;
    
    /* i - the next one to look at
     * j - the next spot in result
     */
    for (i=0,j=0; i<nElem; i++) {
        if (!fFound && strcmp(sbFlag, rgElem[i]) == 0) {  /* found? */
            fFound = 1;
            if (i+1 < nElem) {  /* is there a following arg? */
                sbRtn = rgElem[i+1];
                i++;  /* skip the next arg */
            }
        } else {
            rgElem[j] = rgElem[i];
            rgfDelete[j] = rgfDelete[i];
            j++;  /* leave in result argv */
        }
    }
    nElem -= (i-j);  /* reduce by how many skipped */
    rgElem[nElem] = NULL;
    
    return sbRtn;
}


// ****************************************************************************
//  Name: toString
//  Date: 04/18/2000 : mdb - Created.
// ****************************************************************************
/**
 * Flattens the argv into a string with spaces between args and quotes
 * around args with spaces in them.
 * @param fAddQuotes - add quotes to any arg that has spaces
 * @param strOut - str to fill
 * @return int - 0=success
 */
/*
int Argv::toString(DCString_t &strOut, bool fAddQuotes, int startArg) const
{
    strOut.Set(NULL);
    for (uint i=(uint) startArg; i<nElem; ++i) {
        if (i>0)
            strOut.Cat(' ');
        if (fAddQuotes && strchr(rgElem[i], ' ') && rgElem[i][0] != '\"') {
            strOut.Cat('\"');
            strOut.Cat(rgElem[i]);
            strOut.Cat('\"');
        } else {
            strOut.Cat(rgElem[i]);
        }
    }
    return 0;
}
*/

// ****************************************************************************
//  Name: operator<<
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/** 
 * Output an argv object.
 */
std::ostream &operator<<(std::ostream &os, Argv &argv)
{
    for (uint i=0; i<argv.nElem; i++) {
        os << "[" << i << "] " << argv.rgElem[i] << std::endl;
    }
    return os;
}


// ****************************************************************************


#ifdef __cplusplus
extern "C" {
#endif
    
static char **DBG_envp = NULL;
static char  *DBG_sbProgramBinary = NULL;


#define DBG_JITD_USE_XTERM
#define DBG_JITD_USE_BINARY_IN_PATH

struct DBG_handler_t {
    bool fStarted;
    Argv a;
    char **envp;
    int pfd1[2];
} DBG_handler;

static void sig_handler_start_debugger(int sig, siginfo_t *, void *)
{
#if defined(LINUX)
#if DCDBG>0
    mtrace();
#endif
#endif    
    if (DBG_handler.fStarted)
        return;
    DBG_handler.fStarted = true;
    
    if (fork() == 0) {  // child debugger process
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
        execvp(DBG_handler.a.Get(0), DBG_handler.a.argv());
#else /* DBG_JITD_USE_BINARY_IN_PATH */
        execve(DBG_handler.a.Get(0), DBG_handler.a.argv(), DBG_handler.envp);
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
        _exit(1);
        // If this fails, we are stuck -- no way to let parent know, which
        // will hang reexecuting the bad instruction.  Need a way to stop
        // it from happening.
    } else { // parent buggy main process
        //close(0);  // so we can interact with gdb instead
        // here we just go back and perform the stmt that caused the problem.
        // Once the debugger is attached we will stop as needed.
        if (sig == SIGABRT) while(1) ;  // return from abort will not stop
    }
}

// SIGILL, SIGBUS, SIGFPE, SIGSEGV, SIGSYS, SIGABRT
int DBG_InstallCoreSigHandlers(char *sbProgramBinary_in, char **envp_in)
{
    int wRtn = 0;
    char tmp[80], tmp2[128];
    
    char *sbProgramBinary = (sbProgramBinary_in ?
                             sbProgramBinary_in : DBG_sbProgramBinary);
    
    //cout << "## Installing Core Signal Handlers for JIT debugging ..." << endl;
    DBG_handler.fStarted = false;
#if defined(DBG_JITD_USE_XTERM)
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
    DBG_handler.a.Append("xterm");
#else /* DBG_JITD_USE_BINARY_IN_PATH */
    DBG_handler.a.Append("/usr/X11R6/bin/xterm");
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
    DBG_handler.a.Append("-T");
    gethostname(tmp, 80);
    snprintf(tmp2, 128, "JIT Debugging for %s on host %s",
             sbProgramBinary, tmp);
    DBG_handler.a.Append(strdup(tmp2), true);
    DBG_handler.a.Append("-e");
#endif /* DBG_JITD_USE_XTERM */
    
    // CLH - added support for dbx (06/07/02)
#if defined(DEBUGGER_USE_GDB)
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
    DBG_handler.a.Append("gdb");
#else
    DBG_handler.a.Append("/usr/bin/gdb");
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
    DBG_handler.a.Append(sbProgramBinary);
    snprintf(tmp, 80, "%d", getpid());
    DBG_handler.a.Append(strdup(tmp), true);
#elif defined(DEBUGGER_USE_DBX)
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
    DBG_handler.a.Append("dbx");
#else
    DBG_handler.a.Append("/usr/bin/dbx");
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
    snprintf(tmp, 80, "-a%d", getpid());
    DBG_handler.a.Append(strdupq(tmp), true);
    DBG_handler.a.Append(sbProgramBinary);
#endif
    DBG_handler.envp = envp_in ? envp_in : DBG_envp;
    
    struct sigaction s;
    s.sa_sigaction = sig_handler_start_debugger;
    sigemptyset(&s.sa_mask);
    s.sa_flags = SA_SIGINFO;
    if (sigaction(SIGILL, &s, NULL) != 0) {
        cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGILL handler: " << endl;
        // CLH - not MT safe - << strerror(errno) << endl;
        wRtn = 1;
    }
    s.sa_sigaction = sig_handler_start_debugger;
    sigemptyset(&s.sa_mask);
    s.sa_flags = SA_SIGINFO;
    if (sigaction(SIGBUS, &s, NULL) != 0) {
        cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGBUS handler: " << endl;
        // CLH - not MT safe - << strerror(errno) << endl;
        wRtn = 1;
    }
    s.sa_sigaction = sig_handler_start_debugger;
    sigemptyset(&s.sa_mask);
    s.sa_flags = SA_SIGINFO;
    if (sigaction(SIGFPE, &s, NULL) != 0) {
        cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGFPE handler: " << endl;
        // CLH - not MT safe - << strerror(errno) << endl;
        wRtn = 1;
    }
    s.sa_sigaction = sig_handler_start_debugger;
    sigemptyset(&s.sa_mask);
    s.sa_flags = SA_SIGINFO;
    if (sigaction(SIGSEGV, &s, NULL) != 0) {
        cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGSEGV handler: " << endl;
        // CLH - not MT safe - << strerror(errno) << endl;
        wRtn = 1;
    }
#if !defined(LINUX)  /* (not defined on redhat 6.1 ll.alliance.unm.edu) */
    s.sa_sigaction = sig_handler_start_debugger;
    sigemptyset(&s.sa_mask);
    s.sa_flags = SA_SIGINFO;
    if (sigaction(SIGSYS, &s, NULL) != 0) {
        cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGSYS handler: " << endl;
        // CLH - not MT safe - << strerror(errno) << endl;
        wRtn = 1;
    }
#endif /* !LINUX */
    s.sa_sigaction = sig_handler_start_debugger;
    sigemptyset(&s.sa_mask);
    s.sa_flags = SA_SIGINFO;
    if (sigaction(SIGABRT, &s, NULL) != 0) {
        cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGABRT handler: " << endl;
        // CLH - not MT safe - << strerror(errno) << endl;
        wRtn = 1;
    }
    return wRtn;
}

#ifdef __cplusplus
}
#endif
