/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------
*/

// -*- c++ -*-
// ****************************************************************************
//  Name: uarray.h
//  Athr: Umit V. Catalyurek (UVC)
//  Date: 09/29/2001 : UVC - created based on Mike Beynon's DArray
//
//  Desc: Implementation of a dynamic array class.
// ****************************************************************************

#if !defined(_UARRAY_H_)
#define _UARRAY_H_

#include <iostream>
#include <stdio.h>
#include "usocket.h"


//
// Notes: T must support assignment and have a constructor with no args
//
template <class T>
class UArray
{
private:
#define DefaultExtentSize  10
protected:
    T*    lst;
    int maxsize;
    int size;
    int extentsize;

    void extendArray(int nExtent=0)
        {
        if (!nExtent)
            nExtent = extentsize;
        
        int nNewSize = maxsize + nExtent;
        T* rgTmp = lst;

//        uprintf("newsize=%u\n", nNewSize);
        
        lst = new T[maxsize = nNewSize];
        if (rgTmp)
            {
                // copy over old elements
            for (int i=0; i<size; i++) 
                lst[i] = rgTmp[i];
            delete [] rgTmp;
            }
        }

public:
    UArray(int maxsize_in = 0, int extentsize_in = DefaultExtentSize)
        {
        maxsize = maxsize_in;
        if (maxsize == 0) 
            lst = NULL;
        else 
            lst = new T[maxsize];
        size = 0;
        extentsize = extentsize_in;
        }
    UArray(const UArray<T>& darray)
        {
        size = 0;
        lst = NULL;
        *this = darray;
        }
    
    ~UArray()
        {
        if (lst)
            delete [] lst;
        }
    
    int getSize(void) const                 { return size; }
    void setExtentSize(int extentsize_in)   { extentsize = extentsize_in; }

    T& operator[](int i)
        {
        if (i >= maxsize)
            extendTo(i);
        if (i >= size)
            size = i+1;
        return lst[i];
        }

    const T& operator[](int i) const          { return lst[i]; }

    void extendTo(int newmax)
        {
        extendArray(newmax-maxsize+extentsize);
        }
    
    int index(const T& elem)
	{
	for (int i=0; i<getSize(); ++i)
            if (lst[i]==elem)
                return i;
        return getSize();
	}
    
    int append(const T& elem)
        {
//        cout << "append(const T& elem)" << elem << endl;
        if (size == maxsize)
            extendArray();
        lst[size++] = elem;
        return size-1;  // index of where added
        }

    int appendByVal(const T elem)
        {
//        cout << "appendByVal(const T elem)" << elem << endl;
        return append(elem);
        }
    
            // remove an element at loc 'i'
    int remove(int i)
        {
        if (i >= size)
            return size;  // don't do anything
            // copy all the elements
        for (int j=i+1; j<size; ++j) 
            lst[j-1] = lst[j];
        return --size;
        }

    void clear()
        {
        if (lst)
            delete[] lst;
        
        lst = NULL;
        maxsize = size = 0;
        }

    void deleteAll()
	{
	if (lst)
	    {
	    for (int i=0; i<size; ++i)
		delete lst[i];
	    clear();
	    }
	}


        // whole array assignment
    UArray<T>& operator=(const UArray<T>& darray)
        {
        if (this != &darray)
            {  // avoid self copies
            if (maxsize > 0)
                delete [] lst;
            maxsize = size = darray.size;
            if (maxsize > 0) 
                lst = new T[maxsize];        
            for (int i=0; i<size; i++) 
                lst[i] = darray.lst[i];
            }
        return *this;
        }

    friend std::ostream& operator<<(std::ostream& os, UArray<T>& darray)
        {
        for (int i=0; i<darray.size; i++)
            os << "[" << i << "] " << darray.lst[i] << std::endl;
        return os;
        }
};


#endif 

