/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  /-------------------------------------------------------------\
  |                     File : heap.c                           |
  +-------------------------------------------------------------+
  |           Written by : Umit V. CATALYUREK                   |
  +-------------------------------------------------------------+
  |           Date       : 03.5.97                              |
  +-------------------------------------------------------------+
  |  Description : Heap                                         |
  +-------------------------------------------------------------+
  |  Modification by :                                          |
  |  Date            :                                          |
  |  Description     :                                          |
  \-------------------------------------------------------------/
*/
#include "uheap.h"


/*
#define _DEBUG1
*/

/*************************************************************************
* --------------------------- Heap ---------------------------------------
**************************************************************************/
int u_HeapAlloc(pheap h, int Imaxid)
{
h->cmaxid = h->maxid = Imaxid;
h->size = 0;
h->data = h->ptr = h->vals = NULL;
h->data = (int *) usmalloc((h->maxid+1)*sizeof(int), "HeapAlloc : data");
h->ptr = (int *) usmalloc((h->maxid+1)*sizeof(int), "HeapAlloc : ptr");
h->vals = (HeapValType *) usmalloc((h->maxid+1)*sizeof(HeapValType), "HeapAlloc : vals");
h->usvals = (UnSignedHeapValType *) h->vals;

if (h->data && h->ptr && h->vals) {
    memset(h->ptr, 0xff, (h->maxid+1)*sizeof(int));
    return 0;
} else {
    u_HeapFree(h);
    return -1;
}
}


void u_HeapInit(pheap h, int Imaxid)
{
#ifdef _DEBUG1
int i;
#endif
if (Imaxid > h->maxid)
    errexit("HeapInit : new size (%d) is larger than previous (%d)", Imaxid, h->maxid);
h->cmaxid = Imaxid;
#ifdef _DEBUG1
for (i=1; i<=h->cmaxid; i++)
    if (h->ptr[i]!=-1)
	errexit("HeapInit: for data %d ptr is %d (size=%d)", i, h->ptr[i], h->size);
#endif
}

void u_HeapDeleteAll(pheap h)
{
int i;

if (h->size<h->cmaxid/3)
    {
    for (i=1; i<=h->size; i++)
	{
#ifdef _DEBUG1
	int d=h->data[i];
	if (d<1 || d>h->maxid)
	    errexit("HeapDeleteAll: invalid data %d at %d (size=%d)", d, i, h->size);
#endif
	h->ptr[h->data[i]] = -1;
	}
    }
else
    memset(h->ptr, 0xff, (h->cmaxid+1)*sizeof(int));

#ifdef _DEBUG1
for (i=1; i<=h->cmaxid; i++)
    if (h->ptr[i]!=-1)
	errexit("HeapDeleteAll: for data %d ptr is %d (size=%d)", i, h->ptr[i], h->size);
#endif
h->size = 0;
}


void u_HeapFree(pheap h)
{
usfree(h->data);
usfree(h->ptr);
usfree(h->vals);
h->data = h->ptr = h->vals = NULL;
}



void HeapPrint(pheap h, char *name)
{
int		i;

uprintf("\n There are %d item(s) in %s heap.", h->size, name);
for (i=1; i<=h->size; i++)
    uprintf("\n %s[%3d] = %3d", name, i, h->data[i]);
uprintf("\n");
}


int u_HeapVerify(pheap h, char *st)
{
int   i, ret=1, cnt=0;

for (i=1; i<=h->cmaxid; i++)
    {
    if (h->ptr[i]>h->size)
	errexit("++++ %s\n HeapVerify: pos %d ptr=%d (size=%d)", st, i, h->ptr[i], h->size);
    if (h->ptr[i]!=-1)
	cnt++;
    }
if (cnt != h->size)
    errexit("**** %s\n HeapVerify: there are %d items in heap but there are %d items which have ptr diff then -1", h->size, cnt);
for (i=1; i<=h->size; i++)
    if (i!=h->ptr[h->data[i]])
	{
	uprintf("\n +++++ %s\n Error in heap i : %d data[i] : %d ptr[data] : %d", st, i, h->data[i], h->ptr[h->data[i]]);
	ret = 0;
	}
return ret;
}


void u_HeapAdd(pheap h, int id, HeapValType v)
{
#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_HeapAdd : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_HeapAdd: invalid id %d", id);
#endif

h->size++;
h->data[h->size] = id;
h->ptr[id] = h->size;
h->vals[id] = v;
}


/*************************************************************************
* --------------------- Generalized Heap Functions -----------------------
 Desr: Use with normal heap structure but you can are required to give
 comparison function (use to store multiple value, e.g. 2 words or string
**************************************************************************/
void u_Gen_HeapAdd(pheap h, int id)
{
#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_Gen_HeapAdd : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_Gen_HeapAdd: invalid id %d", id);
#endif

h->size++;
h->data[h->size] = id;
h->ptr[id] = h->size;
}


void u_Gen_HeapInsert(pheap h, int id, gen_heap_compare func)
{
register  int	i, j, d;

#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_Gen_HeapInsert : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_Gen_HeapInsert: invalid id %d", id);
#endif

++(h->size);
j = h->size;
i = j >> 1;
while (i>0 && (*func)(h->data[i], id)>0)
    {
    d = h->data[i];
    h->data[j] = d;
    h->ptr[d] = j;
    j = i;
    i = i >> 1;
    }
h->data[j] = id;
h->ptr[id] = j;
}


void u_Gen_HeapAdjust(pheap h, int i, gen_heap_compare func)
{
register int 	j, par;
int 	        d=h->data[i];

j = i << 1;
while (j<=h->size)
    {
    if (j<h->size && (*func)(h->data[j], h->data[j+1])>0)
	++j;
    if ((*func)(h->data[j], d)>0)
	break;
    else
	{	
	int dd=h->data[j];
	par = j >> 1;
	h->data[par] = dd;
	h->ptr[dd] = par;
	j = j << 1;
	}
    }
par = j >> 1;
h->data[par] = d;
h->ptr[d] = par;
}


void u_Gen_Heapify(pheap h, gen_heap_compare func)
{
int	i;

for (i=h->size>>1; i>0; i--)
    u_Gen_HeapAdjust(h, i, func);
}


int  u_Gen_HeapExtractFirst(pheap h, gen_heap_compare func)
{
if (h->size>1)
    {
    int	ret=h->data[1], d=h->data[h->size];

    h->data[1] = d;
    h->ptr[d] = 1;
    --(h->size);
#ifdef _DEBUG1
    if (h->size<0)
	errexit("u_MinHeapExtractFirst: heap size is less than zero (%d)", h->size);
#endif
    u_Gen_HeapAdjust(h, 1, func);
    h->ptr[ret] = -1;
    return ret;
    }
else if (h->size==1)
    {
    int	ret=h->data[1];
    h->size = 0;
    h->ptr[ret] = -1;
    return ret;    
    }
else
    return -1;   /* nothing in heap*/
}


void u_Gen_HeapModifiedAt(pheap h, int p, gen_heap_compare func)
{
#ifdef _DEBUG1
if (p<1 || p>h->size)
    errexit("u_MinHeapModifiedAt: modification (%d) is out of range of heap (1,%d)", p, h->size);
#endif
if (p>1 && (*func)(h->data[p >> 1], h->data[p])>0)
    {
    int             d = h->data[p];
    int             i, j;

    j = p;
    i = j >> 1;
    while (i>=1)
	{
	if ((*func)(h->data[i], d)>0)
	    {
	    int dd=h->data[i];
	    h->data[j] = dd;
	    h->ptr[dd] = j;
	    j = i;
	    i = j >> 1;
	    }
	else
	    break;
	}
    h->data[j] = d;
    h->ptr[d] = j;
    }
else
    u_Gen_HeapAdjust(h, p, func);
}

void u_Gen_HeapDelete(pheap h, int id, gen_heap_compare func)
{
int	p=h->ptr[id], d=h->data[h->size];

#ifdef _DEBUG1
if (h->ptr[id] == -1)
    errexit("u_MinHeapDelete : id : %d not in heap", id);
#endif

h->data[p] = d;
h->ptr[d] = p;
(h->size)--;
#ifdef _DEBUG1
if (h->size<0)
    errexit("u_MinHeapDelete: heap size is less than zero (%d)", h->size);
#endif

h->ptr[id] = -1;
if (p<=h->size)
    u_Gen_HeapModifiedAt(h, p, func);
}

void u_Gen_HeapUpdateVal(pheap h, int id, gen_heap_compare func)
{
#ifdef _DEBUG
if (id<1 || id>h->cmaxid)
    errexit("u_MinHeapUpdateVal: invalid id %d", id);
if (h->ptr[id]==-1)
    errexit("u_MinHeapUpdateVal: id %d not in heap", id);
#endif
u_Gen_HeapModifiedAt(h, h->ptr[id], func);
}


/*************************************************************************
* --------------------- signed MinHeap Functions -------------------------
**************************************************************************/
void u_MinHeapInsert(pheap h, int id, HeapValType v)
{
register  int	i, j, d;

#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_MinHeapInsert : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_MinHeapInsert: invalid id %d", id);
#endif

++(h->size);
j = h->size;
i = j >> 1;
while (i>0 && h->vals[h->data[i]] >= v)
    {
    d=h->data[i];
    h->data[j] = d;
    h->ptr[d] = j;
    j = i;
    i = i >> 1;
    }
h->vals[id] = v;
h->data[j] = id;
h->ptr[id] = j;
}


void u_MinHeapAdjust(pheap h, int i)
{
register int 	j, par;
int 	        d=h->data[i];
HeapValType 	val=h->vals[h->data[i]];

j = i << 1;
while (j<=h->size)
    {
    if (j<h->size && h->vals[h->data[j]] > h->vals[h->data[j+1]])
	++j;
    if (h->vals[h->data[j]] >= val)
	break;
    else
	{	
	int dd=h->data[j];
	par = j >> 1;
	h->data[par] = dd;
	h->ptr[dd] = par;
	j = j << 1;
	}
    }
par = j >> 1;
h->data[par] = d;
h->ptr[d] = par;
}


void u_MinHeapify(pheap h)
{
int	i;

for (i=h->size>>1; i>0; i--)
    u_MinHeapAdjust(h, i);
}


int u_MinHeapExtractFirst(pheap h)
{
if (h->size>1)
    {
    int	ret=h->data[1], d=h->data[h->size];

    h->data[1] = d;
    h->ptr[d] = 1;
    --(h->size);
#ifdef _DEBUG1
    if (h->size<0)
	errexit("u_MinHeapExtractFirst: heap size is less than zero (%d)", h->size);
#endif
    u_MinHeapAdjust(h, 1);
    h->ptr[ret] = -1;
    return ret;
    }
else if (h->size==1)
    {
    int	ret=h->data[1];
    h->size = 0;
    h->ptr[ret] = -1;
    return ret;    
    }
else
    return -1;   /* nothing in heap*/
}


void u_MinHeapModifiedAt(pheap h, int p) /* p is position */
{
#ifdef _DEBUG1
if (p<1 || p>h->size)
    errexit("u_MinHeapModifiedAt: modification (%d) is out of range of heap (1,%d)", p, h->size);
#endif
if (p>1 && h->vals[h->data[p >> 1]] > h->vals[h->data[p]])
    {
    int             d = h->data[p];
    HeapValType     val = h->vals[h->data[p]];
    int             i, j;

    j = p;
    i = j >> 1;
    while (i>=1)
	{
	if (h->vals[h->data[i]] >= val)
	    {
	    int dd=h->data[i];
	    h->data[j] = dd;
	    h->ptr[dd] = j;
	    j = i;
	    i = j >> 1;
	    }
	else
	    break;
	}
    h->data[j] = d;
    h->ptr[d] = j;
    }
else
    u_MinHeapAdjust(h, p);
}


void u_MinHeapDelete(pheap h, int id)
{
int	p=h->ptr[id], d=h->data[h->size];

#ifdef _DEBUG1
if (h->ptr[id] == -1)
    errexit("u_MinHeapDelete : id : %d not in heap", id);
#endif

h->data[p] = d;
h->ptr[d] = p;
(h->size)--;
#ifdef _DEBUG1
if (h->size<0)
    errexit("u_MinHeapDelete: heap size is less than zero (%d)", h->size);
#endif

h->ptr[id] = -1;
if (p<=h->size)
    u_MinHeapModifiedAt(h, p);
}


void u_MinHeapUpdateVal(pheap h, int id, HeapValType v)
{
#ifdef _DEBUG
if (id<1 || id>h->cmaxid)
    errexit("u_MinHeapUpdateVal: invalid id %d", id);
if (h->ptr[id]==-1)
    errexit("u_MinHeapUpdateVal: id %d not in heap", id);
#endif
h->vals[id] = v; 
u_MinHeapModifiedAt(h, h->ptr[id]);
}


/*************************************************************************
* --------------------- signed MaxHeap Functions -------------------------
**************************************************************************/
void u_MaxHeapInsert(pheap h, int id, HeapValType v)
{
register int	i, j, d;

#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_MaxHeapInsert : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_MaxHeapInsert: invalid id %d", id);
#endif

++(h->size);
j = h->size;
i = j >> 1;
while (i>0 && h->vals[h->data[i]] <= v)
    {
    d = h->data[i];
    h->data[j] = d;
    h->ptr[d] = j;
    j = i;
    i = i >> 1;
    }
h->vals[id] = v;
h->data[j] = id;
h->ptr[id] = j;
#ifdef _DEBUG4
u_HeapVerify(h, "u_MaxHeapInsert");
#endif
}


void u_MaxHeapAdjust(pheap h, int i)
{
register int 	j, par, dd;
int 	        d=h->data[i];
HeapValType 	val=h->vals[h->data[i]];

j = i << 1;
while (j<=h->size)
    {
    if (j<h->size && h->vals[h->data[j]] < h->vals[h->data[j+1]])
	++j;
    if (h->vals[h->data[j]] <= val)
	break;
    else
	{
	par = j >> 1;
	dd = h->data[j];
	h->data[par] = dd;
	h->ptr[dd] = par;
	j = j << 1;
	}
    }
par = j >> 1;
h->data[par] = d;
h->ptr[d] = par;
}


void u_MaxHeapify(pheap h)
{
int	i;

for (i=h->size>>1; i>0; i--)
    u_MaxHeapAdjust(h, i);
#ifdef _DEBUG4
u_HeapVerify(h, "u_MaxHeapify");
#endif
}


int u_MaxHeapExtractFirst(pheap h)
{
if (h->size>1)
    {
    int	ret=h->data[1], d=h->data[h->size];

    h->data[1] = d;
    h->ptr[d] = 1;
    --(h->size);
#ifdef _DEBUG1
if (h->size<0)
    errexit("u_MaxHeapExtractFirst: heap size is less than zero (%d)", h->size);
#endif
    u_MaxHeapAdjust(h, 1);
    h->ptr[ret] = -1;
#ifdef _DEBUG4
    u_HeapVerify(h, "u_MaxHeapExtractFirst");
#endif
    return ret;
    }
else if (h->size==1)
    {
    int	ret=h->data[1];
    h->size = 0;
    h->ptr[ret] = -1;
    return ret;    
    }
else
    return -1;   /* nothing in heap*/
}


void u_MaxHeapModifiedAt(pheap h, int p) /* p is position */
{
#ifdef _DEBUG1
if (p<1 || p>h->size)
    errexit("u_MaxHeapModifiedAt: modification (%d) is out of range of heap (1,%d)", p, h->size);
#endif
if (p>1 && h->vals[h->data[p>>1]] < h->vals[h->data[p]])
    {
    int             d = h->data[p];
    HeapValType     val = h->vals[h->data[p]];
    int             i, j;

    j = p;
    i = j >> 1;
    while (i>=1)
	{
	if (h->vals[h->data[i]] <= val)
	    {
	    int dd=h->data[i];
	    h->data[j] = dd;
	    h->ptr[dd] = j;
	    j = i;
	    i = j >> 1;
	    }
	else
	    break;
	}
    h->data[j] = d;
    h->ptr[d] = j;
    }
else
    u_MaxHeapAdjust(h, p);
}

void u_MaxHeapDelete(pheap h, int id)
{
int	p=h->ptr[id], d=h->data[h->size];

#ifdef _DEBUG1
if (h->ptr[id] == -1)
    errexit("u_MaxHeapDelete : id : %d not in heap", id);
#endif

h->data[p] = d;
h->ptr[d] = p;
(h->size)--;
#ifdef _DEBUG1
if (h->size<0)
    errexit("u_MaxHeapDelete: heap size is less than zero (%d)", h->size);
#endif
h->ptr[id] = -1;
if (p<=h->size)
    u_MaxHeapModifiedAt(h, p);
#ifdef _DEBUG4
u_HeapVerify(h, "u_MaxHeapDelete");
#endif
}


void u_MaxHeapUpdateVal(pheap h, int id, HeapValType v)
{
#ifdef _DEBUG
if (id<1 || id>h->cmaxid)
    errexit("u_MaxHeapUpdateVal: invalid id %d", id);
if (h->ptr[id]==-1)
    errexit("u_MaxHeapUpdateVal: id %d not in heap", id);
#endif
h->vals[id] = v; 
u_MaxHeapModifiedAt(h, h->ptr[id]);
}


 
/*************************************************************************
* --------------------- unsigned MaxHeap Functions -----------------------
**************************************************************************/
void u_UnSigned_HeapAdd(pheap h, int id, UnSignedHeapValType v)
{
#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_UnSigned_HeapAdd : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_UnSigned_HeapAdd: invalid id %d", id);
#endif

h->size++;
h->data[h->size] = id;
h->ptr[id] = h->size;
h->usvals[id] = v;
}

void u_UnSigned_MaxHeapInsert(pheap h, int id, UnSignedHeapValType v)
{
register int	i, j, d;

#ifdef _DEBUG1
if (h->size>=h->maxid || h->ptr[id] != -1)
    errexit("u_UnSigned_MaxHeapInsert : Insert fail for id : %d ptr : %d size is %d", id, h->ptr[id], h->size);
if (id<1 || id>h->cmaxid)
    errexit("u_UnSigned_MaxHeapInsert: invalid id %d", id);
#endif

++(h->size);
j = h->size;
i = j >> 1;
while (i>0 && h->usvals[h->data[i]] <= v)
    {
    d = h->data[i];
    h->data[j] = d;
    h->ptr[d] = j;
    j = i;
    i = i >> 1;
    }
h->usvals[id] = v;
h->data[j] = id;
h->ptr[id] = j;
#ifdef _DEBUG4
u_HeapVerify(h, "u_UnSigned_MaxHeapInsert");
#endif
}


void u_UnSigned_MaxHeapAdjust(pheap h, int i)
{
register int 		j, par, dd;
int 		        d=h->data[i];
UnSignedHeapValType 	val=h->usvals[h->data[i]];

j = i << 1;
while (j<=h->size)
    {
    if (j<h->size && h->usvals[h->data[j]] < h->usvals[h->data[j+1]])
	++j;
    if (h->usvals[h->data[j]] <= val)
	break;
    else
	{
	par = j >> 1;
	dd = h->data[j];
	h->data[par] = dd;
	h->ptr[dd] = par;
	j = j << 1;
	}
    }
par = j >> 1;
h->data[par] = d;
h->ptr[d] = par;
}


void u_UnSigned_MaxHeapify(pheap h)
{
int	i;

for (i=h->size>>1; i>0; i--)
    u_UnSigned_MaxHeapAdjust(h, i);
#ifdef _DEBUG4
    u_HeapVerify(h, "u_UnSigned_MaxHeapify");
#endif
}


int u_UnSigned_MaxHeapExtractFirst(pheap h)
{
if (h->size>1)
    {
    int	ret=h->data[1], d=h->data[h->size];

    h->data[1] = d;
    h->ptr[d] = 1;
    --(h->size);
#ifdef _DEBUG1
if (h->size<0)
    errexit("u_UnSigned_MaxHeapExtractFirst: heap size is less than zero (%d)", h->size);
#endif
    u_UnSigned_MaxHeapAdjust(h, 1);
    h->ptr[ret] = -1;
#ifdef _DEBUG4
    u_HeapVerify(h, "u_UnSigned_MaxHeapExtractFirst");
#endif
    return ret;
    }
else if (h->size==1)
    {
    int	ret=h->data[1];
    h->size = 0;
    h->ptr[ret] = -1;
    return ret;    
    }
else
    return -1;   /* nothing in heap*/
}


void u_UnSigned_MaxHeapModifiedAt(pheap h, int p) /* p is position */
{
#ifdef _DEBUG1
if (p<1 || p>h->size)
    errexit("u_UnSigned_MaxHeapModifiedAt: modification (%d) is out of range of heap (1,%d)", p, h->size);
#endif
if (p>1 && h->usvals[h->data[p>>1]] < h->usvals[h->data[p]])
    {
    int			    d = h->data[p];
    UnSignedHeapValType     val = h->usvals[h->data[p]];
    int			    i, j;

    j = p;
    i = j >> 1;
    while (i>=1)
	{
	if (h->usvals[h->data[i]] <= val)
	    {
	    int dd=h->data[i];
	    h->data[j] = dd;
	    h->ptr[dd] = j;
	    j = i;
	    i = j >> 1;
	    }
	else
	    break;
	}
    h->data[j] = d;
    h->ptr[d] = j;
    }
else
    u_UnSigned_MaxHeapAdjust(h, p);
}

void u_UnSigned_MaxHeapDelete(pheap h, int id)
{
int	p=h->ptr[id], d=h->data[h->size];

#ifdef _DEBUG1
if (h->ptr[id] == -1)
    errexit("u_UnSigned_MaxHeapDelete : id : %d not in heap", id);
#endif

h->data[p] = d;
h->ptr[d] = p;
(h->size)--;
#ifdef _DEBUG1
if (h->size<0)
    errexit("u_UnSigned_MaxHeapDelete: heap size is less than zero (%d)", h->size);
#endif
h->ptr[id] = -1;
if (p<=h->size)
    u_UnSigned_MaxHeapModifiedAt(h, p);
#ifdef _DEBUG4
u_HeapVerify(h, "u_UnSigned_MaxHeapDelete");
#endif
}

void u_UnSigned_MaxHeapUpdateVal(pheap h, int id, UnSignedHeapValType v)
{
#ifdef _DEBUG
if (id<1 || id>h->cmaxid)
    errexit("u_UnSigned_MaxHeapUpdateVal: invalid id %d", id);
if (h->ptr[id]==-1)
    errexit("u_UnSigned_MaxHeapUpdateVal: id %d not in heap", id);
#endif
h->usvals[id] = v; 
u_UnSigned_MaxHeapModifiedAt(h, h->ptr[id]);
}



#ifdef _MSC_VER 
#ifdef _DEBUG1
#pragma message( "Compiling " __FILE__ " with DEBUG1") 
#endif
#ifdef _DEBUG2
#pragma message( "Compiling " __FILE__ " with DEBUG2") 
#endif
#ifdef _DEBUG3
#pragma message( "Compiling " __FILE__ " with DEBUG3") 
#endif
#ifdef _DEBUG4
#pragma message( "Compiling " __FILE__ " with DEBUG4") 
#endif
#endif
