/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  /-------------------------------------------------------------\
  |                     FILE : ulib.c                           |
  +-------------------------------------------------------------+
  |           Written by : Umit V. CATALYUREK                   |
  +-------------------------------------------------------------+
  |           Date       : 27.9.95                              |
  +-------------------------------------------------------------+
  |  Description : Some useful library functions, such as;      |
  |                                                             |
  |     1. File open/close                                      |
  |     2. Timer functions                                      |
  |     3. Safe memory allocation functions                     |
  |     4. Sorting functions                                    |
  |     5. Safe string functions                                |
  |     6. Nice error exit function                             |
  +-------------------------------------------------------------+
  |  Modification by :                                          |
  |  Date            :                                          |
  |  Description     :                                          |
  \-------------------------------------------------------------/
  */

#include "ulib.h"



/*
#define _DEBUG_POOL_MEM
#define _DEBUG_MALLOC_LOW
*/


int _u_seed=847449649;

#define U_MB     (1024*1024)
#define MAXTABSPACELINE  300
#define TABSIZE          3


#ifdef _DEBUG_MALLOC_LOW

#define MAX_ALLOC_CNT	10000

typedef struct UmallocTTag
{
  pvoid  ptr;
  int    size;
  char   *msg;
} UmallocT;

typedef UmallocT	UmallocList[MAX_ALLOC_CNT];

static UmallocList __umallst;

static int __umallstcnt=0;
static long __u_maxtot_size=0;
#endif

static long __u_tot_size=0;


void ReportAllocMem(void)
{
#ifdef _DEBUG_MALLOC_LOW
int i, cnt=0;

for (i=0; i<__umallstcnt; i++)
    if (__umallst[i].ptr)
	cnt++;
uprintf("\n total %d (%d emptied out of %d) ptr allocated with size %dK (maxsize=%dK)", cnt, __umallstcnt-cnt, __umallstcnt,__u_tot_size/1024, __u_maxtot_size/1024);
#else
uprintf("\n Total allocated size = %ldK", __u_tot_size/1024);
#endif
}

void DetailedReportAllocMem()
{
#ifdef _DEBUG_MALLOC_LOW
int i, cnt=0;

uprintf("\n Allocations:");
for (i=0; i<__umallstcnt; i++)
    if (__umallst[i].ptr)
	{
	cnt += __umallst[i].size;
	uprintf("\n%3d. size=%6dK (CurTot:%6dK) %s", i, __umallst[i].size/1024, cnt/1024, __umallst[i].msg);
	}
uprintf("\n---------------------------------------------");
if (cnt != __u_tot_size)
    errexit("DetailedReportAllocMem: cnt=%d tot_size=%d", cnt, __u_tot_size);
ReportAllocMem();
uprintf("\n---------------------------------------------");
#endif
}


/*************************************************************************
 * --------------------------- Useful -------------------------------------
 **************************************************************************/
void uWaitKbdNewLine(void)
{
char st[500];

 scanf("%s\n", st);
 
}

static char _dbg_file_name[512];
#define DEFAULT_DEBUG_FILE "PaToH.log"

void udbgprintf_SetDbgFile(char *name)
{
 if (name) 
     strcpy(_dbg_file_name, name);
 else
     *_dbg_file_name = 0;
}

static void internaldbgprintf(pchar msg)
{
 char *name;
 static int cnt=0;
 FILE *ofp;

 if (!*_dbg_file_name) {
     time_t t;
 
     time(&t);
     sprintf(_dbg_file_name, "%s.", DEFAULT_DEBUG_FILE);
     strcpy(_dbg_file_name+strlen(_dbg_file_name), ctime(&t));
     for (name = _dbg_file_name; *name; ++name)
         if (*name==' ')
             *name='_';
         else if (*name==':')
             *name='-';
         else if (*name=='\n')
             *name = 0;
 }
 name = _dbg_file_name;
     
 if (cnt++==0)
    {
    time_t curtime;
    char    tst[26];

    ofp = ufopen(name, "w", "dbgprint");
    time(&curtime);
    strcpy(tst, ctime(&curtime));
    tst[24] = 0;
    fprintf(ofp, "\n\n\n----------------- %s -----------------\n\n", tst);
    }
 else
     ofp = ufopen(name, "a", "dbgprint");

fprintf(ofp, msg);
ufclose(ofp);
}

void uprintf(pchar f_str,...)
{
va_list argp;
#ifdef _DEBUG
char    *msg=NULL;
#endif

fflush(stdout);
fflush(stderr);
va_start(argp, f_str);
#ifdef _DEBUG
vasprintf(&msg, f_str, argp);
#else
vfprintf(stdout, f_str, argp);
#endif
va_end(argp);

#ifdef _DEBUG    
if (!msg) {
    va_start(argp, f_str);
    vfprintf(stdout, f_str, argp);
    va_end(argp);
} else {
    fprintf(stdout, "%s", msg);
    internaldbgprintf(msg);
    free(msg);
}
#endif
fflush(stdout);
}

void udbgprintf(pchar f_str,...)
{
    char    *msg;
    va_list argp;
    
    va_start(argp, f_str);
    vasprintf(&msg, f_str, argp);
    va_end(argp);
    internaldbgprintf(msg);
    free(msg);
}


/*************************************************************************
 * --------------------------- Time ---------------------------------------
 **************************************************************************/

void u_cleartimer(ptimer tmr)
{
tmr->time = 0.0;
tmr->count = 0L;
tmr->status = TMR_CLEAR;
}

void u_addtimer(ptimer t, ptimer a)
{
t->time += a->time;
t->count++;
}

void u_starttimer(ptimer tmr)
{
if (tmr->status == TMR_START) 
    errexit("Timer already started!\n");

tmr->time -= u_seconds();
tmr->count++;
tmr->status = TMR_START;
}


void u_stoptimer(ptimer tmr)
{
if (tmr->status != TMR_START) 
    errexit("Timer has not been started!\n");

tmr->time += u_seconds();
tmr->status = TMR_STOP;
}


void u_printtimer(ptimer tmr, pchar msg)
{
if (tmr->status == TMR_START) 
    errexit("Timing is in progress!\n");

uprintf("\n%s: %11.3lf sec", msg, tmr->time);
fflush(stdout);
}

double u_gettimer(ptimer tmr)
{
if (tmr->status == TMR_START) 
    errexit("Timing is in progress!\n");

return tmr->time;
}


double u_safegettimer(ptimer tmr)
{
    double r;
if (tmr->status == TMR_START)  {
    u_stoptimer(tmr);
    --tmr->count;
    r = u_gettimer(tmr);
    u_starttimer(tmr);
} else
    r = u_gettimer(tmr);

return r;
}


char *u_gettimestr(ptimer tmr)
{
static char st[20];
double h, m, s;


if (tmr->status == TMR_START) 
    errexit("Timing is in progress!\n");

m = tmr->time / 60.0;
s = tmr->time - (int)m*60;
h = m / 60.0;
m = m - (int)h*60;

sprintf(st, "%d:%02d:%02d", (int) h, (int) m, (int) s);
return st;
}




void u_clearwtimer(pwtimer tmr)
{
tmr->time = 0.0;
tmr->count = 0L;
tmr->status = TMR_CLEAR;
}

void u_addwtimer(pwtimer t, pwtimer a)
{
t->time += a->time;
t->count++;
}

void u_startwtimer(pwtimer tmr)
{
if (tmr->status == TMR_START) 
    errexit("Timer already started!\n");

tmr->time -= u_wseconds();
tmr->count++;
tmr->status = TMR_START;
}


void u_stopwtimer(pwtimer tmr)
{
if (tmr->status != TMR_START) 
    errexit("Timer has not been started!\n");

tmr->time += u_wseconds();
tmr->status = TMR_STOP;
}

void u_printwtimer(pwtimer tmr, pchar msg)
{
if (tmr->status == TMR_START) 
    errexit("Timing is in progress!\n");

uprintf("\n%s: %11.3lf", msg, tmr->time);
fflush(stdout);
}

double u_getwtimer(pwtimer tmr)
{
if (tmr->status == TMR_START) 
    errexit("Timing is in progress!\n");

return tmr->time;
}

char *u_getwtimestr(pwtimer tmr)
{
static char st[20];
double h, m, s;


if (tmr->status == TMR_START) 
    errexit("Timing is in progress!\n");

m = tmr->time / 60.0;
s = tmr->time - (int)m*60;
h = m / 60.0;
m = m - (int)h*60;

sprintf(st, "%d:%02d:%02d", (int) h, (int) m, (int) s);
return st;
}




/*************************************************************************
 * --------------------------- File ---------------------------------------
 **************************************************************************/

FILE *ufopen(pchar fname, pchar mode, pchar msg)
{
FILE *fp;

fp = fopen(fname, mode);
if (fp != NULL)
    return fp;

errexit("file: %s, mode: %s, [%s]", fname, mode, msg);
return NULL;
}


void ufclose(FILE *fp)
{
if (fp)
    fclose(fp);
}

int uopen(pchar fname, int flags, pchar msg)
{
int fid;

fid = open(fname, flags);
if (fid != -1)
    return fid;

errexit("file: %s, flags: %x, [%s]", fname, flags, msg);
return -1;
}


void uclose(int fid)
{
close(fid);
}


/*************************************************************************
 * -------------------------- Memory --------------------------------------
 **************************************************************************/
pint imalloc(long size, pchar msg)
{
pint ptr;

ptr = (pint )umalloc(sizeof(int)*size, msg);
return ptr;
}

pfloat fmalloc(long size, pchar msg)
{
pfloat ptr;

ptr = (pfloat) umalloc(sizeof(float)*size, msg);
return ptr;
}

pdouble dmalloc(long size, pchar msg)
{
pdouble ptr;

ptr = (pdouble) umalloc(sizeof(double)*size, msg);
return ptr;
}


pvoid umalloc(long size, pchar msg)
{
pvoid ptr;
#ifdef _DEBUG_MALLOC_LOW
int i;

if (size<=0)
    errexit("umalloc: invalid alloc request in '%s' with size: %ld", msg, size);
#endif

if (size == 0)
    return NULL;

ptr = (pvoid) malloc(size);
if (!ptr)
    {
#ifdef _DEBUG_MALLOC_LOW
    DetailedReportAllocMem();
#endif
    errexit("Memory allocation failed for '%s'.\nRequested size: %ld byte(s) [%.2lf MB]. Total allocated : %ld byte(s) [%.2lf MB]\n", msg, size, (double)size/(double)U_MB, __u_tot_size, (double)__u_tot_size/(double)U_MB);
    }
__u_tot_size += size;

#ifdef _DEBUG_MALLOC_LOW
if (ptr<(pvoid)1000)
    errexit("Possible Memory allocation failed?? (ptr=%ld) for '%s'.\nRequested size: %ld byte(s). Total allocated : %ld \n", ptr, msg, size, __u_tot_size);

if (__u_tot_size>__u_maxtot_size)
    __u_maxtot_size = __u_tot_size;
for (i=0; i<__umallstcnt && __umallst[i].ptr ; i++);
if (i>=__umallstcnt) {
    if (i>MAX_ALLOC_CNT)
	errexit("MAX_ALLOC_CNT not big enough");
    else
	__umallstcnt++;
}
__umallst[i].ptr = ptr;
__umallst[i].size = size;
__umallst[i].msg = strdup(msg);
#endif

return ptr;
}

pvoid urealloc(pvoid ptr, long size, pchar msg)
{
pvoid nptr;
#ifdef _DEBUG_MALLOC_LOW
int i;

for (i=0; i<__umallstcnt && __umallst[i].ptr!=ptr ; i++);
if (i>__umallstcnt)
    errexit("urealloc: ptr (%ld) not allocated before", ptr);
#endif

nptr = (pvoid) realloc(ptr, size);
if (!nptr)
    errexit("\nRe-alloc failure for '%s'.\nRequested size: %d byte(s).\nTotal allocated: %ld", msg, size, __u_tot_size);

#ifdef _DEBUG_MALLOC_LOW
__u_tot_size -= __umallst[i].size;
__u_tot_size += size;
__umallst[i].ptr = nptr;
__umallst[i].size = size;
free(__umallst[i].msg);
__umallst[i].msg = strdup(msg);
#endif
return nptr;
}


void unfree(int n, ...)
{
  int i;
  va_list va;
  
  va_start(va, n);
  for (i=0; i<n; i++) {
      void **ptr=va_arg(va, void **);
      usfree(*ptr);
      *ptr = NULL;
  }
  va_end(va);
}


void ufree(pvoid ptr1,...)
{
va_list plist;
pvoid ptr;

va_start(plist, ptr1);

while ((ptr = va_arg(plist, pvoid )) != NULL) 
    usfree(ptr);

va_end(plist);
}


pint ismalloc(long size, pchar msg)
{
return (pint )usmalloc(sizeof(int)*size, msg);
}

pfloat fsmalloc(long size, pchar msg)
{
return (pfloat) usmalloc(sizeof(float)*size, msg);
}

pdouble dsmalloc(long size, pchar msg)
{
return (pdouble) usmalloc(sizeof(double)*size, msg);
}


pvoid usmalloc(long size, pchar msg)
{
pvoid ptr;
#ifdef _DEBUG_MALLOC_LOW
int i;

if (size<=0)
    uwarning("umalloc: invalid alloc request in '%s' with size: %ld", msg, size);
#else
(void) msg; /* to avoid unused parameter warning */
#endif

if (size == 0)
    return NULL;

ptr = (pvoid) malloc(size);
if (!ptr)
    {
#ifdef _DEBUG_MALLOC_LOW
    DetailedReportAllocMem();
#endif
#ifdef _DEBUG
    uwarning("Memory allocation failed for '%s'.\nRequested size: %ld byte(s) [%.2lf MB]. Total allocated : %ld byte(s) [%.2lf MB]\n", msg, size, (double)size/(double)U_MB, __u_tot_size, (double)__u_tot_size/(double)U_MB);
#endif
    return ptr;
    }
__u_tot_size += size;

#ifdef _DEBUG_MALLOC_LOW
if (ptr<(pvoid)1000)
    uwarning("Possible Memory allocation failed?? (ptr=%ld) for '%s'.\nRequested size: %ld byte(s). Total allocated : %ld \n", ptr, msg, size, __u_tot_size);

if (__u_tot_size>__u_maxtot_size)
    __u_maxtot_size = __u_tot_size;
for (i=0; i<__umallstcnt && __umallst[i].ptr ; i++);
if (i>=__umallstcnt) {
    if (i>MAX_ALLOC_CNT)
	errexit("MAX_ALLOC_CNT not big enough");
    else
	__umallstcnt++;
}
__umallst[i].ptr = ptr;
__umallst[i].size = size;
__umallst[i].msg = strdup(msg);
#endif

return ptr;
}

pvoid usrealloc(pvoid ptr, long size, pchar msg)
{
pvoid nptr;
#ifdef _DEBUG_MALLOC_LOW
int i;

for (i=0; i<__umallstcnt && __umallst[i].ptr!=ptr ; i++);
if (i>__umallstcnt)
    errexit("urealloc: for '%s' ptr (%ld) not allocated before", msg, ptr);
#else
(void) msg; /* to avoid unused parameter warning */
#endif

nptr = (pvoid) realloc(ptr, size);
if (!nptr) {
#ifdef _DEBUG
    uwarning("\nRe-alloc failure for '%s'.\nRequested size: %d byte(s).\nTotal allocated: %ld", msg, size, __u_tot_size);
#endif
    return nptr;
}

#ifdef _DEBUG_MALLOC_LOW
__u_tot_size -= __umallst[i].size;
__u_tot_size += size;
__umallst[i].ptr = nptr;
__umallst[i].size = size;
free(__umallst[i].msg);
__umallst[i].msg = strdup(msg);
#endif
return nptr;
}


void _usfree(pvoid ptr)    /* Safe free checks if parameter is not null*/
{
if (ptr)
    {
#ifdef _DEBUG_MALLOC_LOW
    int i;
    
    for (i=0; i<__umallstcnt && __umallst[i].ptr!=ptr ; i++);
    if (i>__umallstcnt)
	errexit("usfree: ptr (%ld) not allocated before", ptr);
    __u_tot_size -= __umallst[i].size;
    __umallst[i].ptr = 0;
    free(__umallst[i].msg);
#endif
    free(ptr);
    }
}



void sicopy(pint dest, pint src, int size)
{
if (!dest || !src)
    errexit("icopy : either src or dest is null");
else if (size<=0)
    return;
memcpy((pvoid )dest, (pvoid )src, sizeof(int)*size);
}


static AllocationPools _u_pools;

static AllocationPools _u_pools_check[100];
static int poolcheckcnt=0;

void SavePools2Check(void)
{
int i;
if (poolcheckcnt==100)
    errexit("poolcheck array is too small to save checks");
for (i=0; i<MAXALLOCPOOL; ++i)
    _u_pools_check[poolcheckcnt][i] = _u_pools[i];
++poolcheckcnt;
#ifdef _DEBUG_MALLOC_LOW
uprintf("[PoolSave-%d]", poolcheckcnt);
#endif
}

void CheckPools(void)
{
int i, cnt=0;
--poolcheckcnt;
if (poolcheckcnt<0)
    errexit("poolcheckcnt=%d", poolcheckcnt);
for (i=0; i<MAXALLOCPOOL; ++i)
    {
    if (_u_pools_check[poolcheckcnt][i].base != _u_pools[i].base)
	uprintf("\n ErrCnt:%4d in pool %d check-base=%x org-base=%x diff=%d", ++cnt, i, _u_pools_check[poolcheckcnt][i].base, _u_pools[i].base, _u_pools_check[poolcheckcnt][i].base-_u_pools[i].base);
    if (_u_pools_check[poolcheckcnt][i].ptr != _u_pools[i].ptr)
	uprintf("\n ErrCnt:%4d in pool %d check-ptr=%x org-ptr=%x diff=%d", ++cnt, i, _u_pools_check[poolcheckcnt][i].ptr, _u_pools[i].ptr, _u_pools_check[poolcheckcnt][i].ptr-_u_pools[i].ptr);
    if (_u_pools_check[poolcheckcnt][i].last != _u_pools[i].last)
	uprintf("\n ErrCnt:%4d in pool %d check-last=%x org-last=%x diff=%d", ++cnt, i, _u_pools_check[poolcheckcnt][i].last, _u_pools[i].last, _u_pools_check[poolcheckcnt][i].last-_u_pools[i].last);
    }
if (cnt)
    errexit("pool check error %d", poolcheckcnt);
}


void InitAllocationPools()
{
int i;
for (i=0; i<MAXALLOCPOOL; i++)
    _u_pools[i].isallocated = 0;
}

#define IsPoolAllocated(pool)   (_u_pools[pool].isallocated)


int sAllocatePool(int pool, long size, char *msg)
{
if (!IsPoolAllocated(pool))
    {
    char st[500];
    
    sprintf(st, "Pool-%d.\n\tRequested in %s", pool, msg);
    _u_pools[pool].base = _u_pools[pool].ptr = (pchar) usmalloc(size, st);
    if (!_u_pools[pool].base)
        return -1;
    _u_pools[pool].last = _u_pools[pool].ptr + size;
    _u_pools[pool].isallocated = 1;
#ifdef _DEBUG_POOL_MEM
    _u_pools[pool].maxused = 0;
#endif
    }
return 0;
}


int AllocatePool(int pool, long size, char *msg)
{
if (!IsPoolAllocated(pool))
    {
    char st[500];
    
    sprintf(st, "Pool-%d.\n\tRequested in %s", pool, msg);
    _u_pools[pool].base = _u_pools[pool].ptr = (pchar) umalloc(size, st);
    _u_pools[pool].last = _u_pools[pool].ptr + size;
    _u_pools[pool].isallocated = 1;
#ifdef _DEBUG_POOL_MEM
    _u_pools[pool].maxused = 0;
#endif
    }
return 0;
}

void ReInitPools()
{
int pool;

for (pool=0; pool<MAXALLOCPOOL; pool++)
    if (IsPoolAllocated(pool))
	{
	_u_pools[pool].ptr = _u_pools[pool].base;
#ifdef _DEBUG_POOL_MEM
	_u_pools[pool].maxused = 0;
#endif
	}
}

void ReportMaxPoolUsages(void)
{
#ifdef _DEBUG_POOL_MEM
int i; 
uprintf("\n Max pool usages in Kbyte(s)");
for (i=0; i<MAXALLOCPOOL; i++)
     uprintf("\nPool[%d] = %6ld  of  %6ld  (%3ld%%)", i, _u_pools[i].maxused/1024,
             (_u_pools[i].last-_u_pools[i].base)/1024,
             100*_u_pools[i].maxused/(_u_pools[i].last-_u_pools[i].base));
uprintf("\n-------------------------\n");
#else
uprintf("\n----- Max usage not available--------\n");
#endif
}


void ShortReportMaxPoolUsages(void)
{
#ifdef _DEBUG_POOL_MEM
int i;
uprintf("\tMax Pool Usage (MB):");
for (i=0; i<MAXALLOCPOOL; i++)
     uprintf("\t%5.2lf", (double)_u_pools[i].maxused/(double)(1024*1024));
uprintf("\tMax Pool Usage %%:");
for (i=0; i<MAXALLOCPOOL; i++)
     uprintf("\t%5.2lf", 100.0*(double)_u_pools[i].maxused/(double)(_u_pools[i].last-_u_pools[i].base));
#else
uprintf("\t----- Max usage not available--------");
#endif
}


void FreePool(int pool)
{
if (IsPoolAllocated(pool))
    {
    usfree(_u_pools[pool].base);
    _u_pools[pool].isallocated = 0;
    }
}

void FreePools()
{
int i;
for (i=0; i<MAXALLOCPOOL; i++)
    FreePool(i);
}


static char *_g_pool_names[MAXALLOCPOOL] =
{
"CellNet",
"Pins",
"General"
};

static void CheckPool(int pool, char *msg)
{
#ifdef _DEBUG_POOL_MEM
long used;
#endif

if (_u_pools[pool].ptr>=_u_pools[pool].last)
    {
    ShortReportMaxPoolUsages();
    errexit("%s : pool %d\n allocated %ld byte(s) but used %d byte(s).\nPlease increase MemMul_%s parameter.", msg, 
    	pool, _u_pools[pool].last-_u_pools[pool].base, _u_pools[pool].ptr-_u_pools[pool].base, _g_pool_names[pool]);
    }
#ifdef _DEBUG_POOL_MEM
used = _u_pools[pool].ptr-_u_pools[pool].base;
if (used > _u_pools[pool].maxused)
    _u_pools[pool].maxused = used;
#endif
}

pchar poolalloc(int pool, long size, char *msg)
{
pchar rp=_u_pools[pool].ptr;
#ifdef ALIGN64BIT
size = (long) (ceil((double) size/ 8.0)*8.0);
#endif 
#ifdef _DEBUG_POOL_MEM
if (!IsPoolAllocated(pool))
    errexit("Pool %d is not allocated yet", pool);
#endif
_u_pools[pool].ptr += size;
CheckPool(pool, msg);
return rp;
}



int ispoolavailable(int pool, long size)
{
#ifdef ALIGN64BIT
size = (long) (ceil((double) size/ 8.0)*8.0);
#endif 
return _u_pools[pool].last-_u_pools[pool].ptr>size;
}

void poolfree(int pool, long size, char *msg)
{
#ifdef ALIGN64BIT
size = (long) (floor((double) size/ 8.0)*8.0);
#endif 
#ifdef _DEBUG_POOL_MEM
if (!IsPoolAllocated(pool))
    errexit("Pool %d is not allocated yet", pool);
#endif 
if (size<0)
    errexit("%s : poolfree: pool=%d size=%d", msg, pool, size);
_u_pools[pool].ptr -= size;
}

pchar poolgetptr(int pool)
{
#ifdef _DEBUG_POOL_MEM
if (!IsPoolAllocated(pool))
    errexit("Pool %d is not allocated yet", pool);
#endif
return _u_pools[pool].ptr;
}

void poolfree_upto(int pool, pchar ptr, char *msg)
{
long size=_u_pools[pool].ptr-ptr;

if (!IsPoolAllocated(pool))
    errexit("Pool %d is not allocated yet", pool);

if (ptr && size<0)
    uwarning("%s : poolfree_upto invalid free size=%ld in pool %d", msg, size, pool);

if (ptr && ptr < _u_pools[pool].ptr)
    _u_pools[pool].ptr = ptr;
#ifdef _DEBUG_POOL_MEM
if (_u_pools[pool].ptr < _u_pools[pool].base)
    errexit("%s : poolfree_upto invalid free ptr < base in pool %d", msg, pool);
#endif
}


/*************************************************************************
* -------------------------- String --------------------------------------
**************************************************************************/

int uiswhitespace(char c)
{
return c==' ' || c=='\t' || c=='\n' || c=='\r';
}

pchar ultrim(pchar src)
{
int     i;
pchar   p;

for (p=src, i=0; uiswhitespace(*p); p++, i++);
if (i>0)
    {
    for (; *p; p++)
	*(p-i) = *p;
    *(p-i) = 0;
    }
return src;
}

pchar urtrim(pchar src)
{
pchar   p;
int     i=(int)(strlen(src)-1);

for (p=src+i; i>=0 && uiswhitespace(*p); p--, i--)
    *p = 0;
return src;
}

pchar utrim(pchar src)
{
return ultrim(urtrim(src));
}

pchar ustrcpy(pchar dest, pchar src)
{
if (!dest || !src) /* either one of them is null */
    return NULL;
else
    return (strcpy(dest, src));
}

pchar ustrncpy(pchar dest, pchar src, long n)
{
if (!dest || !src) /* either one of them is null */
    return NULL;
else
    return (strncpy(dest, src, n));
}

pchar ustrdup(pchar st)
{
pchar p;

if (!st)
	return NULL;
else
	{
	p = (pchar) umalloc(strlen(st)+1, "ustrdup");
	ustrcpy(p, st);
	return p;
	}
}


pchar GetTabSpace(int tabcnt)
{
static char spaces[MAXTABSPACELINE];
int i;

for (i=0; i<tabcnt*TABSIZE; i++)
    spaces[i] = ' ';
spaces[i] = 0;
return spaces;
}

/*************************************************************************
* -------------------------- Sort ----------------------------------------
**************************************************************************/

void isorttwo(pint i, pint j)
{
int t=*i;

if (*i>*j)
    {
    *i = *j;
    *j = t;
    }
}

void uwarning(pchar f_str,...)
{
char    *msg;
va_list argp;

va_start(argp, f_str);
vasprintf(&msg, f_str, argp);
va_end(argp);
internaldbgprintf(msg);


fflush(stdout);
fprintf(stderr, "\n*** Warning: %s ***\n", msg);
fflush(stderr);
free(msg);
}




#define SWAP(a, b, tmp)  \
                 do {(tmp) = (a); (a) = (b); (b) = (tmp);} while(0) 


void PermuteInPlace(pint a, int n)
{
#ifdef _USE_DRAND
int i, u, v, tmp;

if (n <= 4)
    return;

for (i=0; i<n; i+=16) 
    {
    u = uRandom(n-4);
    v = uRandom(n-4);
    SWAP(a[v], a[u], tmp);
    SWAP(a[v+1], a[u+1], tmp);
    SWAP(a[v+2], a[u+2], tmp);
    SWAP(a[v+3], a[u+3], tmp);
    }

#else
int i, end = n / 4;

for (i=1; i<end; i++)
    {
    int j=uRandom(n-i);
    int t=a[j];
    a[j] = a[n-i];
    a[n-i] = t;
    }
#endif
}


int ispowerof2(int n)
{
int pwr=2;

while (n>pwr)
    pwr *= 2;
return n==pwr;
}


void SwapByteArray_NC(void *p, int elemsz, int numelem)
{
 int i;
 
 for (i=0; i<numelem; ++i)
     SWAP_BYTES_NC(((char *)p)+i*elemsz, elemsz);
}

#ifdef NO_INLINE
int uGetSeed(void)
{
return _u_seed;
}

int uRandom(int l)
{
int   a = 16807;
int   m = 2147483647;
int   q = 127773;
int   r = 2836;
int   lo, hi, test;
double d;
 
lo = _u_seed % q;
hi = _u_seed / q;
test = (a*lo)-(r*hi);
if (test>0)
    _u_seed = test;
else
    _u_seed = test + m;
d = (double) ((double) _u_seed / (double) m);
return (int) (d*(double)l);
}


void iset(pint a, int to, int sz)
{
    int i;

    for (i=0; i<sz; ++i)
	a[i] = to;
}

void izero(pint a, int sz)
{
    memset(a, 0, sizeof(int)*sz);
}


void iadd(pint sum, pint add, int sz)
{
    int i;

    for (i=0; i<sz; ++i)
	sum[i] += add[i];
}

void isub(pint sum, pint sub, int sz)
{
    int i;

    for (i=0; i<sz; ++i)
	sum[i] -= sub[i];
}


int minindexof(pint lst, int size)
{
    int i, min;

    min = 0;
    for (i=1; i<size; i++)
	if (lst[i]<lst[min])
	    min = i;
    return min;
}

int maxindexof(pint lst, int size)
{
    int i, max;

    max = 0;
    for (i=1; i<size; i++)
	if (lst[i]>lst[max])
	    max = i;
    return max;
}


int minof(pint lst, int size)
{
    int i, min;

    min = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]<min)
	    min = lst[i];
    return min;
}

int maxof(pint lst, int size)
{
    int i, max;

    max = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]>max)
	    max = lst[i];
    return max;
}

int sumof(pint lst, int size)
{
    int i, sum=0;

    for (i=0; i<size; i++)
	sum += lst[i];
    return sum;
}

void iswap(pint p1, pint p2)
{
    int t=*p1;
    *p1 = *p2;
    *p2 = t;
}

void icopy(pint dest, pint src, int size)
{
    memcpy((pvoid )dest, (pvoid )src, sizeof(int)*size);
}


/* float versions */

void fset(pfloat a, float to, int sz)
{
    int i;

    for (i=0; i<sz; ++i)
	a[i] = to;
}

void fzero(pfloat a, int sz)
{
    fset(a, 0.0, sz);
}

void fadd(pfloat sum, pfloat add, int sz)
{
    int i;

    for (i=0; i<sz; ++i)
	sum[i] += add[i];
}

void fsub(pfloat sum, pfloat sub, int sz)
{
    int i;

    for (i=0; i<sz; ++i)
	sum[i] -= sub[i];
}

int fminindexof(pfloat lst, int size)
{
    int i, min;

    min = 0;
    for (i=1; i<size; i++)
	if (lst[i]<lst[min])
	    min = i;
    return min;
}

int fmaxindexof(pfloat lst, int size)
{
    int i, max;

    max = 0;
    for (i=1; i<size; i++)
	if (lst[i]>lst[max])
	    max = i;
    return max;
}


float fminof(pfloat lst, int size)
{
    int   i;
    float min;

    min = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]<min)
	    min = lst[i];
    return min;
}

float fmaxof(pfloat lst, int size)
{
    int   i;
    float max;

    max = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]>max)
	    max = lst[i];
    return max;
}

float fsumof(pfloat lst, int size)
{
    int   i;
    float sum=0;

    for (i=0; i<size; i++)
	sum += lst[i];
    return sum;
}


void fcopy(pfloat dest, pfloat src, int size)
{
    memcpy((pvoid )dest, (pvoid )src, sizeof(float)*size);
}


#endif

int isThisMachineBigEndian()
     /* MIPS  = Big Endian    */
     /* Intel = Little Endian */
{
  union {
    unsigned int testWord;
    char testByte[4];
  } endianTest;

  endianTest.testWord = 1;
  if( endianTest.testByte[0] == 1 )
    return 0;
  else
    return 1;
}


#define MAXLINE	16*1024*1024


pint read_array(char *fname, pint resvec, int n, int base)
{
FILE	*fp;
pchar 	line, newstr, oldstr;
int	i;

fp = ufopen(fname, "r", "read_vector");
if (!fp)
    return NULL;

line = (pchar) umalloc(MAXLINE, "read_vector : line");
for (i=0; i<n;)
    {
    int w;

    do			/* Skip Comments */
	{
	fgets(line, MAXLINE, fp);
	} while (line[0] == '%');
    oldstr = line;
    newstr = NULL;

    if (strlen(line) >= MAXLINE-5)
	errexit("read_vector : Buffer for fgets not big enough.");

    w = (int)strtol(oldstr, &newstr, 10);
    while (oldstr!=newstr)
	{
	resvec[base+i++] = w;
	oldstr = newstr;
	w = (int)strtol(oldstr, &newstr, 10);
	}
    }
usfree(line);
ufclose(fp);
return resvec;
}


#ifdef __cplusplus
#if 0
void* operator new(size_t sz)
{
 puts("new called\n");
 return umalloc(sz, "operator new");
}

void operator delete(void* m)
{
 puts("delete called\n"); 
 usfree(m);
}
#endif
#endif


#ifdef _MSC_VER 
#ifdef _DEBUG1
#pragma message( "Compiling " __FILE__ " with DEBUG1") 
#endif
#ifdef _DEBUG2
#pragma message( "Compiling " __FILE__ " with DEBUG2") 
#endif
#ifdef _DEBUG3
#pragma message( "Compiling " __FILE__ " with DEBUG3") 
#endif
#ifdef _DEBUG4
#pragma message( "Compiling " __FILE__ " with DEBUG4") 
#endif
#ifdef _DEBUG_POOL_MEM
#pragma message( "Compiling " __FILE__ " with _DEBUG_POOL_MEM") 
#endif
#ifdef _DEBUG_MALLOC_LOW
#pragma message( "Compiling " __FILE__ " with _DEBUG_MALLOC_LOW") 
#endif
#endif
