/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  ------------------------------------------------------------------
   File   : uport.h
   Author : Umit V. Catalyurek
   Date   : Jan 12, 2000
   ------------------------------------------------------------------
   Desc   : Utility functions that are OS dependent

   YOU MUST DEFINE ONE OS before compiling this code!!!
   ------------------------------------------------------------------
*/
#include "ulib.h"
#if defined(WIN32)
#include <mmsystem.h>
#endif

long _u_g_jobno=-1;


#if defined(WIN32)
void AppCleanup(void) 
{
 WSACleanup();
}



int AppStartup(void) 
{
 WORD wVersionRequested;
 char buf[500];
 LPWSADATA wsaData=(LPWSADATA)buf;
 int err; 
 wVersionRequested = MAKEWORD( 2, 2 ); 
 err = WSAStartup( wVersionRequested, wsaData );
 if ( err != 0 ) 
     {
         /* Tell the user that we could not find a usable */
         /* WinSock DLL.                                  */
     return err;
     }

 atexit(AppCleanup);  /* be sure this is called at all exit points */

 return 0;
}

int ProcessStartup(void)
{
 return 0;
}


#else /* means not WIN32 */

#include <signal.h>
static void sig_pipe(int arg, siginfo_t *psi, void *p)
{
 printf("sig_pipe(): Handler called ...\n"); 
#if !defined(SILENT)
 printf("sig_pipe(): Handler called ...\n");
#endif
}

int AppStartup(void)
{
 ProcessStartup();
 return 0;
}

int ProcessStartup(void)
{
 struct sigaction s;

     /* install sig handler for SIGPIPE - write to socket with other end gone */
 s.sa_sigaction = sig_pipe;
 s.sa_flags = SA_SIGINFO;
 if (sigaction(SIGPIPE, &s, NULL) != 0) 
     perror("Cannot install SIGPIPE handler");
 return 0;
}

#endif 



#ifdef MEASURE_ONLYCPUTIME

/* Cause to use getrusage function to get time */
#define _USE_RUSAGE 

double u_seconds(void)
{
#ifdef UNIX
#ifdef _USE_RUSAGE
double time;                         /* elapsed time in seconds */
struct rusage rusage;

getrusage(RUSAGE_SELF, &rusage);
time = ((rusage.ru_utime.tv_sec + rusage.ru_stime.tv_sec) +
	1.0e-6*(rusage.ru_utime.tv_usec + rusage.ru_stime.tv_usec));
return(time);
#else
 return (double) clock()*1.0e-6;
#endif
#endif

#ifdef WIN32
 return (double) clock() / (double) CLOCKS_PER_SEC;
#endif
}
#else
double u_seconds(void)
{
#if defined(WIN32)
 return (double) timeGetTime() / 1000.0; 
#else
 struct timeval tp;

 gettimeofday(&tp, NULL);
 
 return (double) tp.tv_sec + (double) tp.tv_usec / 1000000.0;
#endif
}

#endif


double u_wseconds(void)
{
#if defined(WIN32)
 return (double) timeGetTime() / 1000.0; 
#else
 struct timeval tp;

 gettimeofday(&tp, NULL);
 
 return (double) tp.tv_sec + (double) tp.tv_usec / 1000000.0;
#endif
}



/*************************************************************************
* -------------------------- Error Exit ----------------------------------
**************************************************************************/
void errexit(pchar f_str,...)
{
va_list argp;
#ifdef _DEBUG    
char *errst=NULL;
#endif
    
fflush(stdout);
fflush(stderr);
fprintf(stderr, "\n****** Error:\n");
va_start(argp, f_str);
#ifdef _DEBUG    
vasprintf(&errst, f_str, argp);
#else
vfprintf(stderr, f_str, argp);
#endif

va_end(argp);

#ifdef _DEBUG    
if (!errst) {
    va_start(argp, f_str);
    vfprintf(stderr, f_str, argp);
    va_end(argp);
} else {
    fprintf(stderr,"%s ******\n", errst);
    udbgprintf("\nError in Execution\n%s\n----------------------------------------------\n", errst);
    free(errst);
}
#else
fprintf(stderr, " ******\n");
#endif
fflush(stderr);
#ifdef UNIX
if (_u_g_jobno!=-1)
    {
    char hostname[200], donefname[300];
    FILE *donef;

    gethostname(hostname, sizeof(hostname));
    sprintf(donefname, "DONE.%s", hostname);
    donef = fopen(donefname, "w");
    fprintf(donef, "%ld", _u_g_jobno);
    fclose(donef);
    }
#endif


exit(1);
}



void CheckErr(char *file, int line)
{
errexit("Check_failed! File: %s Line: %s", file, line);
}
