/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  ------------------------------------------------------------------
   File   : uport.h
   Author : Umit V. Catalyurek
   Date   : Jan 12, 2000
   ------------------------------------------------------------------
   Desc   : Utility functions that are OS dependent

   YOU MUST DEFINE ONE OS before compiling this code!!!
   ------------------------------------------------------------------
*/
#ifndef _UPORT_H_
#define _UPORT_H_

#if defined(__APPLE__) && defined(__MACH__) && !defined(UNIX)
#define UNIX
#endif

#if defined(LINUX) || defined(_LINUX) || defined(__unix) || defined(__unix__) || defined(unix) || defined(_AIX)  || defined(TRU64UNIX) && !defined(UNIX)
#define UNIX
#endif

#if defined(TRU64UNIX) && !defined(ALIGN64BIT)
#define ALIGN64BIT
#endif


#ifndef UNIX
#ifndef WIN32
YOU MUST DEFINE ONE OS // this hould cause an error
#endif
#endif

#ifdef TRU64UNIX
typedef int socklen_t;
#endif


#define NO_INLINE


/*************************************************************************
* -------------------------- Error Exit ----------------------------------
**************************************************************************/

/* --------------------- Exit : Functions ---------------------------- */
void errexit(char *fmt,...);

void CheckErr(char *file, int line);

#define NEGCHECK(x)    if ((x)<0) CheckErr(__FILE__, __LINE__)
#define ZEROCHECK(x)   if (!(x))  CheckErr(__FILE__, __LINE__)



#ifdef UNIX
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <semaphore.h>

#define UINLINEKEYWORD  inline

typedef socklen_t           USOCKLEN;
typedef int                 USOCKET;
typedef unsigned long       UHOSTIP;
typedef int                 UPORT;


/* semaphore */
typedef	sem_t		    USEMA;

#define USEMA_INIT(s)			bzero(&(s), sizeof(s))
#define USEMA_DONE(s)			sem_destroy(&(s))
#define USEMA_CREATE(s, init, max)	sem_init(&(s), 0, (init));
#define USEMA_WAIT(s)			sem_wait(&(s))
#define USEMA_POST(s)		        sem_post(&(s))
#define USEMA_VAL(s, v)                 sem_getvalue(&(s), &(v))

/* mutex */
typedef pthread_mutex_t	    ULOCK;

#define ULOCK_INIT(u)	    pthread_mutex_init(&u, NULL)
#define ULOCK_DONE(u)	    pthread_mutex_destroy(&u)
#define ULOCK_LOCK(u)	    pthread_mutex_lock(&u)
#define ULOCK_UNLOCK(u)	    pthread_mutex_unlock(&u)

/* thread */
typedef pthread_t		    UTHREAD;



#define UTHREAD_START(t, func, arg) do { \
static pthread_attr_t _g_umitthreadattr; \
 pthread_attr_init(&_g_umitthreadattr); \
 pthread_attr_setdetachstate(&_g_umitthreadattr, PTHREAD_CREATE_DETACHED);\
 pthread_create(&t, &_g_umitthreadattr, func, arg); \
} while (0) 
 
#define UTHREAD_END(t)		    pthread_exit(NULL)

#define UTHREAD_FUNC                void *
#define	UTHREAD_RET		    NULL


#define INVALID_SOCKET -1
#define closesocket(sd)  close(sd)
#define readsocket(SD,BUF,LEN)  read((SD),(BUF),(LEN))
#define writesocket(SD,BUF,LEN) write((SD),(BUF),(LEN))

#endif

#ifdef WIN32

#define FD_SETSIZE	256

#include <io.h>
#include <time.h>
#include <dos.h>
#include <winsock.h>
#include <process.h>

#define UINLINEKEYWORD  __forceinline


#define sleep(n)

typedef int	            USOCKLEN;
typedef SOCKET              USOCKET;
typedef unsigned long       UHOSTIP;
typedef u_short             UPORT;

/* semaphore */
typedef	HANDLE		    USEMA;

#define USEMA_INIT(s)			(s) = 0
#define USEMA_DONE(s)			if (s) CloseHandle(s)
#define USEMA_CREATE(s, init, max)	ZEROCHECK((s) = CreateSemaphoreA(0, (init), (max), 0))
#define USEMA_WAIT(s)			WaitForSingleObject((s), INFINITE)
#define USEMA_POST(s)		        ReleaseSemaphore((s), 1, 0)
#define USEMA_VAL(s, v)                 v=s

/* mutex */
typedef CRITICAL_SECTION    ULOCK;

#define ULOCK_INIT(u)	    InitializeCriticalSection(&u)
#define ULOCK_DONE(u)	    DeleteCriticalSection(&u)
#define ULOCK_LOCK(u)	    EnterCriticalSection(&u)
#define ULOCK_UNLOCK(u)	    LeaveCriticalSection(&u)

/* thread */
typedef unsigned long		    UTHREAD;

#define UTHREAD_START(t, func, arg) _beginthread(func, 0, arg)
#define UTHREAD_END(t)		    _endthread()

#define UTHREAD_FUNC                void 
#define	UTHREAD_RET	

#define readsocket(SD,BUF,LEN)  recv((SD),(BUF),(LEN),0)
#define writesocket(SD,BUF,LEN) send((SD),(BUF),(LEN),0)

#endif





/*************************************************************************
* --------------------------- Time ---------------------------------------
**************************************************************************/

/* --------------------- Time : Functions ------------------------------ */
double u_seconds(void);
double u_wseconds(void);



int ProcessStartup(void);

int AppStartup(void);


#endif








