/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

   ------------------------------------------------------------------
   File   : uqsort.c Modifed from Numerical Recipes Software
   Author : Umit V. Catalyurek
   Date   : Jul 29, 99
   ------------------------------------------------------------------
   Desc   : QuickSort
   ------------------------------------------------------------------
*/

#include "uqsort.h"

#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;

#define M             7
#define NSTACK        50

void uqsort(int n, PSortItem arr)
{
int         i, ir=n, j, k, l=1;
int         jstack=0, istack[NSTACK], aval;
SortItem    a, temp;

--arr;
for (;;) 
    {
    if (ir-l < M) 
	{
	for (j=l+1;j<=ir;j++) 
	    {
	    a=arr[j];
	    aval = a.val;
	    for (i=j-1;i>=1;i--) 
		{
		if (arr[i].val <= aval) 
		    break;
		arr[i+1] = arr[i];
		}
	    arr[i+1]=a;
	    }
	if (jstack == 0) 
	    break;
	ir=istack[jstack--];
	l=istack[jstack--];
	} 
    else 
	{
	k=(l+ir) >> 1;
	SWAP(arr[k],arr[l+1])
	  if (arr[l+1].val > arr[ir].val) 
	      {
	      SWAP(arr[l+1],arr[ir])
		}
	if (arr[l].val > arr[ir].val) 
	    {
	    SWAP(arr[l],arr[ir])
	      }
	if (arr[l+1].val > arr[l].val) 
	    {
	    SWAP(arr[l+1],arr[l])
	      }
	i=l+1;
	j=ir;
	a=arr[l];
	aval = a.val;
	for (;;) 
	    {
	    do i++; while (arr[i].val < aval);
	    do j--; while (arr[j].val > aval);
	    if (j < i) break;
	    SWAP(arr[i],arr[j]);
	    }
	arr[l]=arr[j];
	arr[j]=a;
	jstack += 2;
	if (jstack > NSTACK) 
	    errexit("uqsort: NSTACK too small in sort.");
	if (ir-i+1 >= j-l) 
	    {
	    istack[jstack]=ir;
	    istack[jstack-1]=i;
	    ir=j-1;
	    } 
	else 
	    {
	    istack[jstack]=j-1;
	    istack[jstack-1]=l;
	    l=i;
	    }
	}
    }
}

void uqsort_dec(int n, PSortItem arr)
{
int i, j=n-1;

uqsort(n, arr);
for (i=0; i<j; ++i, --j)
    {
    SortItem a=arr[i];

    arr[i]=arr[j];
    arr[j]=a;
    }
}


void ufqsort(int n, PfSortItem arr)
{
int          i, ir=n, j, k, l=1;
int          jstack=0, istack[NSTACK];
float	     aval;
fSortItem    a, temp;

--arr;
for (;;) 
    {
    if (ir-l < M) 
	{
	for (j=l+1;j<=ir;j++) 
	    {
	    a=arr[j];
	    aval = a.val;
	    for (i=j-1;i>=1;i--) 
		{
		if (arr[i].val <= aval) 
		    break;
		arr[i+1] = arr[i];
		}
	    arr[i+1]=a;
	    }
	if (jstack == 0) 
	    break;
	ir=istack[jstack--];
	l=istack[jstack--];
	} 
    else 
	{
	k=(l+ir) >> 1;
	SWAP(arr[k],arr[l+1])
	  if (arr[l+1].val > arr[ir].val) 
	      {
	      SWAP(arr[l+1],arr[ir])
		}
	if (arr[l].val > arr[ir].val) 
	    {
	    SWAP(arr[l],arr[ir])
	      }
	if (arr[l+1].val > arr[l].val) 
	    {
	    SWAP(arr[l+1],arr[l])
	      }
	i=l+1;
	j=ir;
	a=arr[l];
	aval = a.val;
	for (;;) 
	    {
	    do i++; while (arr[i].val < aval);
	    do j--; while (arr[j].val > aval);
	    if (j < i) break;
	    SWAP(arr[i],arr[j]);
	    }
	arr[l]=arr[j];
	arr[j]=a;
	jstack += 2;
	if (jstack > NSTACK) 
	    errexit("uqsort: NSTACK too small in sort.");
	if (ir-i+1 >= j-l) 
	    {
	    istack[jstack]=ir;
	    istack[jstack-1]=i;
	    ir=j-1;
	    } 
	else 
	    {
	    istack[jstack]=j-1;
	    istack[jstack-1]=l;
	    l=i;
	    }
	}
    }
}


void ufqsort_dec(int n, PfSortItem arr)
{
int i, j=n-1;

ufqsort(n, arr);
for (i=0; i<j; ++i, --j)
    {
    fSortItem a=arr[i];

    arr[i]=arr[j];
    arr[j]=a;
    }
}


void uqsorti(int n, int *arr)
{
int         i, ir=n, j, k, l=1;
int         jstack=0, istack[NSTACK];
int         a, temp;

--arr;
for (;;) 
    {
    if (ir-l < M) 
	{
	for (j=l+1;j<=ir;j++) 
	    {
	    a=arr[j];
	    for (i=j-1;i>=1;i--) 
		{
		if (arr[i] <= a) 
		    break;
		arr[i+1] = arr[i];
		}
	    arr[i+1]=a;
	    }
	if (jstack == 0) 
	    break;
	ir=istack[jstack--];
	l=istack[jstack--];
	} 
    else 
	{
	k=(l+ir) >> 1;
	SWAP(arr[k],arr[l+1])
	  if (arr[l+1] > arr[ir]) 
	      {
	      SWAP(arr[l+1], arr[ir])
		}
	if (arr[l] > arr[ir]) 
	    {
	    SWAP(arr[l], arr[ir])
	      }
	if (arr[l+1] > arr[l]) 
	    {
	    SWAP(arr[l+1], arr[l])
	      }
	i=l+1;
	j=ir;
	a=arr[l];
	for (;;) 
	    {
	    do i++; while (arr[i] < a);
	    do j--; while (arr[j] > a);
	    if (j < i) break;
	    SWAP(arr[i], arr[j]);
	    }
	arr[l]=arr[j];
	arr[j]=a;
	jstack += 2;
	if (jstack > NSTACK) 
	    errexit("uqsort: NSTACK too small in sort.");
	if (ir-i+1 >= j-l) 
	    {
	    istack[jstack]=ir;
	    istack[jstack-1]=i;
	    ir=j-1;
	    } 
	else 
	    {
	    istack[jstack]=j-1;
	    istack[jstack-1]=l;
	    l=i;
	    }
	}
    }
}

#undef M
#undef NSTACK
#undef SWAP


