/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  ------------------------------------------------------------------
   File   : usocket.h
   Author : Umit V. Catalyurek
   Date   : Nov 2, 2000
   ------------------------------------------------------------------
   Desc   : Socket functions

   ------------------------------------------------------------------
*/
#ifndef _USOCKET_H_
#define _USOCKET_H_

#ifdef __cplusplus
extern "C" {
#endif


#include "uport.h"

#define MAX_PORT_TRY 500

/* if option is 0 this means don't set it; leave default */
typedef struct TAGUSocketOptions
{
    int nodelay;
    int sendbufsz;
    int recvbufsz;
    int nconnecttry;
} USocketOptions;

int UInitializeSocketOpt(USocketOptions *sopt);

UHOSTIP UGetLocalHostIP(void);
UHOSTIP UGetHostIP(char *hostname);
    

USOCKET UConnect_to_Host(char *hostname, UPORT port, USocketOptions *sopt);
USOCKET UConnect_to_Host_wIP(UHOSTIP ip, UPORT port, USocketOptions *sopt);

/* give 0 to start a server at any available port */
/* if bind fails it tries following port numbers */
/* it will start listening */
USOCKET UServerSocket(UPORT *pport, USocketOptions *sopt);
        /* waits in select and accepts and returns first connection */
USOCKET UWaitAndAccept_Socket(USOCKET sd, USocketOptions *sopt);
int UWaitAndAcceptAr_Socket(USOCKET *sl, int nsocket, int *acsl, USocketOptions *sopt);
        /* waitns in select and accepts and returns first connection
           with ip */
USOCKET UWaitAndAccept_Socket_IP(USOCKET sd, USocketOptions *sopt, UHOSTIP *ip);

    
int UWrite_Raw(USOCKET sd, const void *buf, int size);
int URead_Raw(USOCKET sd, void *buf, int size);

int UWrite_Str(USOCKET sd, const char *st);
char *URead_Str(USOCKET sd);

    
int UWrite_Int(USOCKET sd, int *ar, int n);
int URead_Int(USOCKET sd, int *ar, int n);

int UWrite_Long(USOCKET sd, long *ar, int n);
int URead_Long(USOCKET sd, long *ar, int n);


int UWrite_Float(USOCKET sd, float *ar, int n);
int URead_Float(USOCKET sd, float *ar, int n);


int UWrite_Double(USOCKET sd, double *ar, int n);
int URead_Double(USOCKET sd, double *ar, int n);

#ifdef __cplusplus
}
#endif

#endif
