#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <assert.h>

#include "dgraph.h"
#include "utils.h"
#include "initialBisection.h"
#include "initialPartitioning.h"
#include "undirPartitioning.h"
#include "dgraphPartition.h"
#include "refinement.h"
#include "refinementBis.h"
#include "rvcycle.h"

void iniPartKerDFS(MLGP_option opt, coarsen*  coars, ecType *edgecut, idxType *partcopy, MLGP_info* info)
{
    idxType* toporder = (idxType*) malloc(sizeof(idxType)*(coars->graph->nVrtx + 2));
    double* bound = (double*) malloc(sizeof(double) * opt.nbPart);
    double* lb_pw = (double*) malloc(sizeof(double) * opt.nbPart);
    int nbpartmin=-1, totbound;
    int allSameUBflag = 1, i, j;

    DFStopsort(coars->graph, toporder);
    coars->nbpart = 0;
    totbound = 0;
    for (i=0; i < opt.nbPart; i++) {
        bound[i] = opt.ub[i];
        totbound += (int) bound[i];
    }
    if (totbound < coars->graph->totvw) {
        totbound =0;
        for (i=0; i < opt.nbPart; i++) {
            bound[i] += 1;
            totbound += (int) bound[i];
        }
    }

    if (totbound < coars->graph->totvw)
        u_errexit("iniPartKerDFS: The problem is unfeasable with tot weight %d and tot bound %d.\n", (int) coars->graph->totvw, (int) totbound);

    for (i=0; i < opt.nbPart; i++) {
        lb_pw[i] = ((2.0 - opt.ratio) / (opt.ratio)) * bound[i];
    }

    while (coars->nbpart != opt.nbPart) {
        allSameUBflag=1;
        for (j=1;j<opt.nbPart;++j)
            if (fabs(bound[j]-bound[0])>0.0000001)
                allSameUBflag=0;
        if (allSameUBflag) {
            coars->nbpart = kernighanSequentialPart(coars->graph, toporder, lb_pw[0], bound[0], partcopy, edgecut, opt);
        }
        if (coars->nbpart != opt.nbPart)
            coars->nbpart = kernighanSequentialExactPart(coars->graph, toporder, opt.nbPart, bound, partcopy, edgecut, opt);

        if ((coars->nbpart != 1)&&(coars->nbpart < opt.nbPart)) {
            *edgecut = ecType_MAX;
            break;
        }
        for (i=0; i < opt.nbPart; i++) {
            bound[i] = bound[i] * 1.1;
            lb_pw[i] = lb_pw[i] * 0.9;
        }
    }

    KwayRefinementStep(opt, info->coars_depth, coars->graph, coars->part, coars->nbpart, info);

    free(lb_pw);
    free(bound);
    free(toporder);
}


void iniPartKerBFS(MLGP_option opt, coarsen*  coars, ecType *edgecut, idxType *partcopy, MLGP_info* info)
{
    idxType* toporder = (idxType*) malloc(sizeof(idxType)*(coars->graph->nVrtx + 2));
    double* bound = (double*) malloc(sizeof(double) * opt.nbPart);
    double* lb_pw = (double*) malloc(sizeof(double) * opt.nbPart);
    int nbpartmin=-1, totbound;
    int allSameUBflag = 1, i, j;

    BFStopsort(coars->graph, toporder);
    coars->nbpart = 0;
    totbound = 0;
    for (i=0; i < opt.nbPart; i++) {
        bound[i] = opt.ub[i];
        totbound += (int) bound[i];
    }
    if (totbound < coars->graph->totvw) {
        totbound =0;
        for (i=0; i < opt.nbPart; i++) {
            bound[i] += 1;
            totbound += (int) bound[i];
        }
    }
    if (totbound < coars->graph->totvw)
        u_errexit("iniPartKerDFS: The problem is unfeasable with tot weight %d and tot bound %d.\n", (int) coars->graph->totvw, (int) totbound);
    for (i=0; i < opt.nbPart; i++) {
        lb_pw[i] = ((2.0 - opt.ratio) / (opt.ratio)) * bound[i];
    }

    while (coars->nbpart != opt.nbPart) {
        allSameUBflag=1;
        for (j=1;j<opt.nbPart;++j)
            if (fabs(bound[j]-bound[0])>0.0000001)
                allSameUBflag=0;
        if (allSameUBflag) {
            coars->nbpart = kernighanSequentialPart(coars->graph, toporder, lb_pw[0], bound[0], partcopy, edgecut, opt);
        }
        if (coars->nbpart != opt.nbPart)
            coars->nbpart = kernighanSequentialExactPart(coars->graph, toporder, opt.nbPart, bound, partcopy, edgecut, opt);
        if ((coars->nbpart != 1)&&(coars->nbpart < opt.nbPart)) {
            *edgecut = ecType_MAX;
            break;
        }
        for (i=0; i < opt.nbPart; i++) {
            bound[i] = bound[i] * 1.1;
            lb_pw[i] = lb_pw[i] * 0.9;
        }
    }
    KwayRefinementStep(opt, info->coars_depth, coars->graph, coars->part, coars->nbpart, info);

    free(bound);
    free(lb_pw);
    free(toporder);
}

void iniPartKerMix(MLGP_option opt, coarsen*  coars, ecType *edgecut, idxType *partcopy, int pass, MLGP_info* info)
{
    idxType* toporder = (idxType*) malloc(sizeof(idxType)*(coars->graph->nVrtx + 2));
    double* bound = (double*) malloc(sizeof(double) * opt.nbPart);
    double* lb_pw = (double*) malloc(sizeof(double) * opt.nbPart);
    int nbpartmin=-1, totbound;
    int allSameUBflag = 1;
    int i,j;
    switch (pass) {
        case 1 : mixtopsort(coars->graph, toporder, 50, 0); break;
        case 2 : mixtopsort(coars->graph, toporder, 25, 0); break;
        case 3 : mixtopsort(coars->graph, toporder, 75, 0); break;
        case 4 : mixtopsort(coars->graph, toporder, 25, 1); break;
        case 5 : mixtopsort(coars->graph, toporder, 75, 1); break;
        default: mixtopsort(coars->graph, toporder, 50, 0); break;
    }

    coars->nbpart = 0;
    totbound = 0;
    for (i=0; i < opt.nbPart; i++) {
        bound[i] = opt.ub[i];
        totbound += (int) bound[i];
    }
    if (totbound < coars->graph->totvw) {
        totbound =0;
        for (i=0; i < opt.nbPart; i++) {
            bound[i] += 1;
            totbound += (int) bound[i];
        }
    }
    if (totbound < coars->graph->totvw)
        u_errexit("iniPartKerMix: The problem is unfeasable with tot weight %d and tot bound %d.\n", (int) coars->graph->totvw, (int) totbound);
    for (i=0; i < opt.nbPart; i++) {
        lb_pw[i] = ((2.0 - opt.ratio) / (opt.ratio)) * bound[i];
    }
    while (coars->nbpart != opt.nbPart) {
        allSameUBflag=1;
        for (j=1;j<opt.nbPart;++j)
            if (fabs(bound[j]-bound[0])>0.0000001)
                allSameUBflag=0;
        if (allSameUBflag) {
            coars->nbpart = kernighanSequentialPart(coars->graph, toporder, lb_pw[0], bound[0], partcopy, edgecut, opt);
        }
        if (coars->nbpart != opt.nbPart)
            coars->nbpart = kernighanSequentialExactPart(coars->graph, toporder, opt.nbPart, bound, partcopy, edgecut, opt);
        if ((coars->nbpart != 1)&&(coars->nbpart < opt.nbPart)) {
            *edgecut = ecType_MAX;
            break;
        }
        for (i=0; i < opt.nbPart; i++) {
            bound[i] = bound[i] * 1.1;
            lb_pw[i] = lb_pw[i] * 0.9;
        }
    }

    KwayRefinementStep(opt, info->coars_depth, coars->graph, coars->part, coars->nbpart, info);

    free(lb_pw);
    free(bound);
    free(toporder);
}


void iniPartForcedBalance(MLGP_option opt, coarsen*  coars, ecType *edgecut, idxType *partcopy, MLGP_info* info)
{
    coars->nbpart = opt.nbPart;
    initialPartitioningForcedBalance(coars->graph, opt.ub, partcopy, opt.nbPart, opt);
    KwayRefinementStep(opt, info->coars_depth, coars->graph, partcopy, coars->nbpart, info);
    *edgecut = edgeCut(coars->graph, partcopy);
}


void iniPartBFSGrowing(MLGP_option opt, MLGP_info* info, coarsen*  coars, ecType *edgecut, idxType *partcopy)
{
    coars->nbpart = opt.nbPart;
    idxType* toporder = (idxType*) malloc(sizeof(idxType)*coars->nbpart);
    idxType* sizes = (idxType*) malloc(sizeof(idxType)*coars->nbpart);

    BFSGrowing(coars->graph, opt.ub, partcopy, opt.nbPart, opt, info);
    randTopSortOnParts(coars->graph, partcopy, toporder, opt.nbPart);
    forcedBalance(coars->graph, toporder, opt.ub, partcopy, opt.nbPart, opt, info, sizes);
    KwayRefinementStep(opt, info->coars_depth, coars->graph, partcopy, coars->nbpart, info);
    *edgecut = edgeCut(coars->graph, partcopy);
    free(sizes);
    free(toporder);
}

void iniPartCocktailShake(MLGP_option opt, MLGP_info* info, coarsen*  coars, ecType *edgecut, idxType *partcopy)
{
    coars->nbpart = opt.nbPart;
    idxType* toporder = (idxType*) malloc(sizeof(idxType)*opt.nbPart);
    idxType* sizes = (idxType*) malloc(sizeof(idxType)*opt.nbPart);
    CocktailShake(coars->graph, opt.ub, partcopy, opt.nbPart, opt);
    randTopSortOnParts(coars->graph, partcopy, toporder, opt.nbPart);
    forcedBalance(coars->graph, toporder, opt.ub, partcopy, opt.nbPart, opt, info, sizes);
    KwayRefinementStep(opt, info->coars_depth, coars->graph, partcopy, coars->nbpart, info);
    *edgecut = edgeCut(coars->graph, partcopy);
    free(sizes);
    free(toporder);
}

void iniPartDiaGreedy(MLGP_option opt, coarsen*  coars, ecType *edgecut, idxType *partcopy, MLGP_info* info)
{
    coars->nbpart = opt.nbPart;
    peripheralRootGreedyGraphGrowing(coars->graph, opt.ub, partcopy, opt.nbPart, opt);
    KwayRefinementStep(opt, info->coars_depth, coars->graph, partcopy, coars->nbpart, info);
    *edgecut = edgeCut(coars->graph, partcopy);
}

void iniPartWCycle(MLGP_option opt, coarsen*  coars, ecType *edgecut, idxType *partcopy, MLGP_info* info)
{
    coars->nbpart = opt.nbPart;
    idxType i;
    for (i = 1; i <= coars->graph->nVrtx; i++) {
        partcopy[i] = coars->flag[i];
    }
    KwayRefinementStep(opt, info->coars_depth, coars->graph, partcopy, coars->nbpart, info);
    *edgecut = edgeCut(coars->graph, partcopy);
}


void initialPartitioning(MLGP_option opt, coarsen*  coars, MLGP_info* info)
{
    int tryAll[] = {IP_UNDIRRAND, IP_UNDIRGGG, IP_GGG_TWO};
    int ntryAll = sizeof(tryAll)/sizeof(tryAll[0]);


    idxType* partcopy = (idxType*) malloc(sizeof(idxType)*(coars->graph->nVrtx + 1));

    ecType edgecutmin = ecType_MAX, edgecut = ecType_MAX;

    for (int run = 0; run<opt.inipart_nrun; ++run) {
        int iniAlg = opt.inipart;

        if (iniAlg == IP_ALL )
            iniAlg = tryAll[run % ntryAll];

        info->timing_inipart_tab[iniAlg] -= u_wseconds();
        switch (iniAlg) {
            case IP_KERDFS :
                iniPartKerDFS(opt, coars, &edgecut, partcopy, info);
                break;

            case IP_KERBFS :
                iniPartKerBFS(opt, coars, &edgecut, partcopy, info);
                break;

            case IP_KERDFSBFS :
                iniPartKerMix(opt, coars, &edgecut, partcopy, run, info);
                break;

            case IP_FORCED :
                iniPartForcedBalance(opt, coars, &edgecut, partcopy, info);
                break;

            case IP_BFS:
                assert(opt.nbPart == 2);
                iniPartBFSGrowing(opt,info,coars, &edgecut, partcopy);
                break;

            case IP_CS:
                assert(opt.nbPart == 2);
                iniPartCocktailShake(opt, info, coars, &edgecut, partcopy);
                break;

            case IP_DIA_GGG: //NOTE: Don't use for testing
                assert(opt.nbPart == 2);
                iniPartDiaGreedy(opt, coars, &edgecut, partcopy, info);
                break;

            case IP_WCYCLE:
                iniPartWCycle(opt, coars, &edgecut, partcopy, info);
                break;

            case IP_RECBISS:
            {
                coars->nbpart = opt.nbPart;
                MLGP_option* optini = copyOpt(opt, opt.nbPart);
                optini->inipart = IP_ALL;

                rMLGP_info infoini;
                infoini.depth=0;
                initRInfoPart(&infoini);

                rcoarsen * rcoars = rVCycle(coars->graph, *optini, &infoini);

                for (int j = 1; j <=coars->graph->nVrtx; j++) {
                    partcopy[j] = rcoars->coars->part[j];
                }

                free_opt(optini);
                edgecut = edgeCut(coars->graph, partcopy);
                /* UVC: TODO: there was no free for rcoars */
                break;
            }

            default:
                u_errexit("initialPartitioning: unknown initial partitioning = %d", opt.inipart);
        }
        info->timing_inipart_tab[iniAlg] += u_wseconds();

        if (opt.print>=PD_HIGH)
            printf("InitPart:%d:%s Cut=%d\n", run, iniAlgName[iniAlg], (int)edgecut);
        info->ec_inipart_tab[iniAlg] = edgecut;
        if (edgecut < edgecutmin) {
            edgecutmin = edgecut;
            info->best_inipart = iniAlg;
            info->iniec = edgecut;
            for (int j = 0; j <= coars->graph->nVrtx; ++j)
                coars->part[j] = partcopy[j];
        }
    }
    if (opt.print>=PD_MEDIUM)
        printf("InitPart:Final:%s Cut=%d\n", iniAlgName[opt.inipart], (int)edgecutmin);
    if (opt.debug){
        int isAcyclic = checkAcyclicity(coars->graph, coars->part, opt.nbPart);
        if(isAcyclic == 0)
            u_errexit("initialPartitioning: the initial partition is not acyclic\n");
        else
            printf("initialPartitioning: the initial partition is acyclic\n");
    }
    
    free(partcopy);
    info->current_edgecut = edgecutmin;
}
