#include "dgraphTraversal.h"

void checkTopsort(dgraph *G, idxType *toporder)
{
    idxType i, j, neigh, nVrtx = G->nVrtx;
    idxType *outEnd = G->outEnd, *outStart = G->outStart, *out = G->out;
    idxType *topindex = (idxType *) malloc(sizeof(idxType)* (nVrtx+1));

    for (i = 1; i <= nVrtx; i++)
        topindex[toporder[i]] = i;

    for (i = 1; i <= nVrtx; i ++)
    {
        for ( j = outStart[i]; j<= outEnd[i]; j++)
        {
            neigh = out[j];
            if (topindex[i] >= topindex[neigh])
                u_errexit("The topological order is not correct %.0f %.0f\n", 1.0*topindex[i], 1.0*topindex[neigh]);
        }
    }
    free(topindex);
}

void natural_with_part(dgraph *G, idxType* toporder, idxType *part, idxType *toporderpart, idxType nbpart){
    idxType i,j, cur_part;
    idxType to = 1;

    idxType *nbnodes = calloc(nbpart, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbnodes[part[i]]++;

    idxType *begin = malloc(nbpart * sizeof(idxType));
    begin[toporderpart[0]] = 1;
    for (i = 1; i < nbpart; i++)
        begin[toporderpart[i]] = begin[toporderpart[i-1]] + nbnodes[toporderpart[i-1]];

    for (i = 1; i <= G->nVrtx; i++)
        toporder[begin[part[i]]++] = i;
}

void topSortOnParts(dgraph *G, idxType *part, idxType *toporder, idxType nbpart)
{
    /*Fill toporder assuming it's already allocated*/
    idxType i, j, k;
    idxType** outpart = (idxType**) malloc(sizeof(idxType*)*nbpart);
    idxType* nbout = (idxType*) malloc(sizeof(idxType)*nbpart);
    idxType** inpart = (idxType**) malloc(sizeof(idxType*)*nbpart);
    idxType* nbin = (idxType*) malloc(sizeof(idxType)*nbpart);
    for (i=0; i<nbpart; i++) {
        outpart[i] = (idxType*) malloc(sizeof(idxType)*nbpart);
        nbout[i] = 0;
        inpart[i] = (idxType*) malloc(sizeof(idxType)*nbpart);
        nbin[i] = 0;
    }
    for (i=1; i<=G->nVrtx; i++) {
        for (j=G->outStart[i]; j<=G->outEnd[i]; j++) {
            idxType outnode = G->out[j];
            if (part[i] == part[outnode])
                continue;
            int is_new = 1;
            for (k=0; k<nbout[part[i]]; k++)
                if (outpart[part[i]][k] == part[outnode]) {
                    is_new = 0;
                    break;
                }
            if (is_new == 1) {
                outpart[part[i]][nbout[part[i]]] = part[outnode];
                nbout[part[i]]++;
            }
            is_new = 1;
            for (k=0; k<nbin[part[outnode]]; k++)
                if (inpart[part[outnode]][k] == part[i]) {
                    is_new = 0;
                    break;
                }
            if (is_new == 1) {
                inpart[part[outnode]][nbin[part[outnode]]] = part[i];
                nbin[part[outnode]]++;
            }
        }
    }
    idxType* ready = (idxType*) malloc(sizeof(idxType)*(nbpart));
    idxType* nbinleft = (idxType*) malloc(sizeof(idxType)*(nbpart));
    int nbready = 0;
    for (i=0; i<nbpart; i++) {
        nbinleft[i] = nbin[i];
        if (nbin[i] == 0)
            ready[nbready++] = i;
    }

    int to = 0;

    while (nbready > 0) {
        idxType pno = ready[nbready-1];
        nbready--;
        toporder[to++] = pno;
        for (i = 0; i < nbout[pno]; i++) {
            idxType succ = outpart[pno][i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0)
                ready[nbready++] = succ;
        }
    }
    if (to != nbpart) {
        u_errexit("In topsortPart, not all sort: to = %d, nbpart = %d\n", to, nbpart);
    }
    for (i=0; i<nbpart; i++) {
        free(outpart[i]);
        free(inpart[i]);
    }
    free(ready);
    free(nbinleft);
    free(nbin);
    free(nbout);
    free(inpart);
    free(outpart);
}

void randTopSortOnParts(dgraph *G, idxType *part, idxType *toporder, idxType nbpart)
{
    /*Fill toporder assuming it's already allocated*/
    idxType i, j, k, ip;
    idxType** outpart = (idxType**) malloc(sizeof(idxType*)*nbpart);
    idxType* nbout = (idxType*) malloc(sizeof(idxType)*nbpart);
    idxType** inpart = (idxType**) malloc(sizeof(idxType*)*nbpart);
    idxType* nbin = (idxType*) malloc(sizeof(idxType)*nbpart);
    for (i=0; i<nbpart; i++) {
        outpart[i] = (idxType*) malloc(sizeof(idxType)*nbpart);
        nbout[i] = 0;
        inpart[i] = (idxType*) malloc(sizeof(idxType)*nbpart);
        nbin[i] = 0;
    }
    for (i=1; i<=G->nVrtx; i++) {
        for (j=G->outStart[i]; j<=G->outEnd[i]; j++) {
            idxType outnode = G->out[j];
            if (part[i] == part[outnode])
                continue;
            int is_new = 1;
            for (k=0; k<nbout[part[i]]; k++)
                if (outpart[part[i]][k] == part[outnode]) {
                    is_new = 0;
                    break;
                }
            if (is_new == 1) {
                outpart[part[i]][nbout[part[i]]] = part[outnode];
                nbout[part[i]]++;
            }
            is_new = 1;
            for (k=0; k<nbin[part[outnode]]; k++)
                if (inpart[part[outnode]][k] == part[i]) {
                    is_new = 0;
                    break;
                }
            if (is_new == 1) {
                inpart[part[outnode]][nbin[part[outnode]]] = part[i];
                nbin[part[outnode]]++;
            }
        }
    }

    idxType* ready = (idxType*) malloc(sizeof(idxType)*(nbpart));
    idxType* nbinleft = (idxType*) malloc(sizeof(idxType)*(nbpart));
    int nbready = 0;
    for (i=0; i<nbpart; i++) {
        nbinleft[i] = nbin[i];
        if (nbin[i] == 0)
            ready[nbready++] = i;
    }

    int to = 0;
    idxType* shuffle = (idxType *) malloc(nbpart * sizeof(idxType));
    while (nbready > 0) {
        idxType pno = ready[nbready-1];
        nbready--;
        toporder[to++] = pno;
        if (nbout[pno]) {
            shuffleTab(0, nbout[pno]-1, shuffle);
            for (ip = 0; ip < nbout[pno]; ip++) {
                i = shuffle[ip];
                idxType succ = outpart[pno][i];
                nbinleft[succ]--;
                if (nbinleft[succ] == 0) {
                    ready[nbready++] = succ;
                }
            }
        }
    }
    for (i=0; i<nbpart; i++) {
        free(outpart[i]);
        free(inpart[i]);
    }
    free(nbinleft);
    free(shuffle);
    free(ready);
    free(nbin);
    free(nbout);
    free(inpart);
    free(outpart);
    if (to != nbpart) {
        u_errexit("In topsortPart, not every nodes are sorted: to = %d, nbpart = %d\n", to, nbpart);
    }
}

void topsortwithTies(dgraph *G, idxType *toporder)
{
    /*topologically sorts the vertices of the dag G, when there are ties,
     chooses the one with the max (indegree -outdegree).

     Assumes toporder is allocated (of size nVrtx + 1).
     */

    idxType i, j, neigh, nVrtx = G->nVrtx;
    idxType  *inEnd = G->inEnd,  *inStart = G->inStart;
    idxType *outEnd = G->outEnd, *outStart = G->outStart, *out = G->out;

    idxType *inDegrees = (idxType*) malloc(sizeof(idxType)*2*(G->nVrtx+1)); /*get working storage*/
    idxType heapsz = 0; /*will be from 1 to heapsz*/
    idxType outtop = 1;
    idxType maxind;

    idxType *heapIds = inDegrees + nVrtx + 1; /*get working storage*/

    for (i = 1; i <= nVrtx; i++)
    {
        inDegrees[i] = inEnd[i]- inStart[i] + 1;
        if (inDegrees[i] == 0)
            heapIds[++heapsz] = i;
    }

    maxBuildHeap(G, heapIds, heapsz);
    /*choose from indegzerolist the one with the largest key. toporder it. reduce the indeg of its neighbors and add them to indegzerolist if necessary*/
    for (outtop = 1; outtop <= nVrtx; outtop ++)
    {
        maxind = maxHeapExtract(G, heapIds, &heapsz);
        toporder[outtop] = maxind;
        for ( j = outStart[maxind]; j<= outEnd[maxind]; j++)
        {
            neigh = out[j];
            inDegrees[neigh] --;
            if (inDegrees[neigh] == 0)
                maxHeapInsert(G, heapIds, &heapsz, neigh);
            else if (inDegrees[neigh] == -1)
                u_errexit("wrong index\n");
        }
    }
    free(inDegrees);
}

void topsortPriorities(dgraph *G, idxType *toporder, ecType *priority)
{
    /*Assume that toporder is already allocated an priorities filled
     Fill toporder to have a topological order of nodes in G
    with DFS*/
    idxType to = 1, i, node;
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    //Build heap of ready tasks
    idxType* heap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* inheap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType hsize = 0;
    for (i =0; i < G->nbsources; i++){
        hsize++;
        heap[hsize] = G->sources[i];
        inheap[G->sources[i]] = hsize;
    }
    heapBuild(heap, priority, hsize, inheap);

    while (hsize > 0) {
        if (to > G->nVrtx)
            u_errexit("Proof topsortPrioritiesto = %d\n", to);
        node = heapExtractMax(heap, priority, &hsize, inheap);
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0)
                heapInsert(heap, priority, &hsize, succ, inheap);
            else if (nbinleft[succ]<0)
                u_errexit("topsortPriorities: negative indegree\n");
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("topsortPriorities : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(heap);
    free(inheap);
}

void DFStopsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS*/
    idxType to = 1;
    idxType* ready = (idxType*) malloc(sizeof(idxType) * (G->nVrtx+1));
    idxType i;
    idxType nbready = G->nbsources;
    for (i = 0; i < nbready; i++)
        ready[i] = G->sources[i];
    sourcesList(G, ready);
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    while (nbready > 0) {
        if (to > G->nVrtx)
            u_errexit("Proof DFStopsort to = %d\n", to);
        idxType node = ready[nbready-1];
        nbready--;
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0)
                ready[nbready++] = succ;
            else if (nbinleft[succ]<0)
                u_errexit("DFStopsort: negative indegree\n");
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("DFStopsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(ready);
}

void BFStopsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS*/
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i;
    idxType endready = sourcesList(G, ready) - 1;
    idxType beginready = 0;
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    while (endready >= beginready) {
        if (to > G->nVrtx)
            u_errexit("Proof to = %d\n", to);
        idxType node = ready[beginready];
        beginready++;
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0) {
                ready[++endready] = succ;
            }
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("BFStopsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(ready);
}

void DFSsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS*/
    idxType to = 1;
    idxType* ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i;
    idxType nbready = sourcesList(G, ready);
    idxType* done = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    while (nbready > 0)
    {
        if (to > G->nVrtx)
            u_errexit("Proof to = %d\n", to);
        idxType node = ready[nbready-1];
        nbready--;
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++)
        {
            idxType succ = G->out[i];
            if (done[succ] == 1)
                continue;
            ready[nbready++] = succ;
            done[succ] = 1;
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("DFSsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(done);
    free(ready);
}

void BFSsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
     with DFS*/
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i;
    idxType endready = sourcesList(G, ready) - 1;
    idxType beginready = 0;
    idxType* done = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    while (endready >= beginready)
    {
        if (to > G->nVrtx)
            u_errexit("Proof to = %d\n", to);
        idxType node = ready[beginready];
        beginready++;
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            if (done[succ] == 1)
                continue;
            ready[++endready] = succ;
            done[succ] = 1;
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("BFSsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(done);
    free(ready);
}

void randDFStopsort(dgraph *G, idxType *toporder)
{
    /*Assume that toporder is already allocated
      Fill toporder to have a random topological order of nodes in G*/
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i,itmp;
    idxType nbready = sourcesList(G, ready);
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    idxType* shuffle = (idxType*) malloc(sizeof(idxType)*(G->maxoutdegree));
    while (nbready > 0)
    {
        idxType node = ready[nbready-1];
        nbready--;
        toporder[to++] = node;
        shuffleTab(G->outStart[node], G->outEnd[node], shuffle);
        for (itmp = 0; itmp <= G->outEnd[node]-G->outStart[node]; itmp++) {
            i = shuffle[itmp];
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0)
                ready[nbready++] = succ;
        }
    }
    free(shuffle);
    free(nbinleft);
    free(ready);
}

void randBFStopsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS*/
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i,itmp;
    idxType endready = sourcesList(G, ready) - 1;
    idxType beginready = 0;
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    idxType* shuffle = (idxType*) malloc(sizeof(idxType)*(G->maxoutdegree));
    while (endready >= beginready) {
        if (to > G->nVrtx)
            u_errexit("Proof to = %d\n", to);
        idxType node = ready[beginready];
        beginready++;
        toporder[to++] = node;
        shuffleTab(G->outStart[node], G->outEnd[node], shuffle);
        for (itmp = 0; itmp <= G->outEnd[node]-G->outStart[node]; itmp++) {
            i = shuffle[itmp];
            idxType succ = G->out[i];
            assert(succ <= G->nVrtx);
            nbinleft[succ]--;
            if (nbinleft[succ] == 0) {
                ready[++endready] = succ;
            }
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("randBFStopsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(shuffle);
    free(ready);
}

void randDFSsort(dgraph *G, idxType *toporder)
{
    /*Assume that toporder is already allocated
      Fill toporder to have a random topological order of nodes in G*/
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i,itmp;
    idxType nbready = sourcesList(G, ready);
    idxType* done = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    idxType* shuffle = (idxType*) malloc(sizeof(idxType)*(G->maxoutdegree+1));
    while (nbready > 0) {
        idxType node = ready[nbready-1];
        nbready--;
        toporder[to++] = node;
        shuffleTab(G->outStart[node], G->outEnd[node], shuffle);
        for (itmp = 0; itmp <= G->outEnd[node]-G->outStart[node]; itmp++) {
            i = shuffle[itmp];
            idxType succ = G->out[i];
            if (done[succ] == 1)
                continue;
            ready[nbready++] = succ;
            done[succ] = 1;
        }
    }
    free(shuffle);
    free(done);
    free(ready);
}

void randBFSsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS*/
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i,itmp;
    idxType endready = sourcesList(G, ready) - 1;
    idxType beginready = 0;
    idxType* done = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    idxType* shuffle = (idxType*) malloc(sizeof(idxType)*(G->maxoutdegree));
    while (endready >= beginready)
    {
        if (to > G->nVrtx)
            u_errexit("Proof to = %d\n", to);
        idxType node = ready[beginready];
        beginready++;
        toporder[to++] = node;
        shuffleTab(G->outStart[node], G->outEnd[node], shuffle);
        for (itmp = 0; itmp <= G->outEnd[node]-G->outStart[node]; itmp++) {
            i = shuffle[itmp];
            idxType succ = G->out[i];
            if (done[succ] == 1)
                continue;
            ready[++endready] = succ;
            done[succ] = 1;
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("randBFSsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(done);
    free(shuffle);
    free(ready);
}


void randTopsort(dgraph* G, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a random topological order of nodes in G
    */
    idxType to = 1;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType i;
    idxType nbready = sourcesList(G, ready);
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    while (nbready > 0) {
        if (to > G->nVrtx)
            u_errexit("Proof to = %d\n", to);
        idxType node_idx = uRandom(nbready);
        idxType node = ready[node_idx];
        ready[node_idx] = ready[nbready-1];
        nbready--;
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0) {
                ready[nbready++] = succ;
            }
        }
    }
    if (to != G->nVrtx + 1)
        u_errexit("randTopsort : Not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(ready);
}


int gcd ( int a, int b )
{
    int c;
    while ( a != 0 ) {
        c = a; a = b%a;  b = c;
    }
    return b;
}

void mixtopsort(dgraph* G, idxType* toporder, int priority, int first)
{
    //first==0 --> bfs first
    //first==1 --> dfs first
    idxType* ready = (idxType*)malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* mark = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    idxType* nbinleft = (idxType *) calloc(G->nVrtx+1, sizeof(idxType));
    idxType readyend = sourcesList(G,ready) - 1;
    idxType readybegin = 0;
    idxType bfrac=0, dfrac=0;
    idxType biter=0, diter=0;
    idxType node=-1;
    idxType i,j,to=1;
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;

    i=gcd(priority,100-priority);
    if (i==0) {
        bfrac=priority/100;
        dfrac=(100-priority)/100;
    }
    else{
        bfrac=priority/i;
        dfrac=(100-priority)/i;
    }
    if (first)
        diter=dfrac;
    else
        biter=bfrac;

    while (readyend>=readybegin) {
        if (to > G->nVrtx+1) {
            printf("Proof to = %d\n", to);
            break;
        }

        //pick from bfs priority
        if (biter>=0) {
            node = ready[readybegin];
            readybegin++;
            biter--;
            if (biter<0)
                diter=dfrac;
        }
        else { //pick from dfs priority
            node = ready[readyend];
            diter--;
            readyend--;
            if (diter<0)
                biter=bfrac;
        }
        mark[node]=1;
        toporder[to++] = node;
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0) {
                readyend++;
                ready[readyend] = succ;
            }
        }
    }
    if (to != G->nVrtx+1)
        u_errexit("Mixtopsort : Not every node concerned!  to = %d, nVrtx = %d\n", to, G->nVrtx);

    free(nbinleft);
    free(mark);
    free(ready);
}

void DFStopsort_with_part(dgraph* G, idxType *part, int nbpart, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS respecting the convexity of the partitioning
    (every nodes of one part, then another part...)*/
    idxType i,j;
    idxType* toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
    topSortOnParts(G, part, toporderpart, nbpart);
    idxType* topsortpart = (idxType*) malloc (sizeof(idxType) * nbpart);

    for (i = 0; i<nbpart; i++) {
        topsortpart[toporderpart[i]] = i;
    }

    idxType to = 1;
    int current_part = 0;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType nbready = sourcesList(G, ready);

    idxType* nbinleft;
    nbinleft = (idxType*) malloc(sizeof(idxType) * (G->nVrtx+1));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    while (nbready > 0)
    {
        idxType node = -1;
        for (i = 1; i<=nbready; i++) {
            if (part[ready[nbready-i]] == toporderpart[current_part]) {
                node = ready[nbready-i];
                break;
            }
        }
        if (node == -1) {
            current_part++;
            continue;
        }
        for (j = nbready-i; j<nbready-1; j++) {
            ready[j] = ready[j+1]; /*BU2JH: a regarder*/
        }

        nbready--;
        toporder[to++] = node;

        //printf("Loop nbready = %d, node = %d, to = %d, part = %d\n", (int) nbready, (int) node, (int) to, part[node]);
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0)
                ready[nbready++] = succ;
        }
    }
    if (to != G->nVrtx +1)
        u_errexit("DFStopsort_with_part: not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(toporderpart);
    free(topsortpart);
    free(ready);
}


void randDFStopsort_with_part(dgraph* G, idxType *part, int nbpart, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS respecting the convexity of the partitioning
    (every nodes of one part, then another part...)*/
    idxType i,j, itmp;
    idxType* toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
    topSortOnParts(G, part, toporderpart, nbpart);
    idxType* topsortpart = (idxType*) malloc (sizeof(idxType) * nbpart);

    for (i = 0; i<nbpart; i++) {
        topsortpart[toporderpart[i]] = i;
    }

    idxType to = 1;
    int current_part = 0;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType nbready = sourcesList(G, ready);
    idxType* shuffle = (idxType*) malloc(sizeof(idxType)*(G->maxoutdegree));

    idxType* nbinleft;
    nbinleft = (idxType*) malloc(sizeof(idxType) * (G->nVrtx+1));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    while (nbready > 0)
    {
        idxType node = -1;
        for (i = 1; i<=nbready; i++) {
            if (part[ready[nbready-i]] == toporderpart[current_part]) {
                node = ready[nbready-i];
                break;
            }
        }
        if (node == -1) {
            current_part++;
            continue;
        }
        for (j = nbready-i; j<nbready-1; j++) {
            ready[j] = ready[j+1]; /*BU2JH: a regarder*/
        }

        nbready--;
        toporder[to++] = node;
        shuffleTab(G->outStart[node], G->outEnd[node], shuffle);

        //printf("Loop nbready = %d, node = %d, to = %d, part = %d\n", (int) nbready, (int) node, (int) to, part[node]);
        for (itmp = 0; itmp <= G->outEnd[node]-G->outStart[node]; itmp++)
        {
            i = shuffle[itmp];
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0)
                ready[nbready++] = succ;
        }
    }
    if (to != G->nVrtx +1)
        u_errexit("RandDFStopsort_with_part, not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(shuffle);
    free(nbinleft);
    free(toporderpart);
    free(topsortpart);
    free(ready);
}


void BFStopsort_with_part(dgraph* G, idxType *part, int nbpart, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS respecting the convexity of the partitioning
    (every nodes of one part, then another part...)*/
    idxType i,j;
    idxType* toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
    topSortOnParts(G, part, toporderpart, nbpart);
    idxType* topsortpart = (idxType*) malloc (sizeof(idxType) * nbpart);

    for (i = 0; i<nbpart; i++) {
        topsortpart[toporderpart[i]] = i;
    }

    idxType to = 1;
    int current_part = 0;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType nbready = sourcesList(G, ready);

    idxType* nbinleft;
    nbinleft = (idxType*) malloc(sizeof(idxType) * (G->nVrtx+1));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;;
    while (nbready > 0)
    {
        idxType node = -1;
        for (i = 1; i<=nbready; i++) {
            if (part[ready[nbready-i]] == toporderpart[current_part]) {
                node = ready[nbready-i];
                break;
            }
        }
        if (node == -1) {
            current_part++;
            continue;
        }
        for (j = nbready-i; j<nbready-1; j++) {
            ready[j] = ready[j+1];  /*BU2JH: a regarder*/
        }

        nbready--;
        toporder[to++] = node;

        //printf("Loop nbready = %d, node = %d, to = %d, part = %d\n", (int) nbready, (int) node, (int) to, part[node]);
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++) {
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0) {
                //		printf("succ = %d is ready\n",succ);
                for (j = nbready-1; j>=0; j--)
                    ready[j+1] = ready[j];  /*BU2JH: a regarder*/
                ready[0] = succ;
                nbready++;
            }
        }
    }
    if (to != G->nVrtx + 1)
        u_errexit("BFStopsort_with_part, not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(toporderpart);
    free(topsortpart);
    free(ready);
}

void randBFStopsort_with_part(dgraph* G, idxType *part, int nbpart, idxType *toporder)
{
    /*Assume that toporder is already allocated
     Fill toporder to have a topological order of nodes in G
    with DFS respecting the convexity of the partitioning
    (every nodes of one part, then another part...)*/
    idxType i,j,itmp;
    idxType* toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
    topSortOnParts(G, part, toporderpart, nbpart);
    idxType* topsortpart = (idxType*) malloc (sizeof(idxType) * nbpart);

    for (i = 0; i<nbpart; i++) {
        topsortpart[toporderpart[i]] = i;
    }

    idxType to = 1;
    int current_part = 0;
    idxType* ready;
    ready = (idxType*) malloc(sizeof(idxType) * G->nVrtx);
    idxType nbready = sourcesList(G, ready);

    idxType* nbinleft;
    nbinleft = (idxType*) malloc(sizeof(idxType) * (G->nVrtx+1));
    idxType* shuffle = (idxType*) malloc(sizeof(idxType)*(G->maxoutdegree));
    for (i = 1; i <= G->nVrtx; i++)
        nbinleft[i] = G->inEnd[i] - G->inStart[i] + 1;
    while (nbready > 0)
    {
        idxType node = -1;
        for (i = 1; i<=nbready; i++) {
            if (part[ready[nbready-i]] == toporderpart[current_part]) {
                node = ready[nbready-i];
                break;
            }
        }
        if (node == -1) {
            current_part++;
            continue;
        }
        for (j = nbready-i; j<nbready-1; j++) {
            ready[j] = ready[j+1];  /*BU2JH: a regarder*/
        }

        nbready--;
        toporder[to++] = node;
        shuffleTab(G->outStart[node], G->outEnd[node], shuffle);

        //printf("Loop nbready = %d, node = %d, to = %d, part = %d\n", (int) nbready, (int) node, (int) to, part[node]);
        for (itmp = 0; itmp <= G->outEnd[node]-G->outStart[node]; itmp++) {
            i = shuffle[itmp];
            idxType succ = G->out[i];
            nbinleft[succ]--;
            if (nbinleft[succ] == 0) {
                //		printf("succ = %d is ready\n",succ);
                for (j = nbready-1; j>=0; j--)
                    ready[j+1] = ready[j];  /*BU2JH: a regarder*/
                ready[0] = succ;
                nbready++;
            }
        }
    }
    if (to != G->nVrtx + 1)
        u_errexit("RandBFStopsort_with_part, not every node concerned! to = %d, nVrtx = %d\n", to, G->nVrtx);
    free(nbinleft);
    free(toporderpart);
    free(topsortpart);
    free(shuffle);
    free(ready);
}