#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "runtime.h"
#include "utils.h"
#include <limits.h>
#include <math.h>
#include "dgraphTraversal.h"


void computePriorities(dgraph* G, ecType* priority, MLGP_option opt){
    idxType* toplevels = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    computeWeightedBottomlevels(G, toplevels);
    for (idxType i =1; i <= G->nVrtx; i++)
        priority[i] = (ecType) toplevels[i];
    free(toplevels);
}

int cmpfunc (const void * a, const void * b) {
    parent *parentA = (parent *)a;
    parent *parentB = (parent *)b;
    return ( parentA->finish_time - parentB->finish_time );
}

void sortParents(parent* table, idxType size){
    idxType i;
    qsort(table, size, sizeof(parent), cmpfunc);
}


ecType EFTruntime(dgraph* G, idxType nbproc, MLGP_option opt){
    //Initialize everything
    idxType i, node, proc, p, dad, child, best_proc, indegree;
    ecType begin_time, best_begin_time, ready, makespan = 0;
    parent *parents = (parent*) malloc((G->nVrtx +1) * sizeof(parent));
    ecType *end_comp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *finish_time = (ecType*) calloc((G->nVrtx +1), sizeof(ecType));
    idxType *mapping = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    idxType *incoming = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    for(i=1; i<=G->nVrtx; i++) {
        mapping[i] = -1;
        incoming[i] = G->inEnd[i] - G->inStart[i] + 1;
    }

    //Compute priorities
    ecType* priority = (ecType*) malloc((G->nVrtx +1) * sizeof(ecType));
    computePriorities(G, priority, opt);

    //Build heap of ready tasks
    idxType* heap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* inheap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType hsize = 0;
    for (i =0; i < G->nbsources; i++){
        hsize++;
        heap[hsize] = G->sources[i];
        inheap[G->sources[i]] = hsize;
    }
    heapBuild(heap, priority, hsize, inheap);

    //Phase 2 of EFT algorithm
    while(hsize > 0){
        node = heapExtractMax(heap, priority, &hsize, inheap);
        //Let sort parents according to their finish time
        indegree = 0;
        //printf("Considering node %d.\n", (int) node);
        for (i = G->inStart[node]; i <= G->inEnd[node]; i++) {
            dad = G->in[i];
            if (mapping[dad] < 0)
                u_errexit("Problem in EFTruntime (running %d before %d)\n", node, dad);
            parents[indegree].idx = i;
            parents[indegree].finish_time = finish_time[dad];
            //printf("Put in parents[%d]: idx %d finish_time %d\n", (int) indegree, (int) parents[indegree].idx, (int) parents[indegree].finish_time);
            indegree++;
        }
        sortParents(parents, indegree);

        //Let find the best proc for node
        best_begin_time = ecType_MAX;
        for (proc = 0; proc < nbproc; proc++){
            //Let compute begin_time if we put node on proc
            begin_time = end_comp[proc];
            for (p = 0; p < nbproc; p++) {
                end_receive_tmp[p] = end_receive[p];
                end_send_tmp[p] = end_send[p];
            }
            for (i = 0; i < indegree; i++){
                dad = G->in[parents[i].idx];
                if (mapping[dad] == proc)
                    ready = finish_time[dad];
                else {
                    ready = finish_time[dad] < end_send_tmp[mapping[dad]] ? end_send_tmp[mapping[dad]] : finish_time[dad];
                    ready = ready < end_receive_tmp[proc] ? end_receive_tmp[proc] : ready;
                    ready += G->ecIn[parents[i].idx];
                    //printf("\t\t(Comm for %d [%d] -> %d [%d] done between %d and %d)\n", (int) dad, (int) mapping[dad], (int) node, (int) proc, (int) (ready - G->ecIn[parents[i].idx]), (int) ready);
                    end_send_tmp[mapping[dad]] = ready;
                    end_receive_tmp[proc] = ready;
                }
                begin_time = begin_time < ready ? ready : begin_time;
            }
            //printf("\t(Begin time of node %d on proc %d : %d)\n", (int) node, (int) proc, (int) begin_time);
            //Let update best_proc
            if (best_begin_time > begin_time){
                best_begin_time = begin_time;
                best_proc = proc;
            }
        }
        //We put node on best_proc
        mapping[node] = best_proc;
        begin_time = end_comp[best_proc];
        for (i = 0; i < indegree; i++){
            dad = G->in[parents[i].idx];
            if (mapping[dad] == best_proc)
                ready = finish_time[dad];
            else{
                ready = finish_time[dad] < end_send[mapping[dad]] ? end_send[mapping[dad]] : finish_time[dad];
                ready = ready < end_receive[best_proc] ? end_receive[best_proc] : ready;
                ready += G->ecIn[parents[i].idx];
                end_send[mapping[dad]] = ready;
                end_receive[best_proc] = ready;
                //printf("Comm for %d [%d] -> %d [%d] done between %d and %d\n", (int) dad, (int) mapping[dad], (int) node, (int) best_proc, (int) (ready - G->ecIn[parents[i].idx]), (int) ready);
            }
            begin_time = begin_time < ready ? ready : begin_time;
        }
        end_comp[best_proc] = begin_time + (ecType) G->vw[node];
        //printf("Node %d on proc %d (%d -> %d)\n", (int) node, (int) best_proc, (int) begin_time, (int) end_comp[best_proc]);
        finish_time[node] = end_comp[best_proc];
        makespan = makespan < finish_time[node] ? finish_time[node] : makespan;
        //We update heap
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++){
            child = G->out[i];
            incoming[child]--;
            if (incoming[child] == 0)
                heapInsert(heap, priority, &hsize, child, inheap);
        }
    }
    free(inheap);
    free(heap);
    free(priority);
    free(incoming);
    free(mapping);
    free(finish_time);
    free(end_send_tmp);
    free(end_send);
    free(end_receive_tmp);
    free(end_receive);
    free(end_comp);
    free(parents);
    return makespan;
}


ecType EFTPartruntime(dgraph* G, idxType nbproc, idxType* part, idxType nbpart, MLGP_option opt){
    //Initialize everything
    idxType i, node, proc, p, dad, child, best_proc, indegree;
    ecType begin_time, best_begin_time, ready, makespan = 0;
    parent *parents = (parent*) malloc((G->nVrtx +1) * sizeof(parent));
    ecType *end_comp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *finish_time = (ecType*) calloc((G->nVrtx +1), sizeof(ecType));
    idxType *mapping = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    idxType *incoming = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    idxType *mapping_part = (idxType*) malloc(nbpart * sizeof(idxType));
    for(i=1; i<=G->nVrtx; i++) {
        mapping[i] = -1;
        incoming[i] = G->inEnd[i] - G->inStart[i] + 1;
    }
    for (i = 0; i < nbpart; i++){
        mapping_part[i] = -1;
    }

    //Compute priorities
    ecType* priority = (ecType*) malloc((G->nVrtx +1) * sizeof(ecType));
    computePriorities(G, priority, opt);

    //Build heap of ready tasks
    idxType* heap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* inheap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType hsize = 0;
    for (i =0; i < G->nbsources; i++){
        hsize++;
        heap[hsize] = G->sources[i];
        inheap[G->sources[i]] = hsize;
    }
    heapBuild(heap, priority, hsize, inheap);

    //Phase 2 of EFT algorithm
    while(hsize > 0){
        node = heapExtractMax(heap, priority, &hsize, inheap);
        //Let sort parents according to their finish time
        indegree = 0;
        //printf("Considering node %d.\n", (int) node);
        for (i = G->inStart[node]; i <= G->inEnd[node]; i++) {
            dad = G->in[i];
            if (mapping[dad] < 0)
                u_errexit("Problem in EFTruntime (running %d before %d)\n", node, dad);
            parents[indegree].idx = i;
            parents[indegree].finish_time = finish_time[dad];
            //printf("Put in parents[%d]: idx %d finish_time %d\n", (int) indegree, (int) parents[indegree].idx, (int) parents[indegree].finish_time);
            indegree++;
        }
        sortParents(parents, indegree);

        if (mapping_part[part[node]] != -1){
            //Some nodes of this part has already been mapped
            //We map the node on the same processor
            best_proc = mapping_part[part[node]];
            //printf("Part %d already reserved for proc %d\n", part[node], best_proc);
        } else {
            //Else Let find the best proc for node
            best_begin_time = ecType_MAX;
            for (proc = 0; proc < nbproc; proc++) {
                //Let compute begin_time if we put node on proc
                begin_time = end_comp[proc];
                for (p = 0; p < nbproc; p++) {
                    end_receive_tmp[p] = end_receive[p];
                    end_send_tmp[p] = end_send[p];
                }
                for (i = 0; i < indegree; i++) {
                    dad = G->in[parents[i].idx];
                    if (mapping[dad] == proc)
                        ready = finish_time[dad];
                    else {
                        ready = finish_time[dad] < end_send_tmp[mapping[dad]] ? end_send_tmp[mapping[dad]]
                                                                              : finish_time[dad];
                        ready = ready < end_receive_tmp[proc] ? end_receive_tmp[proc] : ready;
                        ready += G->ecIn[parents[i].idx];
                        //printf("\t\t(Comm for %d [%d] -> %d [%d] done between %d and %d)\n", (int) dad, (int) mapping[dad], (int) node, (int) proc, (int) (ready - G->ecIn[parents[i].idx]), (int) ready);
                        end_send_tmp[mapping[dad]] = ready;
                        end_receive_tmp[proc] = ready;
                    }
                    begin_time = begin_time < ready ? ready : begin_time;
                }
                //printf("\t(Begin time of node %d on proc %d : %d)\n", (int) node, (int) proc, (int) begin_time);
                //Let update best_proc
                if (best_begin_time > begin_time) {
                    best_begin_time = begin_time;
                    best_proc = proc;
                }
            }
        }
        //We put node on best_proc
        mapping[node] = best_proc;
        mapping_part[part[node]] = best_proc;
        begin_time = end_comp[best_proc];
        for (i = 0; i < indegree; i++){
            dad = G->in[parents[i].idx];
            if (mapping[dad] == best_proc)
                ready = finish_time[dad];
            else{
                ready = finish_time[dad] < end_send[mapping[dad]] ? end_send[mapping[dad]] : finish_time[dad];
                ready = ready < end_receive[best_proc] ? end_receive[best_proc] : ready;
                ready += G->ecIn[parents[i].idx];
                end_send[mapping[dad]] = ready;
                end_receive[best_proc] = ready;
                //printf("Comm for %d [%d] -> %d [%d] done between %d and %d\n", (int) dad, (int) mapping[dad], (int) node, (int) best_proc, (int) (ready - G->ecIn[parents[i].idx]), (int) ready);
            }
            begin_time = begin_time < ready ? ready : begin_time;
        }
        end_comp[best_proc] = begin_time + (ecType) G->vw[node];
        //printf("Node %d on proc %d (%d -> %d)\n", (int) node, (int) best_proc, (int) begin_time, (int) end_comp[best_proc]);
        finish_time[node] = end_comp[best_proc];
        makespan = makespan < finish_time[node] ? finish_time[node] : makespan;
        //We update heap
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++){
            child = G->out[i];
            incoming[child]--;
            if (incoming[child] == 0)
                heapInsert(heap, priority, &hsize, child, inheap);
        }
    }
    free(inheap);
    free(heap);
    free(priority);
    free(mapping_part);
    free(incoming);
    free(mapping);
    free(finish_time);
    free(end_send_tmp);
    free(end_send);
    free(end_receive_tmp);
    free(end_receive);
    free(end_comp);
    free(parents);
    return makespan;
}

ecType EFTPartruntimeBusy(dgraph* G, idxType nbproc, idxType* part, idxType nbpart, MLGP_option opt){
    //Initialize everything
    idxType i, node, proc, p, dad, child, best_proc, indegree, all_busy;
    ecType begin_time, best_begin_time, ready, makespan = 0;
    parent *parents = (parent*) malloc((G->nVrtx +1) * sizeof(parent));
    ecType *end_comp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *finish_time = (ecType*) calloc((G->nVrtx +1), sizeof(ecType));
    idxType *mapping = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    idxType *incoming = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    idxType *partsize = (idxType*) calloc(nbpart, sizeof(idxType));
    idxType *mapping_part = (idxType*) malloc(nbpart * sizeof(idxType));
    idxType *is_busy = (idxType*) calloc(nbproc, sizeof(idxType));
    for(i=1; i<=G->nVrtx; i++) {
        mapping[i] = -1;
        incoming[i] = G->inEnd[i] - G->inStart[i] + 1;
        partsize[part[i]]++;
    }
    for (i = 0; i < nbpart; i++){
        mapping_part[i] = -1;
    }

    //Compute priorities
    ecType* priority = (ecType*) malloc((G->nVrtx +1) * sizeof(ecType));
    computePriorities(G, priority, opt);

    //Build heap of ready tasks
    idxType* heap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* inheap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType hsize = 0;
    for (i =0; i < G->nbsources; i++){
        hsize++;
        heap[hsize] = G->sources[i];
        inheap[G->sources[i]] = hsize;
    }
    heapBuild(heap, priority, hsize, inheap);

    //Phase 2 of EFT algorithm
    while(hsize > 0){
        node = heapExtractMax(heap, priority, &hsize, inheap);
        //Let sort parents according to their finish time
        indegree = 0;
        //printf("Considering node %d.\n", (int) node);
        for (i = G->inStart[node]; i <= G->inEnd[node]; i++) {
            dad = G->in[i];
            if (mapping[dad] < 0)
                u_errexit("Problem in EFTruntime (running %d before %d)\n", node, dad);
            parents[indegree].idx = i;
            parents[indegree].finish_time = finish_time[dad];
            //printf("Put in parents[%d]: idx %d finish_time %d\n", (int) indegree, (int) parents[indegree].idx, (int) parents[indegree].finish_time);
            indegree++;
        }
        sortParents(parents, indegree);

        if (mapping_part[part[node]] != -1){
            //Some nodes of this part has already been mapped
            //We map the node on the same processor
            best_proc = mapping_part[part[node]];
            //printf("Part %d already reserved for proc %d\n", part[node], best_proc);
        } else {
            //Else Let find the best proc for node
            best_begin_time = ecType_MAX;
            all_busy = 1;
            for (proc = 0; proc < nbproc; proc++)
                if (is_busy[proc] == 0)
                    all_busy = 0;
            for (proc = 0; proc < nbproc; proc++) {
                if ((is_busy[proc] > 0)&&(all_busy == 0))
                    continue;
                //Let compute begin_time if we put node on proc
                begin_time = end_comp[proc];
                for (p = 0; p < nbproc; p++) {
                    end_receive_tmp[p] = end_receive[p];
                    end_send_tmp[p] = end_send[p];
                }
                for (i = 0; i < indegree; i++) {
                    dad = G->in[parents[i].idx];
                    if (mapping[dad] == proc)
                        ready = finish_time[dad];
                    else {
                        ready = finish_time[dad] < end_send_tmp[mapping[dad]] ? end_send_tmp[mapping[dad]]
                                                                              : finish_time[dad];
                        ready = ready < end_receive_tmp[proc] ? end_receive_tmp[proc] : ready;
                        ready += G->ecIn[parents[i].idx];
                        //printf("\t\t(Comm for %d [%d] -> %d [%d] done between %d and %d)\n", (int) dad, (int) mapping[dad], (int) node, (int) proc, (int) (ready - G->ecIn[parents[i].idx]), (int) ready);
                        end_send_tmp[mapping[dad]] = ready;
                        end_receive_tmp[proc] = ready;
                    }
                    begin_time = begin_time < ready ? ready : begin_time;
                }
                //printf("\t(Begin time of node %d on proc %d : %d)\n", (int) node, (int) proc, (int) begin_time);
                //Let update best_proc
                if (best_begin_time > begin_time) {
                    best_begin_time = begin_time;
                    best_proc = proc;
                }
            }
        }
        //We put node on best_proc
        mapping[node] = best_proc;
        if (mapping_part[part[node]] == -1)
            is_busy[best_proc] += partsize[part[node]];
        is_busy[best_proc]--;
        mapping_part[part[node]] = best_proc;
        begin_time = end_comp[best_proc];
        for (i = 0; i < indegree; i++){
            dad = G->in[parents[i].idx];
            if (mapping[dad] == best_proc)
                ready = finish_time[dad];
            else{
                ready = finish_time[dad] < end_send[mapping[dad]] ? end_send[mapping[dad]] : finish_time[dad];
                ready = ready < end_receive[best_proc] ? end_receive[best_proc] : ready;
                ready += G->ecIn[parents[i].idx];
                end_send[mapping[dad]] = ready;
                end_receive[best_proc] = ready;
                //printf("Comm for %d [%d] -> %d [%d] done between %d and %d\n", (int) dad, (int) mapping[dad], (int) node, (int) best_proc, (int) (ready - G->ecIn[parents[i].idx]), (int) ready);
            }
            begin_time = begin_time < ready ? ready : begin_time;
        }
        end_comp[best_proc] = begin_time + (ecType) G->vw[node];
        //printf("Node %d on proc %d (%d -> %d)\n", (int) node, (int) best_proc, (int) begin_time, (int) end_comp[best_proc]);
        finish_time[node] = end_comp[best_proc];
        makespan = makespan < finish_time[node] ? finish_time[node] : makespan;
        //We update heap
        for (i = G->outStart[node]; i <= G->outEnd[node]; i++){
            child = G->out[i];
            incoming[child]--;
            if (incoming[child] == 0)
                heapInsert(heap, priority, &hsize, child, inheap);
        }
    }
    free(inheap);
    free(heap);
    free(priority);
    free(is_busy);
    free(mapping_part);
    free(partsize);
    free(incoming);
    free(mapping);
    free(finish_time);
    free(end_send_tmp);
    free(end_send);
    free(end_receive_tmp);
    free(end_receive);
    free(end_comp);
    free(parents);
    return makespan;
}


ecType EFTPartruntimeMacro(dgraph* G, idxType nbproc, idxType* part, idxType nbpart, MLGP_option opt){
    //Initialize everything
    idxType i, j, node, proc, p, dad, child, best_proc, all_busy, hsize_part = 0, current_part;
    ecType begin_time, best_end_comp, ready, makespan = 0;
    parent **parents = (parent**) malloc((G->nVrtx +1) * sizeof(parent*));
    idxType *nbparents = (idxType*) calloc(G->nVrtx +1, sizeof(idxType));
    ecType *end_comp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_comp_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_receive_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *end_send_tmp = (ecType*) calloc(nbproc, sizeof(ecType));
    ecType *finish_time = (ecType*) calloc((G->nVrtx +1), sizeof(ecType));
    idxType *mapping = (idxType*) malloc((G->nVrtx +1) * sizeof(idxType));
    idxType *incoming_part = (idxType*) calloc(nbpart, sizeof(idxType));
    idxType* heap_part = (idxType*) malloc(sizeof(idxType)*(nbpart));
    idxType* inheap_part = (idxType*) calloc(nbpart, sizeof(idxType));
    idxType *partsize = (idxType*) calloc(nbpart, sizeof(idxType));
    idxType *toporder = (idxType*) malloc((G->nVrtx+1) * sizeof(idxType));
    idxType **part_list = (idxType**) malloc(nbpart * sizeof(idxType*));
    idxType *part_list_size = (idxType*) calloc(nbpart, sizeof(idxType*));
    for(i=1; i<=G->nVrtx; i++) {
        mapping[i] = -1;
        partsize[part[i]]++;
        parents[i] = (parent*) malloc((G->inEnd[i] - G->inStart[i] + 2) * sizeof(parent));
        for (j=G->outStart[i]; j <= G->outEnd[i]; j++)
            if (part[i] != part[G->out[j]])
                incoming_part[part[G->out[j]]]++;
    }
    for (i = 0; i < nbpart; i++){
        part_list[i] = (idxType*) malloc((partsize[i]) * sizeof(idxType));
        if (incoming_part[i] == 0){
            hsize_part++;
            heap_part[hsize_part] = i;
            inheap_part[i] = hsize_part;
        }
    }

    //Compute priorities
    ecType* priority = (ecType*) malloc((G->nVrtx +1) * sizeof(ecType));
    computePriorities(G, priority, opt);

    //Compute part related stuff
    topsortPriorities(G, toporder, priority);
    for (i=1; i <= G->nVrtx; i++)
        part_list[part[i]][part_list_size[part[i]]++] = i;
    ecType* priority_parts = (ecType*) malloc((nbpart) * sizeof(ecType));
    for (i = 0; i < nbpart; i++) {
        if (partsize[i] > 0)
            priority_parts[i] = priority[part_list[i][0]];
        else
            priority_parts[i] = 0;
    }
    heapBuild(heap_part, priority_parts, hsize_part, inheap_part);

    //Phase 2 of EFT algorithm
    while(hsize_part > 0){
        current_part = heapExtractMax(heap_part, priority_parts, &hsize_part, inheap_part);
        //printf("Let's map part %d (%d ready)\n", current_part, hsize_part+1);
        //Let sort parents according to their finish time
        for (i = 0; i < part_list_size[current_part]; i++){
            node = part_list[current_part][i];
            nbparents[node] = 0;
            for (j = G->inStart[node]; j <= G->inEnd[node]; j++) {
                dad = G->in[j];
                if (part[dad] == current_part)
                    continue;
                if (mapping[dad] < 0)
                    u_errexit("Problem in EFTruntime (running %d before %d)\n", node, dad);
                parents[node][nbparents[node]].idx = j;
                parents[node][nbparents[node]].finish_time = finish_time[dad];
                nbparents[node]++;
            }
            sortParents(parents[node], nbparents[node]);
        }

        //Let find the best proc for part
        best_end_comp = ecType_MAX;
        for (proc = 0; proc < nbproc; proc++) {
            //Let compute ending_time if we put part on proc
            for (p = 0; p < nbproc; p++) {
                end_comp_tmp[proc] = end_comp[proc];
                end_receive_tmp[p] = end_receive[p];
                end_send_tmp[p] = end_send[p];
            }
            for (i = 0; i < part_list_size[current_part]; i++){
                node = part_list[current_part][i];
                begin_time = end_comp_tmp[proc];
                for (j = 0; j < nbparents[node]; j++) {
                    dad = G->in[parents[node][j].idx];
                    if (mapping[dad] == proc)
                        ready = finish_time[dad];
                    else {
                        ready = finish_time[dad] < end_send_tmp[mapping[dad]] ? end_send_tmp[mapping[dad]]
                                                                              : finish_time[dad];
                        ready = ready < end_receive_tmp[proc] ? end_receive_tmp[proc] : ready;
                        ready += G->ecIn[parents[node][j].idx];
                        end_send_tmp[mapping[dad]] = ready;
                        end_receive_tmp[proc] = ready;
                    }
                    begin_time = begin_time < ready ? ready : begin_time;
                }
                end_comp_tmp[proc] = begin_time + (ecType) G->vw[node];
            }
            if (best_end_comp > end_comp_tmp[proc]) {
                best_end_comp = end_comp_tmp[proc];
                best_proc = proc;
            }
        }

        //We put part on best_proc
        for (i = 0; i < part_list_size[current_part]; i++){
            node = part_list[current_part][i];
            mapping[node] = best_proc;
            begin_time = end_comp[best_proc];
            for (j = 0; j < nbparents[node]; j++) {
                dad = G->in[parents[node][j].idx];
                if (mapping[dad] == best_proc)
                    ready = finish_time[dad];
                else {
                    ready = finish_time[dad] < end_send[mapping[dad]] ? end_send[mapping[dad]]
                                                                          : finish_time[dad];
                    ready = ready < end_receive[best_proc] ? end_receive[best_proc] : ready;
                    ready += G->ecIn[parents[node][j].idx];
                    end_send[mapping[dad]] = ready;
                    end_receive[best_proc] = ready;
                }
                begin_time = begin_time < ready ? ready : begin_time;
            }
            end_comp[best_proc] = begin_time + (ecType) G->vw[node];
            finish_time[node] = end_comp[best_proc];
            makespan = makespan < finish_time[node] ? finish_time[node] : makespan;
            for (j=G->outStart[node]; j <= G->outEnd[node]; j++)
                if (part[node] != part[G->out[j]]) {
                    incoming_part[part[G->out[j]]]--;
                    if (incoming_part[part[G->out[j]]] == 0)
                        heapInsert(heap_part, priority_parts, &hsize_part, part[G->out[j]], inheap_part);
                }
        }
        //printf("Node %d on proc %d (%d -> %d)\n", (int) node, (int) best_proc, (int) begin_time, (int) end_comp[best_proc]);
        //We update heap

    }
    free(priority_parts);
    free(priority);
    for (i = 0; i < nbpart; i++)
        free(part_list[i]);
    free(part_list);
    free(part_list_size);
    free(toporder);
    free(partsize);
    free(inheap_part);
    free(heap_part);
    free(incoming_part);
    free(mapping);
    free(finish_time);
    free(end_send_tmp);
    free(end_send);
    free(end_receive_tmp);
    free(end_receive);
    free(end_comp_tmp);
    free(end_comp);
    free(nbparents);
    for(i=1; i<=G->nVrtx; i++)
        free(parents[i]);
    free(parents);
    return makespan;
}