#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"
#include "cache.h"
#include "dgraphTraversal.h"

int cmpfunc (const void * a, const void * b) {
    return ( *(int*)b - *(int*)a );
}

void sortNextUsage(idxType* table, idxType size){
    idxType i;
    qsort(table, size, sizeof(idxType), cmpfunc);
}

idxType cacheMissLUFO(dgraph* G, idxType* traversal, idxType cache_limit){
    /*Compute the number of cache misses for the traversal of G
     * with a given cache_size
     * for a last use first out strategy
     */

    //Warning!!! Here each task produce one unitary data
    //Otherwise we don't know how to handle multiple out edges
  idxType i, j, k, datasize, tmp, dad;
    /*printf("In cache miss computation: ");
    for (k=1; k <= G->nVrtx; k++)
        printf("%d ", traversal[k]);
    printf("\n");*/
    idxType cache_size = 0;
    idxType cache_miss = 0;
    idxType* nbout = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    int* marked = (int*) calloc(G->nVrtx+1, sizeof(int));
    idxType** nextusage = (idxType**) malloc(sizeof(idxType*)*(G->nVrtx+1));
    idxType* traversalidx = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* cache = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* incache = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType nbcache = 0;
    ecType* key = (ecType*) calloc(G->nVrtx+1, sizeof(ecType));
    for (i=1; i<=G->nVrtx; i++)
        traversalidx[traversal[i]] = i;
    for (i=1; i<=G->nVrtx; i++) {
        nbout[i] = G->outEnd[i] - G->outStart[i] + 1;
        nextusage[i] = (idxType*) malloc(sizeof(idxType)*(nbout[i]+2));
        for (j = G->outStart[i]; j <= G->outEnd[i]; j++)
            nextusage[i][j-G->outStart[i]] = traversalidx[G->out[j]];
        sortNextUsage(nextusage[i], nbout[i]);
        if (nbout[i] > 0)
            key[i] = nextusage[i][nbout[i]-1];
        else
            key[i] = G->nVrtx+1;
    }
    for (i=1; i<=G->nVrtx; i++) {
        idxType node = traversal[i];
        for (j = G->inStart[node]; j <= G->inEnd[node]; j++) {
            dad = G->in[j];
            if (marked[dad] == 0)
                u_errexit("In cacheMissLUFO, traversal is not a topological order (executing %d and %d not done)\n", node, dad);
            if (incache[dad] == 0) {
                datasize = 1; //here unitary
                /*printf("We have a cache miss executing %d (%d not in cache)\n",node,dad);
                printf("Cache has: ");
                for (k=1; k <= G->nVrtx; k++)
                    if (incache[k])
                        printf("%d ", k);
                printf("\n");*/
                while (datasize + cache_size > cache_limit){
                    tmp = heapExtractMax(cache, key, &nbcache, incache);
                    //printf("\tSo we discard %d\n",tmp);
                    cache_size -= 1; //here unitary
                }
                //printf("Insert dad %d (incahe %d)\n", dad, incache[dad]);
                heapInsert(cache, key, &nbcache, dad, incache);
                cache_size += datasize;
                cache_miss += datasize;
            }
        }
        marked[node] = 1;
        datasize = 1; //here unitary
        while (datasize + cache_size > cache_limit){
            tmp = heapExtractMax(cache, key, &nbcache, incache);
            //printf("Output of %d exeed limit (datasize %d cache_size %d) so discard %d\n",node, datasize, cache_size, tmp);
            cache_size -= 1; //here unitary
        }
        //printf("Insert output %d (incahe %d)\n", node, incache[node]);
        heapInsert(cache, key, &nbcache, node, incache);
        cache_size += datasize;
        //printf("Done executing %d\n",node);
        for (j = G->inStart[node]; j <= G->inEnd[node]; j++) {
            dad = G->in[j];
            nbout[dad]--;
            if ((nbout[dad] == 0)&&(incache[dad] != 0)) {
                //printf("We delete %d from cache\n", dad);
                heapDelete(cache, key, &nbcache, dad, incache);
                cache_size -= 1; //here unitary
            }
            if ((nbout[dad] != 0)&&(incache[dad] != 0))
                heapUpdateKey(cache, incache, key, nbcache, dad, nextusage[dad][nbout[dad] - 1]);
            if ((nbout[dad] != 0)&&(incache[dad] == 0))
                key[dad] = nextusage[dad][nbout[dad]-1];
        }
        /*printf("Cache has: ");
        for (k=1; k <= G->nVrtx; k++)
            if (incache[k])
                printf("%d ", k);
        printf("\n");*/
    }
    free(nbout);
    free(marked);
    for (i=1; i<=G->nVrtx; i++)
        free(nextusage[i]);
    free(nextusage);
    free(traversalidx);
    free(cache);
    free(incache);
    free(key);
    return cache_miss;
}

idxType cacheMiss(dgraph* G, MLGP_option opt) {
  idxType *toporder = (idxType*) malloc((G->nVrtx + 2) * sizeof(idxType));
    idxType i, cachemiss;
    switch (opt.live_traversal){
        case LIVE_TRAV_NAT:
            for (i=1; i <= G->nVrtx; i++)
                toporder[i] = i;
            break;

        case LIVE_TRAV_DFS:
            DFStopsort(G, toporder);
            break;

        case LIVE_TRAV_BFS:
            BFStopsort(G, toporder);
            break;

        case LIVE_TRAV_MIX:
            mixtopsort(G, toporder, 50, 1);
            break;

        case LIVE_TRAV_RAND:
            randTopsort(G, toporder);
            break;

        default:
            u_errexit("Live_traversal %d is not recognized!\n", opt.live_traversal);
    }

    cachemiss = cacheMissLUFO(G, toporder, opt.maxliveset);

    free(toporder);
    return cachemiss;
}

idxType cacheMiss_with_part(dgraph *G, idxType* part, int nbpart, MLGP_option opt){
  idxType *toporder = (idxType*) malloc((G->nVrtx + 2) * sizeof(idxType));
    idxType i, cachemiss;
    idxType *toporderpart;
    switch (opt.live_traversal){
        case LIVE_TRAV_NAT:
            toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
            topSortOnParts(G, part, toporderpart, nbpart);
            natural_with_part(G, toporder, part, toporderpart, nbpart);
            free(toporderpart);
            break;

        case LIVE_TRAV_DFS:
            DFStopsort_with_part(G, part, nbpart, toporder);
            break;

        case LIVE_TRAV_BFS:
            BFStopsort_with_part(G, part, nbpart, toporder);
            break;

        case LIVE_TRAV_MIX:
            //TODO mixtopsort_with_part
            DFStopsort_with_part(G, part, nbpart, toporder);
            break;

        case LIVE_TRAV_RAND:
            randBFStopsort_with_part(G, part, nbpart, toporder);
            break;

        default:
            u_errexit("Live_traversal %d is not recognized!\n", opt.live_traversal);
    }

    /*printf("toporder: ");
    for (i = 1; i<= G->nVrtx; i++)
        printf("%d ", toporder[i]);
    printf("\n");*/

    cachemiss = cacheMissLUFO(G, toporder, opt.maxliveset);

    free(toporder);
    return cachemiss;

}
