#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/stat.h>

#include <getopt.h>

#include "dgraph.h"
#include "dgraphReader.h"
#include "cache.h"
#include "scheduling.h"
#include "utils.h"
#include "rvcyclelive.h"
#include "dgraphPartition.h"


void printUsage_rMLGPlive(char *exeName)
{
    printMLGPusage(exeName);

    //live_cut
    printf("\t--live_cut int \t(default = 0)\n");
    printf("\t\tDetermine the cuting strategy for rMLGPlive\n");
    printf("\t\t= 0 : recursively cut in half until the live set value is smaller than opt.maxliveset\n");
    printf("\t\t= 1 : cut in half and recut in 2/3 1/3 if liveset smaller than opt.maxliveset / 2\n");

    //live_cut_merge
    printf("\t--live_cut_merge int \t(default = 1)\n");
    printf("\t\tDetermine if we merge parts if the max live set is not violated after the recursive cut\n");
    printf("\t\t= 0 : no\n");
    printf("\t\t= 1 : yes\n");
}


int processArgs_rMLGPlive(int argc, char **argv, MLGP_option* opt)
{
    if (argc < 3) {
	    printUsage_rMLGPlive(argv[0]);
	    u_errexit("%s: There is a problem with the arguments.\n", argv[0]);
    }

    opt->maxliveset = atoi(argv[2]);
    initMLGPoptions(opt, atoi(argv[2]));
    processMLGPargs(argc, argv, opt);

    return 0;
}


void run_rMLGPlive(char* file_name, MLGP_option opt)
{
    dgraph G;
    opt.ratio = 1.1;

    readDGraph(&G, file_name, opt.use_binary_input);
    set_dgraph_info(&G);
    idxType percentage = opt.maxliveset;

    int maxindegree, minindegree, maxoutdegree, minoutdegree;
    double aveindegree, aveoutdegree;
    dgraph_info(&G, &maxindegree, &minindegree, &aveindegree, &maxoutdegree, &minoutdegree, &aveoutdegree);
    G.maxindegree = maxindegree;
    G.maxoutdegree = maxoutdegree;

    printf("%d %d %d %d\n%lf %lf\n", maxindegree,minindegree, maxoutdegree,minoutdegree,aveindegree,aveoutdegree);

    //opt.maxliveset=400;

    printf("maxliveset: %d debug:%d\n", opt.maxliveset,opt.debug);
    printf("Nb node: %d\nNb Edges: %d\n\n", G.nVrtx, G.nEdge);

    ecType* edgecut = (ecType*) malloc(sizeof(ecType) * opt.runs);
    int* nbcomm = (int*) malloc(sizeof(int) * opt.runs);
    double* latencies = (double*) malloc(sizeof(double) * opt.runs);
    idxType* traversal = (idxType*) malloc(sizeof(idxType)*(G.nVrtx + 2));

    int r,i;
    rcoarsen* rcoars;
    int dumpnbpartMain=-1;
    idxType cachemiss;
    idxType ls;

    for (r = 0; r<opt.runs; r++){
        rMLGP_info* info = (rMLGP_info*)  malloc (sizeof (rMLGP_info));
        initRInfoPart(info);

        printf("########################## RUN %d ########################\n", r);
    	usRandom(opt.seed + r);
        if(opt.debug > -1){
	        ls = liveSize(&G, opt);
	        printf("Initial Graph LiveSize: %d\n", ls);
            opt.maxliveset = (percentage * ls) / 100;

            printf("New maxliveset: %d\n", opt.maxliveset);

            cachemiss = cacheMiss(&G, opt);
            printf("Initial Graph cache misses: %d\n", cachemiss);
        }

        rcoars = rVCycleLive(&G, opt, info, &dumpnbpartMain);
        //printf("Number of partitions: %d\n", dumpnbpartMain);

        if(opt.debug > 3)
            printf("dumpnbpartMain:%d\n",dumpnbpartMain );
    	edgecut[r] = edgeCut(&G, rcoars->coars->part);
    	latencies[r] = computeLatency(&G, rcoars->coars->part, 1.0, 11.0);
    	printf("Final run %d\t%d\t%d\t%d\t%d\t%f\n", r, G.nVrtx, G.nEdge, (int) edgecut[r], (int) nbcomm[r], latencies[r]);
        printPartWeights(rcoars->coars->graph, rcoars->coars->part);
    	opt.seed++;
        /*
        char a[20]="rectpart.dot";
        dgraph_to_dot(&G,rcoars->coars->part,a);
        */
        int* partsize = (int*) malloc(sizeof(int) * dumpnbpartMain);
        int nn = nbPart(&G, rcoars->coars->part, partsize);
        printf("Number of partitions: %d\n", nn);
        printf("Edgecut: %d\n", (int) edgecut[r]);
        printf("Partitions size: ");
        for (i = 0; i<nn; i++)
            printf("%d ", partsize[i]);
        printf("\n");
        ls = liveSize_with_part(&G, rcoars->coars->part, nn, opt);
        printf("Partitioned Graph LiveSize: %d\n", ls);

        cachemiss = cacheMiss_with_part(&G, rcoars->coars->part, nn, opt);
        printf("Partitioned Graph cache misses: %d\n", cachemiss);

        if(opt.debug > 0) {
            DFStopsort(&G, traversal);
	        printf("Done topsort\n");
            ls = calcLiveSizeMix_with_part_Multi(&G, rcoars->coars->part, nn, traversal, 10, opt);
            printf("Partitioned LiveSize: %d\n", ls);
        }

        if(opt.debug > 0) {
            printf("==================\n");
            printf("Nbpart = %d\nPartsize: ", nn);
            int k;
            for (k = 0; k<nn; k++)
                printf("%d ", partsize[k]);
            printf("\n");
            printf("==================\n");

            printf("==================\n");
            printf("Nbpart = %d\nPartLiveSet: ", nn);
            for (k = 0; k<nn; k++)
                printf("%d ", calcLiveSizePartMixMulti(&G, rcoars->coars->part, k, 10));
            printf("\n");
            printf("==================\n");

        }

    	printf("######################## END RUN %d ########################\n\n", r);
    	FILE *fin=fopen("partassign.txt","w");
        for (i=1; i<=rcoars->coars->graph->nVrtx; i++){
            fprintf(fin,"part[%d] %d\n",i, rcoars->coars->part[i]);
        }
        fclose(fin);

        free(partsize);

        freeRInfoPart(info);
    }
    double edgecutave = 0.0, nbcommave = 0.0, edgecutsd = 0.0, nbcommsd = 0.0;
    for (r = 0; r<opt.runs; r++){
    	edgecutave += edgecut[r];
    	nbcommave += nbcomm[r];
    }
    edgecutave = edgecutave / opt.runs;
    nbcommave = nbcommave / opt.runs;
    for (r = 0; r<opt.runs; r++){
    	edgecutsd = edgecut[r] - edgecutave < 0? edgecutave - edgecut[r] : edgecut[r] - edgecutave;
    	nbcommsd = nbcomm[r] - nbcommave < 0? nbcommave - nbcomm[r] : nbcomm[r] - nbcommave;
    }
    edgecutsd = edgecutsd / opt.runs;
    nbcommsd = nbcommsd / opt.runs;
    printf("Final runs\t%d\t%d\t%f\t%f\t%f\t%f\n", G.nVrtx, G.nEdge, edgecutave, edgecutsd, nbcommave, nbcommsd);

    /*
    FILE *file;
    char name_tmp[200];
    sprintf(name_tmp,".rMLGP.part.%d.seed.%d", nbpart, opt.seed);
    char res_name[200];
    strcpy(res_name, file_name);
    strcat(res_name, name_tmp);
    file = fopen(res_name, "w");
    idxType i;
    for (i=1; i<=G.nVrtx; i++){
	fprintf(file, "%d\n", (int) rcoars->coars->part[i]);
    }
    fclose(file);
    */
    free(G.inStart);
    free(G.inEnd);
    free(G.in);
    free(G.outStart);
    free(G.outEnd);
    free(G.out);
    free(G.hollow);
    free(G.vw);
    free(G.ecIn);
    free(G.ecOut);
    free(edgecut);
    free(nbcomm);
    free(latencies);
    free(traversal);
    return;
}

int main(int argc, char *argv[])
{
    MLGP_option opt;

    processArgs_rMLGPlive(argc, argv, &opt);

    run_rMLGPlive(opt.file_name, opt);

    return 0;
}
