#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "rvcycle.h"
#include "rvcyclelive.h"
#include "vcycle.h"
#include "option.h"
#include "utils.h"
#include "debug.h"
#include "clustering.h"
#include "refinement.h"
#include "info.h"
#include "dgraph.h"
#include "scheduling.h"
#include "undirPartitioning.h"
#include "dgraphPartition.h"




void splitCoarsenLive(rcoarsen* rcoars, dgraph* G1, dgraph* G2, int* part1, int* part2)
{
    //printf("splitCoarsen\n");
    /*Build G1 and G2 and fill rcoars->next_index based on rcoars->coars->part*/
    coarsen* C = rcoars->coars;
    idxType* part = C->part;
    idxType nVrtx = C->graph->nVrtx;
    idxType* inStart = C->graph->inStart;
    idxType* inEnd = C->graph->inEnd;
    idxType* outStart = C->graph->outStart;
    idxType* outEnd = C->graph->outEnd;
    idxType* next_index = rcoars->next_index;
    dgraph* graph = C->graph;
    idxType *weightmarker1,*weightmarker2;
    weightmarker1 = (idxType*) calloc ((nVrtx + 2),sizeof(idxType));
    weightmarker2 = (idxType*) calloc ((nVrtx + 2),sizeof(idxType));
    int* both = (int*) calloc ((nVrtx + 2),sizeof(int));
    idxType nVrtx1 = 1, nVrtx2 = 1;
    idxType nEdge1 = 0, nEdge2 = 0;
    idxType i,j;
    int degree;
    int partin = -1, partout = -1;

    for (i=1; i <= nVrtx; i++) {
        for (j=inStart[i]; j<=inEnd[i]; j++) {
            if (part[i] != part[C->graph->in[j]]) {
                partin = part[C->graph->in[j]];
                partout = part[i];
                break;
            }
        }
        if (partin != -1)
            break;
    }
    if ((partin == -1)||(partout == -1)) {
        partin = 0;
        partout = 1;
    }
    *part1 = partin;
    *part2 = partout;

    //idxType hin1=0,hin2=0;
    //idxType hout1=0,hout2=0;
    /*Fill rcoars->next_index and compute nVrtx1, nVrtx2, nEdge1 and nEdge2*/
    for (i=1; i <= nVrtx; i++) {
        if ( part[i] == partin ) {
            for (j=inStart[i]; j<=inEnd[i]; j++) {
                if (part[graph->in[j]]==partout) {
                    both[i] = 1;
                    break;
                }
            }
            for (j=outStart[i]; j<=outEnd[i]; j++) {
                if (part[graph->out[j]]==partout) {
                    both[i] = 1;
                    break;
                }
            }
        }
        else {
            for (j=inStart[i]; j<=inEnd[i]; j++) {
                if (part[graph->in[j]]==partin) {
                    both[i]=1;
                    break;
                }
            }

            for (j=outStart[i]; j<=outEnd[i]; j++) {
                if (part[graph->out[j]]==partin) {
                    both[i]=1;
                    break;
                }
            }
        }
    }
    for (i=1; i <= nVrtx; i++) {
        if (both[i]==1) {
            if (part[i]==partin)
                next_index[i]=nVrtx1;
            else
                next_index[i]=nVrtx2;

            weightmarker1[i]=nVrtx1;
            weightmarker2[i]=nVrtx2;
            nVrtx1++;
            nVrtx2++;
        }
        else{
            if (part[i]==partin) {
                next_index[i]=nVrtx1;
                weightmarker1[i]=nVrtx1;
                nVrtx1++;
            }
            else{ //part[i]==1
                next_index[i]=nVrtx2;
                weightmarker2[i]=nVrtx2;
                nVrtx2++;
            }
        }
    }
    for (i=1; i <= nVrtx; i++) {
        for (j=inStart[i]; j<=inEnd[i]; j++) {

            if (weightmarker1[i]>0 && weightmarker1[graph->in[j]]>0)
                if (part[i]!=partout || part[graph->in[j]]!=partout)
                    nEdge1++;
            if (weightmarker2[i]>0 && weightmarker2[graph->in[j]]>0)
                if (part[i]!=partin || part[graph->in[j]]!=partin)
                    nEdge2++;

        }
    }
    nVrtx1--;
    nVrtx2--;
    //printf("nVrtx: %d n: %d %d e: %d %d\n",nVrtx, nVrtx1,nVrtx2,nEdge1,nEdge2 );
    //printf("hollow nodes in out1:: in out2:: %d %d %d %d\n",hin1,hout1,hin2,hout2);
    //printf("G1 G2 sizes: %d %d\n",nVrtx1,nVrtx2);
    //printf("Let's allocate everything, nVrtx1 = %d, nEdge1 = %d, nVrtx2 = %d, nEdge2 = %d\n", nVrtx1, nEdge1, nVrtx2, nEdge2);
    /*Allocate everything*/
    G1->inStart = (idxType*) malloc (sizeof(idxType) * (nVrtx1 + 2));
    G1->inEnd = (idxType*) malloc (sizeof(idxType) * (nVrtx1 + 2));
    G1->in = (idxType*) malloc (sizeof(idxType) * (nEdge1 + 2));
    G1->outStart = (idxType*) malloc (sizeof(idxType) * (nVrtx1 + 2));
    G1->hollow = (int*) calloc ((nVrtx1 + 2),sizeof(int));
    G1->outEnd = (idxType*) malloc (sizeof(idxType) * (nVrtx1 + 2));
    G1->out = (idxType*) malloc (sizeof(idxType) * (nEdge1 + 2));
    G1->vw = (idxType*) calloc ((nVrtx1 + 2),sizeof(idxType));
    G1->ecIn = (ecType*) malloc (sizeof(ecType) * (nEdge1 + 2));
    G1->ecOut = (ecType*) malloc (sizeof(ecType) * (nEdge1 + 2));
    G2->inStart = (idxType*) malloc (sizeof(idxType) * (nVrtx2 + 2));
    G2->inEnd = (idxType*) malloc (sizeof(idxType) * (nVrtx2 + 2));
    G2->in = (idxType*) malloc (sizeof(idxType) * (nEdge2 + 2));
    G2->outStart = (idxType*) malloc (sizeof(idxType) * (nVrtx2 + 2));
    G2->hollow = (int*) calloc ((nVrtx2 + 2), sizeof(int));
    G2->outEnd = (idxType*) malloc (sizeof(idxType) * (nVrtx2 + 2));
    G2->out = (idxType*) malloc (sizeof(idxType) * (nEdge2 + 2));
    G2->vw = (idxType*) calloc ((nVrtx2 + 2),sizeof(idxType));
    G2->ecIn = (ecType*) malloc (sizeof(ecType) * (nEdge2 + 2));
    G2->ecOut = (ecType*) malloc (sizeof(ecType) * (nEdge2 + 2));

    /*Fill inStart1, inStart2, inEnd1, inEnd2, in1 and in2*/

    idxType idx1 = 0, idx2 = 0;

    for (i=1; i <= nVrtx; i++) {
        if (both[i]) {
            if (part[i]==partin) {
                G2->hollow[weightmarker2[i]]=1;
            }
            else{
                G1->hollow[weightmarker1[i]]=1;
            }
        }
    }
    for (i=1; i <= nVrtx; i++) {
        if (weightmarker1[i]>0) { //part 0 or both
            G1->inEnd[weightmarker1[i]-1] = idx1-1;
            G1->inStart[weightmarker1[i]] = idx1;
            for (j=inStart[i]; j<=inEnd[i]; ++j) {
                if (weightmarker1[graph->in[j]]>0) {
                    if (part[i] != partout || part[graph->in[j]] != partout) {
                        G1->in[idx1] = weightmarker1[graph->in[j]];
                        if (part[i]==partin && part[graph->in[j]]==partin)
                            G1->ecIn[idx1]=graph->ecIn[j];
                        else
                            //G1->ecIn[idx1]=0;
                            G1->ecIn[idx1]=1;
                        ++idx1;
                    }
                }
            }
        }
    }
    for (i=1; i <= nVrtx; i++) {
        if (weightmarker2[i]>0) { //part 1 or both
            G2->inEnd[weightmarker2[i]-1] = idx2-1;
            G2->inStart[weightmarker2[i]] = idx2;
            for (j=inStart[i]; j<=inEnd[i]; ++j) {
                if (weightmarker2[graph->in[j]]>0) {
                    if (part[i]!=partin || part[graph->in[j]]!=partin) {
                        G2->in[idx2]=weightmarker2[graph->in[j]];
                        if (part[i]==partout && part[graph->in[j]]==partout)
                            G2->ecIn[idx2]=graph->ecIn[j];
                        else
                            //G2->ecIn[idx2]=0;
                            G2->ecIn[idx2]=1;
                        ++idx2;
                    }
                }
            }
        }
    }
    //printf("after loop %d %d %d %d\n",idx1,nEdge1,idx2,nEdge2);
    G1->inEnd[nVrtx1] = nEdge1-1;
    G1->inStart[nVrtx1+1] = nEdge1;
    G1->inEnd[nVrtx1+1] = nEdge1-1;
    G2->inEnd[nVrtx2] = nEdge2-1;
    G2->inStart[nVrtx2+1] = nEdge2;
    G2->inEnd[nVrtx2+1] = nEdge2-1;

    G1->maxindegree = 0;
    G2->maxindegree = 0;
    for (i=1; i <= nVrtx; i++) {
        if ((part[i] == partin)||(both[i] == 1)) {
            degree = G1->inEnd[weightmarker1[i]] - G1->inStart[weightmarker1[i]] + 1;
            G1->maxindegree = G1->maxindegree < degree ? degree : G1->maxindegree;
        }
        if ((part[i] == partout)||(both[i] == 1)) {
            degree = G2->inEnd[weightmarker2[i]] - G2->inStart[weightmarker2[i]] + 1;
            G2->maxindegree = G2->maxindegree < degree ? degree : G2->maxindegree;
        }
    }

    vwType G1weight = 0, G2weight = 0;
    /*Fill vw1 and vw2*/
    for (i=1; i<=graph->nVrtx; i++) {
        //if (weightmarker1[i] ==0 && weightmarker2[i]==0)
        //printf("big problem\n");
        if (part[i] == partin) {
            //printf("node %d is mapped to p0 node %d  -- p1 node %d\n",i,weightmarker1[i],weightmarker2[i] );
            G1->vw[weightmarker1[i]] = C->graph->vw[i];
            G1weight += G1->vw[weightmarker1[i]];
        }
        else if (part[i] == partout) {
            //printf("node %d is mapped to p0 node %d -- p1 node %d\n",i,weightmarker1[i],weightmarker2[i] );
            G2->vw[weightmarker2[i]] = C->graph->vw[i];
            G2weight += G2->vw[weightmarker2[i]];
        }
        else{
            //printf("wms: %d %d \n", weightmarker1[i],weightmarker2[i]);
            //printf("weight in G1 G2: %d %d\n",G1->vw[weightmarker1[i]],G2->vw[weightmarker2[i]]);
        }
    }
    //don't assign weights to hollow nodes, calloc makes them zero
    /*
    printf("FIRST IN\n");
    for (i=0;i<nEdge1;++i)
        printf("%d ", G1->in[i]);
    printf("\nFIRST\n");
    for (i=1;i<=nVrtx;++i)
        if (part[i]==0 || both[i])
            printf("i=%d,wm1[i]=%d, (%d)-(%d)\n",i,weightmarker1[i], G1->inStart[weightmarker1[i]],G1->inEnd[weightmarker1[i]]);
    printf("\nSECOND IN\n");
    for (i=0;i<nEdge2;++i)
        printf("%d ", G2->in[i]);
    printf("\nSECOND\n");
    for (i=1;i<=nVrtx;++i)
        if (part[i]==1 || both[i])
            printf("i=%d,wm2[i]=%d, (%d)-(%d)\n",i,weightmarker2[i], G2->inStart[weightmarker2[i]],G2->inEnd[weightmarker2[i]]);
    printf("\n");
    */
    /*Fill G1 and G2*/
    G1->frmt = graph->frmt;
    G1->nVrtx = nVrtx1;
    G1->nEdge = nEdge1;
    G1->totvw = G1weight;
    G2->frmt = graph->frmt;
    G2->nVrtx = nVrtx2;
    G2->nEdge = nEdge2;
    G2->totvw = G2weight;
    fillOutFromIn(G1);
    //printf("first done\n");
    fillOutFromIn(G2);

    G1->sources  = (idxType * ) malloc(sizeof(idxType) * (nVrtx1 + 1));
    G1->targets  = (idxType * ) malloc(sizeof(idxType) * (nVrtx1 + 1));
    G1->nbsources = 0;
    G1->nbtargets = 0;
    for (i=1; i<=nVrtx1; i++){
        if (G1->inEnd[i] < G1->inStart[i])
            G1->sources[G1->nbsources++] = i;
        if (G1->outEnd[i] < G1->outStart[i])
            G1->targets[G1->nbtargets++] = i;
    }
    G2->sources  = (idxType * ) malloc(sizeof(idxType) * (nVrtx2 + 1));
    G2->targets  = (idxType * ) malloc(sizeof(idxType) * (nVrtx2 + 1));
    G2->nbsources = 0;
    G2->nbtargets = 0;
    for (i=1; i<=nVrtx2; i++){
        if (G2->inEnd[i] < G2->inStart[i])
            G2->sources[G2->nbsources++] = i;
        if (G2->outEnd[i] < G2->outStart[i])
            G2->targets[G2->nbtargets++] = i;
    }

    //printf("second done\n");
    free(weightmarker2);
    free(weightmarker1);
    free(both);
}


rcoarsen* rVCycleLive(dgraph *G, MLGP_option opt, rMLGP_info* info, int *resultingNbPart)
{
    info->timing_global -= u_wseconds();
    info->depth++;
    //printf("in rVCycle first line\n");
    double ratio = opt.ratio;
    idxType nVrtx = G->nVrtx;
    idxType* traversal = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));

    coarsen* C;
    idxType i;
    if (opt.debug > 0) {
        //for (i=0; i<info->depth; i++)
        //    printf("@");
        printf("In rVcycle nVrtx = %d first line\n", nVrtx);
    }

    idxType ls = liveSize(G, opt);

    if (opt.debug > 0) {
        //for (i=0; i<info->depth; i++)
        //    printf("@");
        printf("Begin rVCycleLive liveSize: %d\n", ls);
    }

    if (opt.debug > 10)
        if (ls <= opt.maxliveset) {
            printf("ls=%d maxliveset=%d actual size: %d\n",ls,opt.maxliveset, G->nVrtx);
        }

    vwType Gweight = (vwType) G->totvw;

    if ((ls <= opt.maxliveset) || (Gweight <= opt.maxliveset)) {
        C = initializeCoarsen(G);
        for (i=1; i<=G->nVrtx; i++)
            C->part[i] = 0;
        C->nbpart = 1;

        *resultingNbPart = 1;
        info->depth--;
        info->timing_global += u_wseconds();
        info->timing_coars = 0;
        info->timing_inipart = 0;
        info->timing_uncoars = 0;
        return initializeRCoarsen(C);
    }

    opt.nbPart = 2;
    opt.ratio = 1.5;
    opt.ub[0] = (Gweight/2.) * opt.ratio;
    opt.ub[1] = (Gweight/2.) * opt.ratio;
    opt.lb[0] = ((2.0 - opt.ratio)/opt.ratio) * opt.ub[0];
    opt.lb[1] = ((2.0 - opt.ratio)/opt.ratio) * opt.ub[1];

    if (opt.debug > 0) {
        //for (i=0; i<info->depth; i++)
        //    printf("@");
        printf("In recursive totvw = %d,nVrtx = %d, nbpart = %d, lb = %f, ub = %f\n",Gweight, nVrtx, opt.nbPart, opt.lb[0], opt.ub[0]);
    }

    //wcycle
    idxType* flag = (idxType * ) calloc(G->nVrtx +3, sizeof(idxType));
    dgraph* igraph;
    dgraph Gsingle;
    if ((opt.wcycle > 0)||(opt.anchored > 0)) {
        wcycle(G, flag, opt, info->info);
        if (opt.wcycle_fix)
            printf("Edgecut Metis + fix : %d\n", (int) edgeCut(G, flag));
        else
            printf("Edgecut Metis : %d\n", (int) edgeCut(G, flag));
    }
    if (opt.anchored == 1) {
        copyDgraph(G, &Gsingle);
        addSingleSourceTarget(&Gsingle, flag);
    }
    if (opt.wcycle == 0)
        for (i = 0; i <= G->nVrtx + 2; i++)
            flag[i] = 0;

    if (opt.anchored == 1)
        igraph = &Gsingle;
    else
        igraph = G;

    //Now bisection
    C = VCycle2way(igraph, flag, opt, info->info);
    //printf("End vcycle in 2\n");

    rcoarsen *rcoars = initializeRCoarsen(C);

    //We build the two subgraphes based on the bisection
    int part1, part2;
    dgraph *G1 = (dgraph*) malloc(sizeof(dgraph));
    dgraph *G2 = (dgraph*) malloc(sizeof(dgraph));
    splitCoarsenLive(rcoars, G1, G2, &part1, &part2);
    //printf("Done spliting nVrtx1 %d nVrtx2 %d\n", G1->nVrtx, G2->nVrtx);

    idxType liveset1 = liveSize(G1, opt);
    idxType liveset2 = liveSize(G2, opt);
    int nbrecut = 0;

    //printf("G1 live set %d\n", calcLiveSizeMix(G1));
    //printf("G2 live set %d\n", calcLiveSizeMix(G2));

    switch (opt.live_cut) {
        case LIVE_CUT_RECUT:
            while ((((liveset1 < opt.maxliveset / 2)&&(liveset2 > opt.maxliveset))||((liveset2 < opt.maxliveset / 2)&&(liveset1 > opt.maxliveset)))&&(nbrecut < 1)) {
                if (liveset1 < opt.maxliveset / 2) {
                    opt.nbPart = 2;
                    opt.ub[0] = (2*Gweight/3.) * opt.ratio < 0.9 * Gweight ? (2*Gweight/3.) * opt.ratio : 0.9 * Gweight;
                    opt.ub[1] = (Gweight/3.) * opt.ratio;
                    opt.lb[0] = ((2.0 - opt.ratio)/opt.ratio) * opt.ub[0];
                    opt.lb[1] = ((2.0 - opt.ratio)/opt.ratio) * opt.ub[1];

                    //printf("Now first recursion\n");
                    C = VCycle2way(igraph, flag, opt, info->info);
                    //printf("End first recursion\n");
                    rcoars = initializeRCoarsen(C);
                    splitCoarsenLive(rcoars, G1, G2, &part1, &part2);
                    liveset1 = liveSize(G1, opt);
                    liveset2 = liveSize(G2, opt);
                }
                if (liveset2 < opt.maxliveset / 2) {
                    opt.nbPart = 2;
                    opt.ub[0] = (Gweight/3.) * opt.ratio;
                    opt.ub[1] = (2*Gweight/3.) * opt.ratio < 0.9 * Gweight ? (2*Gweight/3.) * opt.ratio : 0.9 * Gweight;
                    opt.lb[0] = ((2.0 - opt.ratio)/opt.ratio) * opt.ub[0];
                    opt.lb[1] = ((2.0 - opt.ratio)/opt.ratio) * opt.ub[1];
                    //printf("Now second recursion\n");

                    C = VCycle2way(igraph, flag, opt, info->info);
                    //printf("End second recursion\n");

                    rcoars = initializeRCoarsen(C);
                    splitCoarsenLive(rcoars, G1, G2, &part1, &part2);
                    liveset1 = liveSize(G1, opt);
                    liveset2 = liveSize(G2, opt);
                }
                nbrecut++;
            }
            break;
    }

    vwType G1weight = 0;
    for (i=1; i<=G1->nVrtx; i++)
        G1weight += G1->vw[i];
    vwType G2weight = 0;
    for (i=1; i<=G2->nVrtx; i++)
        G2weight += G2->vw[i];

    if (opt.debug) {
        for (i=0; i<info->depth; i++)
            printf("@");
        printf("In recursive G1 size = %d, G1 weight=%d, G2 size = %d, G2 weight = %d\n", G1->nVrtx,G1weight,G2->nVrtx,G2weight);
    }

    int dumpnbpart1=-1,dumpnbpart2=-1;

    //We compute recursive bisection for G1
    MLGP_option* opt1 = copyOpt(opt, -1);
    info->rec1 = (rMLGP_info*) malloc (sizeof (struct rMLGP_info));
    initRInfoPart(info->rec1);
    //printf("Now first recursion\n");
    rcoarsen *rcoars1 = rVCycleLive(G1, opt, info->rec1, &dumpnbpart1);
    //printf("End first recursion\n");
    free_opt(opt1);
    free(opt1);

    //printf("first subpart done! size:%d weight: %d\n", G1->nVrtx,G1->totvw);
    //printf("dump part 1: %d\n",dumpnbpart1);

    //We compute recursive bisection for G2
    MLGP_option* opt2 = copyOpt(opt, -1);
    info->rec2 = (rMLGP_info*) malloc (sizeof (struct rMLGP_info));
    initRInfoPart(info->rec2);
    //printf("Now second recursion\n");
    rcoarsen *rcoars2 = rVCycleLive(G2, opt, info->rec2, &dumpnbpart2);
    //printf("End second recursion\n");
    free_opt(opt2);
    free(opt2);

    if (G1->nVrtx == 0)
        dumpnbpart1 = 0;
    if (G2->nVrtx == 0)
        dumpnbpart2 = 0;

    rcoars->next_rcoarsen1 = rcoars1;
    rcoars->next_rcoarsen2 = rcoars2;
    rcoars1->previous_rcoarsen = rcoars;
    rcoars2->previous_rcoarsen = rcoars;

    for (i=1; i<=nVrtx; i++) {
        if (C->part[i] == part1)
            rcoars1->previous_index[rcoars->next_index[i]] = i;
        if (C->part[i] == part2)
            rcoars2->previous_index[rcoars->next_index[i]] = i;
    }

    //printf("Part: ");
    for (i=1; i<=nVrtx; i++) {
        if (C->part[i] == part1)
            C->part[i] = rcoars1->coars->part[rcoars->next_index[i]];
        else
            C->part[i] = dumpnbpart1 + rcoars2->coars->part[rcoars->next_index[i]];
        //printf("%d ", C->part[i]);
    }
    //printf("\n");

    int nbmerge, livesetpart;
    nbmerge = 0;
    if (opt.live_cut_merge) {
        nbmerge = 0;
        livesetpart = calcLiveSizePartMixMulti(igraph, C->part, dumpnbpart1-1, 10);
        idxType* part_save = (idxType*) malloc(sizeof(idxType) * (igraph->nVrtx+1));
        //printf("liveset before merge %d\n", livesetpart);
        while ((livesetpart < opt.maxliveset)&&(nbmerge < dumpnbpart2)) {
            for (i=1; i<=nVrtx; i++) {
                part_save[i] = C->part[i];
                if (C->part[i] >= dumpnbpart1)
                    C->part[i] -= 1;
            }
            nbmerge++;
            livesetpart = calcLiveSizePartMixMulti(igraph, C->part, dumpnbpart1-1, 10);
            //printf("liveset after merge %d\n", livesetpart);
        }
        if (nbmerge > 0) {
            nbmerge--;
            for (i = 1; i <= nVrtx; i++)
                C->part[i] = part_save[i];
        }
        if (opt.debug > 1) {
            //for (i=0; i<info->depth; i++)
            //    printf("@");
            printf("We did %d merge\n", nbmerge);
        }
    }

    //printf("second subpart done! size:%d\n", G2->nVrtx);
    if (opt.debug > -1) {
        //for (i=0; i<info->depth; i++)
        //    printf("@");
        //printf("dumps: %d %d\n",dumpnbpart1,dumpnbpart2);
    }
    //We rebuild and return final solution

    //----------------
    *resultingNbPart = dumpnbpart2 + dumpnbpart1 - nbmerge;
    //----------------

    opt.ratio = ratio;
    free(flag);
    free(traversal);

    info->depth--;
    info->timing_global += u_wseconds();
    info->timing_coars = info->rec1->timing_coars + info->rec2->timing_coars;
    info->timing_inipart = info->rec1->timing_inipart + info->rec2->timing_inipart;
    info->timing_uncoars = info->rec1->timing_uncoars + info->rec2->timing_uncoars;
    return rcoars;
}



