#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "scheduling.h"
#include "utils.h"
#include <limits.h>
#include <math.h>
#include "dgraphTraversal.h"

idxType calcLiveSizeMixMulti(dgraph *G, int rep, idxType* traversal, MLGP_option opt)
{
    idxType liveset = idxType_MAX, l;
    idxType* besttrav = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    int i, j;
    for (i=0; i < rep; i++) {
        l = calcLiveSizeMix(G, besttrav, opt);
        if (liveset > l){
            liveset = l;
            for (j = 1; j <= G->nVrtx; j++){
                traversal[j] = besttrav[j];
            }
        }
    }
    free(besttrav);
    return liveset;
}

idxType calcLiveSizeMix(dgraph *G, idxType* traversal, MLGP_option opt)
{
    idxType i;
    int print =0;
    idxType* DFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    DFStopsort(G, DFStop);
    idxType DFSlive = calcLiveSize(G, DFStop, opt);
    if (print) {
        printf("DFSlive: %d\n", DFSlive);
        for (i = 1; i <= G->nVrtx; i++)
            printf("%d ", (int) DFStop[i]);
        printf("\n");
    }

    idxType* randtop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    randTopsort(G, randtop);
    idxType randToplive = calcLiveSize(G, randtop, opt);
    if (print) {
        printf("randToplive: %d\n", randToplive);
        for (i = 1; i <= G->nVrtx; i++)
            printf("%d ", (int) randtop[i]);
        printf("\n");
    }

    idxType* BFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    BFStopsort(G, BFStop);
    idxType BFSlive = calcLiveSize(G, BFStop, opt);
    if (print) {
        printf("BFSlive: %d\n", BFSlive);
        for (i = 1; i <= G->nVrtx; i++)
            printf("%d ", (int) BFStop[i]);
        printf("\n");
    }

    idxType* mixtop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    mixtopsort(G, mixtop, 50, 1);
    idxType mixlive = calcLiveSize(G, mixtop, opt);
    if (print) {
        printf("mixlive: %d\n", mixlive);
        for (i = 1; i <= G->nVrtx; i++)
            printf("%d ", (int) mixtop[i]);
        printf("\n");
    }

    idxType liveset = idxType_MAX;
    if (liveset > DFSlive){
        liveset = DFSlive;
        for (i = 1; i <= G->nVrtx; i++){
            traversal[i] = DFStop[i];
        }
    }
    if (liveset > BFSlive){
        liveset = BFSlive;
        for (i = 1; i <= G->nVrtx; i++){
            traversal[i] = BFStop[i];
        }
    }
    if (liveset > randToplive){
        liveset = randToplive;
        for (i = 1; i <= G->nVrtx; i++){
            traversal[i] = randtop[i];
        }
    }
    if (liveset > mixlive){
        liveset = mixlive;
        for (i = 1; i <= G->nVrtx; i++){
            traversal[i] = mixtop[i];
        }
    }
    free(DFStop);
    free(BFStop);
    free(randtop);
    free(mixtop);

    return liveset;
}

idxType calcLiveSizeMix_with_part_Multi(dgraph *G, idxType* part, int nbpart, idxType* traversal, int rep, MLGP_option opt)
{
    idxType liveset = idxType_MAX, l;
    idxType* besttrav = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    int i,j;
    for (i=0; i < rep; i++) {
        l = calcLiveSizeMix_with_part(G, part, nbpart, besttrav, opt);
        if (liveset > l){
            liveset = l;
            for (j = 1; j <= G->nVrtx; j++){
                traversal[j] = besttrav[j];
            }
        }
    }
    free(besttrav);
    return liveset;
}

idxType calcLiveSizeMix_with_part(dgraph *G, idxType* part, int nbpart, idxType* toporder, MLGP_option opt)
{
    idxType i;

    idxType* DFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    DFStopsort_with_part(G, part, nbpart, DFStop);
    idxType DFSlive = calcLiveSize(G, DFStop, opt);

    idxType* randDFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    randDFStopsort_with_part(G, part, nbpart, randDFStop);
    idxType randDFSlive = calcLiveSize(G, randDFStop, opt);

    idxType* BFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    BFStopsort_with_part(G, part, nbpart, BFStop);
    idxType BFSlive = calcLiveSize(G, BFStop, opt);

    idxType* randBFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    randBFStopsort_with_part(G, part, nbpart, randBFStop);
    idxType randBFSlive = calcLiveSize(G, randBFStop, opt);

    idxType liveset = idxType_MAX;
    if (DFSlive < liveset) {
        liveset = DFSlive;
        for (i=1; i <= G->nVrtx; i++)
            toporder[i] = DFStop[i];
    }
    if (randDFSlive < liveset) {
        liveset = randDFSlive;
        for (i=1; i <= G->nVrtx; i++)
            toporder[i] = randDFStop[i];
    }
    if (BFSlive < liveset) {
        liveset = BFSlive;
        for (i=1; i <= G->nVrtx; i++)
            toporder[i] = BFStop[i];
    }
    if (randBFSlive < liveset) {
        liveset = randBFSlive;
        for (i=1; i <= G->nVrtx; i++)
            toporder[i] = randBFStop[i];
    }
    free(DFStop);
    free(randDFStop);
    free(BFStop);
    free(randBFStop);

    return liveset;
}


/*idxType calcLiveSize(dgraph *g,idxType* toporder, MLGP_option opt)
{
    idxType i,j,k,node, liveSize = 0, maxLiveSize = 0;
    idxType* outgoing = (idxType*) calloc(g->nVrtx + 2, sizeof(idxType));

    for (j = 1; j <= g->nVrtx; j++) {
        i = toporder[j];
        liveSize++;
        maxLiveSize = liveSize > maxLiveSize ? liveSize : maxLiveSize;
        //printf("%d is alive now\n", i);
        for (k = g->inStart[i]; k <= g->inEnd[i]; k++){
            node = g->in[k];
            outgoing[node]++;
            if (outgoing[node] >= g->outEnd[node] - g->outStart[node] + 1)
                liveSize--;
        }
        if (g->outEnd[i] < g->outStart[i])
            liveSize--;
    }

    free(outgoing);
    return maxLiveSize;
}*/

idxType calcLiveSizeNatural(dgraph *G, MLGP_option opt)
{
    idxType* traversal = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    idxType i;
    for (i=1; i <= G->nVrtx; i++)
        traversal[i] = i;
    checkTopsort(G, traversal);
    idxType ls = calcLiveSize(G, traversal, opt);
    free(traversal);
    return ls;
}

idxType calcLiveSizeNatural_with_part(dgraph *G, idxType *part, int nbpart, MLGP_option opt)
{
    idxType* traversal = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    idxType* toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
    topSortOnParts(G, part, toporderpart, nbpart);
    natural_with_part(G, traversal, part, toporderpart, nbpart);
    checkTopsort(G, traversal);
    idxType ls = calcLiveSize(G, traversal, opt);
    free(toporderpart);
    free(traversal);
    return ls;
}

idxType calcLiveSize_Lmax(dgraph *g,idxType* toporder)
{
    idxType i,j,k,node, liveSize = 0, maxLiveSize = 0;
    idxType* outgoing = (idxType*) calloc(g->nVrtx + 2, sizeof(idxType));
    int* marked = (int*) calloc(g->nVrtx+1, sizeof(int));

    idxType nbhollowinput = 0;
    int* ishollowinput = (int*) calloc(g->nVrtx+1, sizeof(int));

    for (i = 1; i <= g->nVrtx; i++)
        if(g->hollow[i] == 1)
            for(j = g->outStart[i]; j <= g->outEnd[i]; j++)
                if(g->hollow[g->out[j]] == 0){
                    nbhollowinput += 1; //here unitary
                    ishollowinput[i] = 1;
                    break;
                }
    liveSize = maxLiveSize = nbhollowinput;

    for (j = 1; j <= g->nVrtx; j++) {
        i = toporder[j];
        marked[i] = 1;
        if (g->hollow[i])
            continue;
        liveSize++;
        maxLiveSize = liveSize > maxLiveSize ? liveSize : maxLiveSize;
        for (k = g->inStart[i]; k <= g->inEnd[i]; k++){
            node = g->in[k];
            if (marked[node] == 0)
                u_errexit("In calcLiveSize_with_keep_hollow, traversal is not a topological order (executing %d and %d not done)\n", i, node);
            outgoing[node]++;
            if ((outgoing[node] >= g->outEnd[node] - g->outStart[node] + 1)&&(g->hollow[node] == 0))
                liveSize--;
        }
        if ((g->outEnd[i] < g->outStart[i])&&(g->hollow[node] == 0))
            liveSize--;
    }

    free(outgoing);
    free(marked);
    return maxLiveSize;
}

idxType calcLiveSize_L(dgraph *g,idxType* toporder)
{
    idxType i,j,k,node, liveSize = 0, maxLiveSize = 0;
    idxType* outgoing = (idxType*) calloc(g->nVrtx + 2, sizeof(idxType));
    int* marked = (int*) calloc(g->nVrtx+1, sizeof(int));

    for (j = 1; j <= g->nVrtx; j++) {
        i = toporder[j];
        liveSize++; //here unitary
        marked[i] = 1;
        maxLiveSize = liveSize > maxLiveSize ? liveSize : maxLiveSize;
        for (k = g->inStart[i]; k <= g->inEnd[i]; k++){
            node = g->in[k];
            if (marked[node] == 0)
                u_errexit("In calcLiveSize_with_keep_hollow, traversal is not a topological order (executing %d and %d not done)\n", i, node);
            outgoing[node]++;
            if ((outgoing[node] >= g->outEnd[node] - g->outStart[node] + 1)&&(g->hollow[node] == 0))
                liveSize--; //here unitary
        }
        if ((g->outEnd[i] < g->outStart[i])&&(g->hollow[node] == 0))
            liveSize--; //here unitary
    }

    free(outgoing);
    free(marked);
    return maxLiveSize;
}

idxType calcLiveSize_Lmin(dgraph *g,idxType* toporder){
    idxType i,j,k,node, liveSize = 0, maxLiveSize = 0;
    idxType* outgoing = (idxType*) calloc(g->nVrtx + 2, sizeof(idxType));
    int* marked = (int*) calloc(g->nVrtx+1, sizeof(int));

    for (j = 1; j <= g->nVrtx; j++)
        for (k = g->outStart[j]; k <= g->outEnd[j]; k++)
            if (g->hollow[g->out[k]] == 0){
                outgoing[j]++;
            }

    for (j = 1; j <= g->nVrtx; j++) {
        i = toporder[j];
        marked[i] = 1;
        if (g->hollow[i])
            continue;
        liveSize++;//here unitary

        maxLiveSize = liveSize > maxLiveSize ? liveSize : maxLiveSize;
        for (k = g->inStart[i]; k <= g->inEnd[i]; k++){
            node = g->in[k];
            if (marked[node] == 0)
                u_errexit("In calcLiveSize_with_discard_hollow, traversal is not a topological order (executing %d and %d not done)\n", i, node);
            outgoing[node]--;
            if (outgoing[node] ==0)
                liveSize--; //here unitary
        }
        if (g->outEnd[i] < g->outStart[i])
            liveSize--; //here unitary
    }

    free(outgoing);
    free(marked);
    return maxLiveSize;
}


idxType calcLiveSizePartMixMulti(dgraph *G, idxType* part, idxType part_idx, int rep)
{
    idxType liveset = idxType_MAX, l;
    int i;
    for (i=0; i < rep; i++) {
        l = calcLiveSizePartMix(G, part, part_idx);
        liveset = liveset < l ? liveset : l;
    }

    return liveset;
}

idxType calcLiveSizePartMix(dgraph *G, idxType* part, idxType part_idx)
{
    idxType* DFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    DFStopsort(G, DFStop);
    idxType DFSlive = calcLiveSizePart(G, DFStop, part, part_idx);

    idxType* BFStop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    BFStopsort(G, BFStop);
    idxType BFSlive = calcLiveSizePart(G, BFStop, part, part_idx);

    idxType* randtop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    randTopsort(G, randtop);
    idxType randtoplive = calcLiveSizePart(G, randtop, part, part_idx);

    idxType* mixtop = (idxType*) malloc(sizeof(idxType)*(G->nVrtx + 2));
    mixtopsort(G, mixtop, 50, 1);
    idxType mixlive = calcLiveSizePart(G, mixtop, part, part_idx);

    idxType liveset = idxType_MAX;
    liveset = liveset < DFSlive ? liveset : DFSlive;
    liveset = liveset < BFSlive ? liveset : BFSlive;
    liveset = liveset < randtoplive ? liveset : randtoplive;
    liveset = liveset < mixlive ? liveset : mixlive;
    free(DFStop);
    free(BFStop);
    free(randtop);
    free(mixtop);

    return liveset;
}

idxType calcLiveSizePart(dgraph *G, idxType* toporder, idxType* part, idxType part_idx)
{
    idxType* incoming = (idxType*) calloc((G->nVrtx + 2), sizeof(idxType));
    idxType* outgoing = (idxType*) calloc((G->nVrtx + 2), sizeof(idxType));
    int i=0,j=0;
    int r=0,rend=0;
    int looper=0,lend=0;
    idxType maxLiveSize=0;
    idxType liveSize=0;
    int no = 1;
    idxType node;

    for (j= 1; j<= G->nVrtx; j++){
        node = toporder[j];
        //printf("Node %d\n", (int) i);
        if (G->hollow[node]) {
            //printf("%d is hollow \n",i);
            no=1;
            if (G->inEnd[i]<G->inStart[i]) // no in edge
                no=0; // don't skip
            /*
            for (k=G->inStart[i];k<=G->inEnd[i];++k) {
                if (G->hollow[G->in[k]]==0) { // out node, skip
                    no=1;
                    break;
                }
            }
            */
            if (no) { // skip
                //printf("i %d no\n",i);
                continue;
            }
        }
        int no_in_part = 1;
        for (r = G->inStart[node]; r <= G->inEnd[node]; r++)
            if (part[G->in[r]] == part_idx)
                no_in_part = 0;
        int no_out_part = 1;
        for (r = G->outStart[node]; r <= G->outEnd[node]; r++)
            if (part[G->out[r]] == part_idx)
                no_out_part = 0;
        if ((part[node] != part_idx)&&no_out_part&&no_in_part) {
            continue;
        }
        ++liveSize;
        //printf("%d is alive now\n", i);
        if (liveSize>maxLiveSize) {
            maxLiveSize=liveSize;
            //printf("maxLiveSize updated to %d\n", maxLiveSize);
        }
        r = G->inStart[node];
        rend = G->inEnd[node];
        while(r <= rend) {
            ++outgoing[G->in[r]];
            //printf("increasing out of %d %d\n",G->in[r], outgoing[G->in[r]]);
            if (outgoing[G->in[r]]>=G->outEnd[G->in[r]]-G->outStart[G->in[r]]+1) {
                //printf("%d dies\n", G->in[r] );
                --liveSize;
            }
            else{
                //printf(" %d still alive %d --  %d\n", G->in[r],outgoing[G->in[r]],G->outEnd[G->in[r]]-G->outStart[G->in[r]]+1);
            }
            ++r;
        }
        looper = G->outStart[node];
        lend = G->outEnd[node];
        if (lend<looper) {
            //printf("no child, %d dies, too\n", (int) i);
            --liveSize;
        }
        while(looper<=lend) {
            ++incoming[G->out[looper]];
            ++looper;
            //printf("looper %d\n", (int) looper);

        }
    }

    //printf("end\n");
    free(outgoing);
    return maxLiveSize;
}

idxType liveSize(dgraph *G, MLGP_option opt){
    idxType *toporder = (idxType*) malloc((G->nVrtx + 2) * sizeof(idxType));
    idxType i, ls=-1;
    switch (opt.live_traversal){
        case LIVE_TRAV_NAT:
            for (i=1; i <= G->nVrtx; i++)
                toporder[i] = i;
            break;

        case LIVE_TRAV_DFS:
            DFStopsort(G, toporder);
            break;

        case LIVE_TRAV_BFS:
            BFStopsort(G, toporder);
            break;

        case LIVE_TRAV_MIX:
            mixtopsort(G, toporder, 50, 1);
            break;

        case LIVE_TRAV_RAND:
            randTopsort(G, toporder);
            break;

        default:
            u_errexit("Live_traversal %d is not recognized!\n", opt.live_traversal);
    }

    switch (opt.live_comp){
        case 0:
            ls = calcLiveSize_Lmin(G, toporder);
            break;

        case 1:
            ls = calcLiveSize_L(G, toporder);
            break;

        case 2:
            ls = calcLiveSize_Lmax(G, toporder);
            break;

        default:
            u_errexit("Live_comp %d is not recognized!\n", opt.live_comp);
    }

    free(toporder);
    return ls;
}

idxType liveSize_with_part(dgraph *G, idxType* part, int nbpart, MLGP_option opt){
    idxType *toporder = (idxType*) malloc((G->nVrtx + 2) * sizeof(idxType));
    idxType i, ls=-1;
    idxType *toporderpart;
    switch (opt.live_traversal){
        case LIVE_TRAV_NAT:
            toporderpart = (idxType*) malloc (sizeof(idxType) * nbpart);
            topSortOnParts(G, part, toporderpart, nbpart);
            natural_with_part(G, toporder, part, toporderpart, nbpart);
            free(toporderpart);
            break;

        case LIVE_TRAV_DFS:
            DFStopsort_with_part(G, part, nbpart, toporder);
            break;

        case LIVE_TRAV_BFS:
            BFStopsort_with_part(G, part, nbpart, toporder);
            break;

        case LIVE_TRAV_MIX:
            //TODO mixtopsort_with_part
            DFStopsort_with_part(G, part, nbpart, toporder);
            break;

        case LIVE_TRAV_RAND:
            randBFStopsort_with_part(G, part, nbpart, toporder);
            break;

        default:
            u_errexit("Live_traversal %d is not recognized!\n", opt.live_traversal);
    }

    switch (opt.live_comp){
        case 0:
            ls = calcLiveSize_Lmin(G, toporder);
            break;

        case 1:
            ls = calcLiveSize_L(G, toporder);
            break;

        case 2:
            ls = calcLiveSize_Lmax(G, toporder);
            break;

        default:
            u_errexit("Live_comp %d is not recognized!\n", opt.live_comp);
    }

    free(toporder);
    return ls;
}

idxType calcLiveSize(dgraph *G,idxType* toporder, MLGP_option opt){
    idxType ls = -1;
    switch (opt.live_comp){
        case 0:
            ls = calcLiveSize_Lmin(G, toporder);
            break;

        case 1:
            ls = calcLiveSize_L(G, toporder);
            break;

        case 2:
            ls = calcLiveSize_Lmax(G, toporder);
            break;

        default:
            u_errexit("Live_comp %d is not recognized!\n", opt.live_comp);
    }
    return ls;
}



